/**
 * <p>Title: OWAMP Protocol</p>
 * <p>Description: This section of this project contains the class used to implement the OWAMP Control-Client.</p>
 * <p>Copyright: Copyright (c) 2005</p>
 * <p>Company: University of Aveiro / Institute of Telecommunications - Aveiro pole</p>
 * @authors: Helder Veiga and Teresa Pinho
 * @version: 2.1
 * @Last updated: 9-12-2005
 *
 * Copyright (C) 2005  Helder Veiga and Teresa Pinho
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.  A HTML version of
 * the GNU General Public License can be seen at
 * http://www.gnu.org/licenses/gpl.html.
 *
 */

import java.io.*;
import java.net.*;
import java.util.*;
import java.sql.*;  // import needed for JDBC access

public class OWAMP_ControlClient extends OWAMP_Base{

  // Constructors

  /**
   * Constructor for class OWAMP_ControlClient.
   * @param serverAddress InetAddress - represents the Server address.
   * @param serverTCPportNumber int - represents the Server's TCP port to use in OWAMP-Control communications
   * with the Server.
   * @param systemInReader BufferedReader - represents the System.in
   * @param hiResTimer HiResTimer - represents the high resolution timer used to get the time.
   */

  public OWAMP_ControlClient(InetAddress serverAddress, int serverTCPportNumber, BufferedReader systemInReader, HiResTimer hiResTimer){

    rand = new Random();
    this.serverAddress = serverAddress;
    this.serverTCPportNumber = serverTCPportNumber;
    buffer = new byte[DEFAULT_PACKET_LENGTH];

    arrayOfSocketToSender = new Socket[1];
    arrayOfSenderTCPportNumber = new int[1];
    arrayOfdataRS = new RequestSessionToBeSend[1];
    arrayOfdataSSD = new ScheduleSlotDescriptionsToBeSend[1][];
    socketToServer = new Socket();

    seed = 0;
    this.hiResTimer = hiResTimer;

    dataSSs = new StopSessionsToBeSend((short)0,0);
    dataSSr = new StopSessionsReceiver();

    this.systemInReader = systemInReader;

    try{
      serverSocket = new ServerSocket();    // Used to verify if a socket connection is closed
                                            // using method isSocketConnected.
    }catch(IOException e){
      System.err.println(e);
    }

  }

  /**
   * Constructor for class OWAMP_ControlClient.
   * The user can execute this module with all the configurations parameters in the command line. In
   * this case the user must instroduce 29 parameters. This mode of execution is also used in the
   * Web interface.
   *
   * @param serverAddress String - represents the Server address.
   * The server address should be in the following format:
   *
   *     IPv4_address:TCPportNumber    // for IPv4 address
   *     [IPv6_address]:TCPportNumber  // for IPv6 address
   *
   * The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the
   * Control-Client and the respective OWAMP-Server. If the user wants to use the default TCPportNumber (22368)
   * the server address can be in the following format:
   *
   *     IPv4_address    // for IPv4 address
   *     [IPv6_address]  // for IPv6 address
   *
   * @param _mbz_iPVN String
   * @param _confSender String
   * @param _confReceiver String
   * @param _numberOfScheduleSlots String
   * @param _numberOfPackets String
   * @param _senderPort String
   * @param _receiverPort String
   * @param _senderAddress String - The sender address should be in the following format:
   *
   *      IPv4_address:TCPportNumber    // for IPv4 address
   *      [IPv6_address]:TCPportNumber  // for IPv6 address
   *
   * The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the
   * Control-Client and the respective OWAMP-Sender. If the user wants to use the default TCPportNumber (4181)
   * the sender address can be in the following format:
   *
   *      IPv4_address    // for IPv4 address
   *      [IPv6_address]  // for IPv6 address
   *
   * @param _receiverAddress String - The receiver address should be in the following format:
   *
   *      IPv4_address     // for IPv4 address
   *      [IPv6_address]   // for IPv6 address
   *
   * The TCP port to be used in OWAMP-Control communication between the Server and receiver should be defined
   * when running OWAMP_Server, or the default TCP port (28181) will be used.
   *
   * @param _paddingLenght String
   * @param _startTime long
   * @param _timeOut String
   * @param _typePDescriptor_Type String
   * @param _typePDescriptor_Value String
   * @param _slotType String
   * @param _slotParameter String
   * @param systemInReader BufferedReader - represents the System.in
   * @param hiResTimer HiResTimer - represents the high resolution timer used to get the time.
   */

  public OWAMP_ControlClient(String serverAddress,String _confSender,
                             String _confReceiver, String _numberOfScheduleSlots, String _numberOfPackets,
                             String _senderPort, String _receiverPort, String _senderAddress,
                             String _receiverAddress, String _paddingLenght, long _startTime, String _timeOut,
                             String _typePDescriptor_Type, String _typePDescriptor_Value, String _slotType,
                             String _slotParameter, BufferedReader systemInReader, HiResTimer hiResTimer){

    rand = new Random();

    String address =  serverAddress.trim();
    String addressAux = address;
    serverTCPportNumber = DEFAULT_SEVER_PORT;  // TCPportNumber to use in OWAMP-Control with Server.
    boolean IPv6 = false;
    int i=0;

    try{

      if(address.indexOf("[") != -1){  // IPv6 address
        IPv6 = true;
        if(address.indexOf("]:") != -1){  // TCPportNumber to use in OWAMP-Control with sender specified.
          serverTCPportNumber = Integer.parseInt(address.substring(address.indexOf("]:")+2));
        }
        address = address.substring(address.indexOf("[")+1,address.indexOf("]"));
      }else{  // IPv4 address
        if(address.indexOf(":") != -1){  // TCPportNumber to use in OWAMP-Control with sender specified.
          serverTCPportNumber = Integer.parseInt(address.substring(address.indexOf(":")+1));
          address = address.substring(0,address.indexOf(":"));
        }
      }

      if (serverTCPportNumber < 0 || serverTCPportNumber >= 65536) {
        System.err.println("Error: Sender TCP port number out of range (" +serverTCPportNumber+").");
        System.err.println();
        System.exit(1);
      }

      InetAddress inetAddress[] = InetAddress.getAllByName(address);     // The array of all the IP addresses for a given host name.

      for(i=0;i<inetAddress.length;i++){
        if(IPv6 && (inetAddress[i].getAddress().length == 16)){  // IPv6 address
          this.serverAddress = inetAddress[i];
          break;
        }

        if(!IPv6 && (inetAddress[i].getAddress().length == 4)){  // IPv4 address
          this.serverAddress = inetAddress[i];
          break;
        }

        if(i == (inetAddress.length-1)){  // last cicle.
          System.err.println("Error: There isn't a valid IPv"+(IPv6?6:4)+" address for the server address introduced ("+address+").");
          System.err.println();
          System.exit(1);
        }
      }

    }catch(NumberFormatException e){
      System.err.println("Error: The the server address read is not a valid address ("+addressAux+").");
      System.err.println("The server address should be in the following format:\n\n"+
                         "    IPv4_address:TCPportNumber    // for IPv4 address\n"+
                         "    [IPv6_address]:TCPportNumber  // for IPv6 address\n\n"+
                         "The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the \n"+
                         "Control-Client and the respective OWAMP-Server. If the user wants to use the default TCPportNumber (22368) \n"+
                         "the server address can be in the following format:\n\n"+
                         "    IPv4_address    // for IPv4 address\n"+
                         "    [IPv6_address]  // for IPv6 address\n");
      System.err.println();
      System.exit(1);
    }catch(Exception e2){
      System.err.println("Error: The the server address read is not a valid address ("+addressAux+").");
      System.err.println("The server address should be in the following format:\n\n"+
                         "    IPv4_address:TCPportNumber    // for IPv4 address\n"+
                         "    [IPv6_address]:TCPportNumber  // for IPv6 address\n\n"+
                         "The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the \n"+
                         "Control-Client and the respective OWAMP-Server. If the user wants to use the default TCPportNumber (22368) \n"+
                         "the server address can be in the following format:\n\n"+
                         "    IPv4_address    // for IPv4 address\n"+
                         "    [IPv6_address]  // for IPv6 address\n");
      System.err.println();
      System.exit(1);
    }

    buffer = new byte[DEFAULT_PACKET_LENGTH];
    arrayOfSocketToSender = new Socket[1];
    arrayOfSenderTCPportNumber = new int[1];
    arrayOfdataRS = new RequestSessionToBeSend[1];
    arrayOfdataRS[0] = new RequestSessionToBeSend();
    arrayOfdataSSD = new ScheduleSlotDescriptionsToBeSend[1][];

    socketToServer = new Socket();

    seed = 0;
    this.hiResTimer = hiResTimer;

    dataSSs = new StopSessionsToBeSend((short)0,0);
    dataSSr = new StopSessionsReceiver();

    this.systemInReader = systemInReader;

    numRS = 0;    // Number of Request Session made - there are no Request Session yet.
    mode = 1;     // The mode value to be used in the current OWAMP control and test sessions.

    try {
      arrayOfdataRS[numRS].confSender = (byte)Integer.parseInt(_confSender);
    }catch(NumberFormatException e){
      System.err.println(e);
      arrayOfdataRS[numRS].confSender = -1;
    }

    if((arrayOfdataRS[numRS].confSender != 0)&&(arrayOfdataRS[numRS].confSender != 1)){
      System.err.println("Error: The value of confSender read is not valid.");
      System.exit(1);
    }


    try {
      arrayOfdataRS[numRS].confReceiver = (byte)Integer.parseInt(_confReceiver);
    }catch(NumberFormatException e){
      System.err.println(e);
      arrayOfdataRS[numRS].confReceiver = -1;
    }

    if((arrayOfdataRS[numRS].confReceiver != 0)&&(arrayOfdataRS[numRS].confReceiver != 1)){
      System.err.println("Error: The value of confReceiver read is not valid.");
      System.exit(1);
    }

    try {
      arrayOfdataRS[numRS].numberOfScheduleSlots = Integer.parseInt(_numberOfScheduleSlots);
    }catch(NumberFormatException e){
      System.err.println(e);
      arrayOfdataRS[numRS].numberOfScheduleSlots = -1;
    }
    if(arrayOfdataRS[numRS].numberOfScheduleSlots == -1){
      System.err.println("Error: The value of numberOfScheduleSlots read is not valid.");
      System.exit(1);
    }

    try {
      arrayOfdataRS[numRS].numberOfPackets = Integer.parseInt(_numberOfPackets);
    }catch(NumberFormatException e){
      System.err.println(e);
      arrayOfdataRS[numRS].numberOfPackets = -1;
    }
    if(arrayOfdataRS[numRS].numberOfPackets == -1){
      System.err.println("Error: The value of numberOfPackets read is not valid.");
      System.exit(1);
    }

    try {
      arrayOfdataRS[numRS].senderPort = Integer.parseInt(_senderPort);
    }catch(NumberFormatException e){
      System.err.println(e);
      System.err.println("Error: The value of senderPort read is not valid.");
      System.err.println();
      System.exit(1);
    }

    if(arrayOfdataRS[numRS].senderPort < 0 || arrayOfdataRS[numRS].senderPort >= 65536){
      System.err.println("Error: Sender port number out of range " + arrayOfdataRS[numRS].senderPort);
      System.exit(1);
    }

    try {
      arrayOfdataRS[numRS].receiverPort = Integer.parseInt(_receiverPort);
    }catch(NumberFormatException e){
      System.err.println(e);
      System.err.println("Error: The value of receiverPort read is not valid.");
      System.err.println();
      System.exit(1);
    }

    if(arrayOfdataRS[numRS].receiverPort < 0 || arrayOfdataRS[numRS].receiverPort >= 65536){
      System.err.println("Error: Receiver port number out of range " + arrayOfdataRS[numRS].receiverPort);
      System.exit(1);
    }

    // sender address

    /* The sender address should be in the following format:

          IPv4_address:TCPportNumber    // for IPv4 address
          [IPv6_address]:TCPportNumber  // for IPv6 address

       The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the
       Control-Client and the respective OWAMP-Sender. If the user wants to use the default TCPportNumber
       the sender address can be in the following format:

          IPv4_address    // for IPv4 address
          [IPv6_address]  // for IPv6 address
    */

    address =  _senderAddress;
    int senderTCPportNumber = DEFAULT_SENDER_PORT;  // TCPportNumber to use in OWAMP-Control with sender.
    IPv6 = false;

    try{

      if(address.indexOf("[") != -1){  // IPv6 address
        IPv6 = true;
        if(address.indexOf("]:") != -1){  // TCPportNumber to use in OWAMP-Control with sender specified.
          senderTCPportNumber = Integer.parseInt(address.substring(address.indexOf("]:")+2));
        }
        address = address.substring(address.indexOf("[")+1,address.indexOf("]"));
      }else{  // IPv4 address
        if(address.indexOf(":") != -1){  // TCPportNumber to use in OWAMP-Control with sender specified.
          senderTCPportNumber = Integer.parseInt(address.substring(address.indexOf(":")+1));
          address = address.substring(0,address.indexOf(":"));
        }
      }

      if (senderTCPportNumber < 0 || senderTCPportNumber >= 65536) {
        System.err.println("Error: Sender TCP port number out of range (" +senderTCPportNumber+").");
        System.err.println();
        System.exit(1);
      }
      arrayOfSenderTCPportNumber[numRS] = senderTCPportNumber;

      InetAddress inetAddress[] = InetAddress.getAllByName(address);     // The array of all the IP addresses for a given host name.

      for(i=0;i<inetAddress.length;i++){
        if(IPv6 && (inetAddress[i].getAddress().length == 16)){  // IPv6 address
          senderAddress = inetAddress[i].getAddress();
          break;
        }

        if(!IPv6 && (inetAddress[i].getAddress().length == 4)){  // IPv4 address
          senderAddress = inetAddress[i].getAddress();
          break;
        }

        if(i == (inetAddress.length-1)){  // last cicle.
          System.err.println("Error: There isn't a valid IPv"+(IPv6?6:4)+" address for the sender address introduced ("+address+").");
          System.err.println();
          System.exit(1);
        }
      }

      // The sender and receiver addresses can only be both loopback addresses if the Control-Client
      // is on the same machine were the Server is running.
      if(InetAddress.getByAddress(senderAddress).isLoopbackAddress() && !isAnyLocalAddress(this.serverAddress)){
        System.err.println("Error: The sender address can only be loopback address if the Control-Client is on the same machine were the Server is running.");
        System.err.println();
        System.exit(1);
      }

      System.arraycopy(senderAddress,0,arrayOfdataRS[numRS].senderAddress,0,senderAddress.length);

    }catch(NumberFormatException e){
      System.err.println("Error: The the sender address read is not a valid address ("+_senderAddress+").");
      System.err.println("The sender address should be in the following format:\n\n"+
                         "    IPv4_address:TCPportNumber    // for IPv4 address\n"+
                         "    [IPv6_address]:TCPportNumber  // for IPv6 address\n\n"+
                         "The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the \n"+
                         "Control-Client and the respective OWAMP-Sender. If the user wants to use the default TCPportNumber \n"+
                         "the sender address can be in the following format:\n\n"+
                         "    IPv4_address    // for IPv4 address\n"+
                         "    [IPv6_address]  // for IPv6 address\n");
      System.err.println();
      System.exit(1);
    }catch(Exception e2){
      System.err.println("Error: The the sender address read is not a valid address ("+_senderAddress+").");
      System.err.println("The sender address should be in the following format:\n\n"+
                         "    IPv4_address:TCPportNumber    // for IPv4 address\n"+
                         "    [IPv6_address]:TCPportNumber  // for IPv6 address\n\n"+
                         "The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the \n"+
                         "Control-Client and the respective OWAMP-Sender. If the user wants to use the default TCPportNumber \n"+
                         "the sender address can be in the following format:\n\n"+
                         "    IPv4_address    // for IPv4 address\n"+
                         "    [IPv6_address]  // for IPv6 address\n");
      System.err.println();
      System.exit(1);
    }


    // receiver address

    /* The receiver address should be in the following format:

           IPv4_address     // for IPv4 address
           [IPv6_address]   // for IPv6 address

       The TCP port to be used in OWAMP-Control communication between the Server and receiver should be defined
       when running OWAMP_Server, or the default TCP port will be used.
    */

    address =  _receiverAddress;
    IPv6 = false;

    try{

      if(address.indexOf("[") != -1){  // IPv6 address
        IPv6 = true;
        arrayOfdataRS[numRS].mbz_iPVN = (byte)6;
        if(address.indexOf("]:") != -1){  // TCPportNumber to use in OWAMP-Control with receiver specified.
          System.err.println("Error: The TCP port to be used in OWAMP-Control communication between the Server and receiver should be defined when running OWAMP_Server.");
          System.err.println();
          System.exit(1);
        }
        address = address.substring(address.indexOf("[")+1,address.indexOf("]"));
      }else{  // IPv4 address
        arrayOfdataRS[numRS].mbz_iPVN = (byte)4;
        if(address.indexOf(":") != -1){  // TCPportNumber to use in OWAMP-Control with receiver specified.
          System.err.println("Error: The TCP port to be used in OWAMP-Control communication between the Server and receiver should be defined when running OWAMP_Server.");
          System.err.println();
          System.exit(1);
        }
      }

      InetAddress inetAddress[] = InetAddress.getAllByName(address);     // The array of all the IP addresses for a given host name.

      for(i=0;i<inetAddress.length;i++){
        if(IPv6 && (inetAddress[i].getAddress().length == 16)){  // IPv6 address
          receiverAddress = inetAddress[i].getAddress();
          break;
        }

        if(!IPv6 && (inetAddress[i].getAddress().length == 4)){  // IPv4 address
          receiverAddress = inetAddress[i].getAddress();
          break;
        }

        if(i == (inetAddress.length-1)){  // last cicle.
          System.err.println("Error: There isn't a valid IPv"+(IPv6?6:4)+" address for the receiver address introduced ("+address+").");
          System.err.println();
          System.exit(1);
        }
      }

      // If sender address is loopback address receiver should be loopback address too. If receiver address is loopback address sender should be loopback address too.
      if((InetAddress.getByAddress(senderAddress).isLoopbackAddress()&&!InetAddress.getByAddress(receiverAddress).isLoopbackAddress())||(!InetAddress.getByAddress(senderAddress).isLoopbackAddress()&&InetAddress.getByAddress(receiverAddress).isLoopbackAddress())){
        if(InetAddress.getByAddress(senderAddress).isLoopbackAddress())
          System.err.println("Error: The sender address is loopback address and receiver address isn't.");
        if(InetAddress.getByAddress(receiverAddress).isLoopbackAddress())
          System.err.println("Error: The receiver address is loopback address and sender address isn't.");
        System.err.println();
        System.exit(1);
      }

      if(senderAddress.length != receiverAddress.length){  // The sender and receiver address should have the same IP version.
        System.err.println("Error: The sender and receiver address should have the same IP version.");
        System.err.println();
        System.exit(1);
      }

      System.arraycopy(receiverAddress,0,arrayOfdataRS[numRS].receiverAddress,0,receiverAddress.length);

    }catch(NumberFormatException e){
      System.err.println("Error: The the receiver address read is not a valid address ("+_receiverAddress+").");
      System.err.println("The receiver address should be in the following format:\n\n"+
                         "    IPv4_address    // for IPv4 address\n"+
                         "    [IPv6_address]  // for IPv6 address\n");
      System.err.println();
      System.exit(1);
    }catch(Exception e2){
      System.err.println("Error: The the receiver address read is not a valid address ("+_receiverAddress+").");
      System.err.println("The receiver address should be in the following format:\n\n"+
                         "    IPv4_address    // for IPv4 address\n"+
                         "    [IPv6_address]  // for IPv6 address\n");
      System.err.println();
      System.exit(1);
    }


    try {
      arrayOfdataRS[numRS].paddingLenght = Integer.parseInt(_paddingLenght);
    }catch(NumberFormatException e){
      System.err.println(e);
      arrayOfdataRS[numRS].paddingLenght = -1;
    }
    if(arrayOfdataRS[numRS].paddingLenght == -1){
      System.err.println("Error: The value of paddingLenght read is not valid.");
      System.exit(1);
    }

    arrayOfdataRS[numRS].startTime = new Timestamp(_startTime, (short)1); // Time in milliseconds since 1Jan1970

    double timeOut = 0;
    try {
      timeOut = Double.parseDouble(_timeOut);
    }catch(NumberFormatException e){
      System.err.println(e);
      timeOut = -1;
    }
    if(timeOut < 0){
      System.err.println("Error: The value of timeOut read is not valid.");
      System.exit(1);
    }

    arrayOfdataRS[numRS].timeout = new Timestamp(timeOut, (short)0);    // Timeout in seconds since 1Jan1900 = timeout seconds * 1000


    byte typePDescriptor_Type = 0;
    try {
      typePDescriptor_Type = (byte)Integer.parseInt(_typePDescriptor_Type);
    }catch(NumberFormatException e){
      System.err.println(e);
      typePDescriptor_Type = -1;
    }
    if((typePDescriptor_Type == -1)||(typePDescriptor_Type != 0)&&(typePDescriptor_Type != 1)){
      System.err.println("Error: The type of typePDescriptor read is not valid.");
      System.exit(1);
    }

    int typePDescriptor_Value = 0;
    try {
      typePDescriptor_Value = Integer.parseInt(_typePDescriptor_Value);
    }catch(NumberFormatException e){
      System.err.println(e);
      typePDescriptor_Value = -1;
    }

    if(typePDescriptor_Value == -1){
      System.err.println("Error: The value of typePDescriptor read is not valid.");
      System.exit(1);
    }

    arrayOfdataRS[numRS].typeP_Descriptor = generateTypeP_Descriptor(typePDescriptor_Type,typePDescriptor_Value);


    arrayOfdataSSD[numRS] = new ScheduleSlotDescriptionsToBeSend[arrayOfdataRS[numRS].numberOfScheduleSlots];

    for(i=0;i<arrayOfdataRS[numRS].numberOfScheduleSlots;i++){

      arrayOfdataSSD[numRS][i] = new ScheduleSlotDescriptionsToBeSend();

      try {
        arrayOfdataSSD[numRS][i].slotType = (byte)Integer.parseInt(_slotType);
      }catch(NumberFormatException e){
        System.err.println(e);
        arrayOfdataSSD[numRS][i].slotType = -1;
      }

      if((arrayOfdataSSD[numRS][i].slotType != 0)&&(arrayOfdataSSD[numRS][i].slotType != 1)){
        System.err.println("Error: The value of slotType "+i+" read is not valid.");
        System.exit(1);
      }

      double parameter = 0;
      try {
        parameter = Double.parseDouble(_slotParameter);
      }catch(NumberFormatException e){
        System.err.println(e);
        parameter = -1;
      }
      if(parameter < 0){
        System.err.println("Error: The value of slot parameter "+i+" read is not valid.");
        System.exit(1);
      }

      if(arrayOfdataSSD[numRS][i].slotType == 0)
        arrayOfdataSSD[numRS][i].slotParameter = generateSlotParameter(parameter);
      else
        arrayOfdataSSD[numRS][i].slotParameter = new Timestamp(parameter, (short)0);

    }

    try{
      serverSocket = new ServerSocket();    // Used to verify if a socket connection is closed
                                            // using method isSocketConnected.
    }catch(IOException e){
      System.err.println(e);
    }

  }


  // Instance variables

  protected Socket socketToServer;               // Socket to Server. It's not defined as private because
                                                 // we must has access to it value in Fetch Session.

  private Socket[] arrayOfSocketToSender;        // This array will have all the socket opened
                                                 // to the senders, in case of several RS.
  private InetAddress serverAddress;             // The Server's address.
  private int serverTCPportNumber;               // The Server's port for TCP connection.
  private int option;                            // The user's option.
  private byte[] senderAddress = new byte[16];
  private byte[] receiverAddress = new byte[16];
  private int[] arrayOfSenderTCPportNumber;  // Array of TCPportNumber to use in OWAMP-Control with the sender for a
                                             // given RS made.

  private RequestSessionToBeSend[] arrayOfdataRS;    /* Array of datagrams Request Session (it has
                                                      * all request session made by the client).
                                                      */

  private ScheduleSlotDescriptionsToBeSend[][] arrayOfdataSSD; // For each Request Session made by the client
                                                               // we have an array of ScheduleSlotDescriptions.
  private static File clientRSdir = new File("clientRSdir");   // The directory of all RS made using this Control-Client.

  private long seed;                                           // seed used to generate random values.

  private AcceptSessionReceiver dataASr = new AcceptSessionReceiver();
  private StartingTestSessionsToBeSend dataSTS = new StartingTestSessionsToBeSend();
  private Start_AckReceiver dataSA = new Start_AckReceiver();

  private SessionDescriptionRecordReceiver[] arrayOfdataSDRr;  // Used to receive the SDR from Session-Sender
  private SessionDescriptionRecordToBeSend[] arrayOfdataSDRs;  // Array of SDR of all SDR of each Session-Sender
                                                               // used to send the SS to server.
  private  int numSDR = 0;                                     // Number of SDR received from Sender. At the end of
                                                               // all session it should be equal to numRS.

  private BufferedReader systemInReader; // Used to implement the System.in.


  // Instance methods


  /**
   * This  Method prints the menu on the screen.
   */

  public void showMenu(){

    boolean test = true;  // To test if the option entered is ok.
    String line = "";     // Used to receive the string read from keybord.

    do{
      System.out.println("Active measurement using OWAMP: ");
      System.out.println(" ");
      System.out.println("0  -  Connection-Setup");
      System.out.println("1  -  Request-session");
      System.out.println("2  -  Start-session ");
      System.out.println("3  -  Stop-session ");
      System.out.println("4  -  Fetch-session ");
      System.out.println("5  -  Start-session and Fetch ");
      System.out.println("6  -  Finish ");
      System.out.println(" ");
      System.out.print("Choose one of the options: ");

      try {
        line = systemInReader.readLine().trim();
        option = Integer.parseInt(line);
      }catch(NumberFormatException e){
        if(line.length() == 0) // 0 is the default option  -  Connection-Setup
          option = 0;
        else{
          System.err.println(e);
          option = -1;
        }
      }catch (IOException e) {
        System.err.println(e);
        if(numRS != 0){  // Stop all Sessions
          System.out.println("All sessions requested so far will be finished.");
          System.err.println();
          StopSessions.stopAllSessions(socketToServer,arrayOfSocketToSender,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
        }
        System.exit(1);
      }

      if(option == 6){   // Finish
        if(numRS != 0 && isSocketConnected(socketToServer)){  // Stop all Sessions
          System.out.println("All sessions requested so far will be finished.");
          StopSessions.stopAllSessions(socketToServer,arrayOfSocketToSender,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
        }
        System.exit(1);
      }

      if((option >= 0) && (option <= 5))
        test = false;
      else{
        System.out.println("You have to choose one of the listed options.");
        System.out.println("Press Enter to continue...");
        try {
          systemInReader.readLine();
        }
        catch (IOException e) {
          System.err.println(e);
        }
      }

      if((option == 5)&&(numRS > 1)){
        test = true;
        System.out.println("This option is only available if you make just one RS.");
        System.out.println("With more than one RS you should choose option 2.");
        System.out.println("Press Enter to continue...");
        try {
          systemInReader.readLine();
        }
        catch (IOException e) {
          System.err.println(e);
        }
      }

    }while(test == true);

  }


  /**
   * Method to send the packet Request Session to the server. Individual one-way active measurement sessions
   * are established using a simple request/response protocol. An OWAMP client may issue zero or more
   * Request-Session messages to an OWAMP server, which MUST respond to each with an Accept-Session message.
   * An Accept-Session message may refuse a request.
   * @param closeConnection boolean - indicates if the connection to the server should be closed if a
   * negative response come from the server (Server doesn't accept the request).
   * @return boolean - It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean requestSession(boolean closeConnection){    // process Request Session

    // Test if if the startTime is in the past, and separated from the present by more than
    // Timeout time. In this case, the RS should be rejected.

    if ( (arrayOfdataRS[numRS].startTime.timeInMillisecondsSince1970() +
          arrayOfdataRS[numRS].timeout.timeInMillisecondsSince1900()) <
        (long) hiResTimer.milliTime()) {
      System.err.println("Error: The startTime is in the past, and separated from the present by more than Timeout.");
      System.err.println();
      return false;
    }

    // Test socket connection

    if(!isSocketConnected(socketToServer)){
      System.err.println("Error: You should connect to a server first, before trying this operation.");
      System.err.println();

      if(numRS != 0)      // send the SS to Session-Senders. They will free the UDP port allocated to this session.
        StopSessions.stopAllSessions(socketToServer,arrayOfSocketToSender,numRS,false);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)

      return false;
    }

    // Test confSender and confReceiver - At least one of Conf-Sender and Conf-Receiver MUST be 1.

    if((arrayOfdataRS[numRS].confSender==0) && (arrayOfdataRS[numRS].confReceiver==0)){
      System.err.println("Error: At least one of Conf-Sender and Conf-Receiver MUST be 1.");
      System.err.println();
      return false;
    }

    /**
     * SID is the session identifier.  It can be used in later sessions as
     * an argument for Fetch-Session command.  It is meaningful only if
     * Conf-Receiver is 0.  This way, the SID is always generated by the
     * receiving side.
     */


    if(arrayOfdataRS[numRS].confReceiver == 0)    // If confReceiver is set, the value of sid is generated by the server.
      arrayOfdataRS[numRS].sid = generateSid();   // generates the value for the variable sid


    // Connect and send the RS  to Session-Sender to verify if it exists

    arrayOfSocketToSender[numRS] = connectionSetupRequest(mode,determineSeReAddress(arrayOfdataRS[numRS].mbz_iPVN, arrayOfdataRS[numRS].senderAddress),arrayOfSenderTCPportNumber[numRS],TIMEOUT);

    if(arrayOfSocketToSender[numRS] == null){
      System.err.println("Error: The Session-Sender specified doesn't exist.");
      System.err.println();
      return false;
    }

    // After the connection Setup process, a CC should send the RS packet to sender. It will be used to read the
    // sender IP address to be used to open the datagram socket, used to send the test packets to receiver.
    arrayOfdataRS[numRS].sendRequestSession(arrayOfSocketToSender[numRS]);


    // After this process, a sender should send an AcceptSession packet to the ControlClient.
    // This packet has the possible UDP port to be used to send test packets to Receiver, in the port field
    // of the AcceptSession packet. This port will only be used if confSender is zero and confReceiver is 1.

    if(!dataASr.receiveAS(arrayOfSocketToSender[numRS])){
      System.err.println("Error: Error receiving packet AcceptSession from sender, in operation Request session.");
      System.err.println();
      return false;
    }

    // Verify if the Sender has a free port to process the test session. When a range of UDP port is specified,
    // all the ports can be binded. If the sender has not a free UDP port the accept = 1. In this case this
    // session should be rejected.

    if(dataASr.accept == 1){
      System.err.println("The sender rejected the request for the UDP port number to be used in this session.");
      System.err.println();
      System.err.println("  Possible errors:\n\n"+
                         "  i)   The Session-Sender has no free UDP port to process this test session;"+
                         "  ii)  All the UDP ports in the range specified when the Session-Sender was started are binded;"+
                         "  iii) The Session-Sender couldn't bind the UDP socket to the requested address.");
      System.err.println();
      return false;
    }

    if(arrayOfdataRS[numRS].senderPort == 0) // If senderPort = 0 the free UDP port is used.
      arrayOfdataRS[numRS].senderPort = dataASr.port;

    // If Sender exists, Send the RequestSession packet to Server

    // Process the complete RequestSession command

    arrayOfdataRS[numRS].sendCompleteRequestSession(arrayOfdataSSD[numRS],socketToServer);

    /**
     *  To each Request-Session message, an OWAMP server MUST respond with an Accept-Session message.
     *  If the server rejects a Request-Session command, it SHOULD not close the TCP connection.  The
     *  client MAY close it if it gets negative response to Request-Session.
     */

    // Receive the AcceptSession packet

    if(!dataASr.receiveAS(socketToServer)){
      System.err.println("Error: Error receiving packet AcceptSession in operation Request session.");
      System.err.println("All sessions requested so far will be finished.");
      System.err.println();
      // Send the Stop All Session to Server. In case of failure,
      // the Client should stop all sessions and stop this OWAMP-Control session.

      StopSessions.stopAllSessions(socketToServer,arrayOfSocketToSender,numRS+1,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)

      return false;
    }

    String errorStr = "  Possible errors:\n\n"+
                      "  i)   The Session-Receiver doesn't exist.\n"+
                      "  ii)  An error occurred processing connection to receiver.\n"+
                      "  iii) The Session-Receiver has no free UDP port to process this test session because all\n"+
                      "       the UDP ports in the range specified when the Session-Receiver was started are binded.\n"+
                      "  iv)  An error occurred processing Request Session with receiver.\n"+
                      "  v)   The receiver rejected the requested session.\n";

    if(verifyAcceptValue(dataASr.accept,"Server",socketToServer,errorStr)){  // If an error occurred, return true.
      System.err.println("The server rejected the requested session.");
      System.err.println();

      if (closeConnection) {
        System.err.println("All sessions requested so far will be finished.");
        System.err.println();
        // Send the Stop All Session to Server. The Client should stop all sessions and stop this
        // OWAMP-Control session.
        StopSessions.stopAllSessions(socketToServer, arrayOfSocketToSender, numRS+1, true); // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
      }else  // Close the socket to this Session-Senders. It will free the UDP port allocated to this session.
        closeSocket(arrayOfSocketToSender[numRS]);

      return false;
    }

    /**
     * The meaning of Port in the response depends on the values of Conf-Sender and Conf-Receiver in the
     * query that solicited the response. If both were set, Port field is unused.  If only Conf-Sender was
     * set, Port is the port to expect OWAMP-Test packets from.  If only Conf- Receiver was set, Port is the
     * port to send OWAMP-Test packets to.
     * If only Conf-Sender was set, SID field in the response is unused. Otherwise, SID is a unique server-
     * generated session identifier.  It can be used later as handle to fetch the results of a session.
     */

    switch(arrayOfdataRS[numRS].confReceiver){
      case 0:
        if(arrayOfdataRS[numRS].confSender == 1)
          arrayOfdataRS[numRS].senderPort = dataASr.port;
        break;
      case 1:
        if(arrayOfdataRS[numRS].confSender == 0)
          arrayOfdataRS[numRS].receiverPort = dataASr.port;
        System.arraycopy(dataASr.sid,0,arrayOfdataRS[numRS].sid,0,dataASr.sid.length);
    }

    // Now we can send the RS to Sender because the Receiver accepted it.

    if(!requestSessionToSender(numRS)){  // requestSessionToSender(int index)
      System.err.println("Error: Error processing Request Session to Session-Sender.");
      System.err.println("All sessions requested so far will be finished.");
      System.err.println();
      StopSessions.stopAllSessions(socketToServer,arrayOfSocketToSender,numRS+1,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
      return false;
    }

    // An OWAMP client may issue zero or more Request-Session messages to an OWAMP server. So we have to increment
    // the number of Request Session made (numRS) and the size of the arrays used to keep the sessions
    // configurations. The new incremented position will be used by the next session configured.

    numRS++;   // Increment the number of Request Session made

    Socket[] arrayOfSocketToSenderAux = new Socket[numRS+1];
    System.arraycopy(arrayOfSocketToSender,0,arrayOfSocketToSenderAux,0,numRS);
    arrayOfSocketToSender = arrayOfSocketToSenderAux;

    int[] arrayOfSenderTCPportNumberAux = new int[numRS+1];
    System.arraycopy(arrayOfSenderTCPportNumber,0,arrayOfSenderTCPportNumberAux,0,numRS);
    arrayOfSenderTCPportNumber = arrayOfSenderTCPportNumberAux;

    RequestSessionToBeSend[] arrayOfdataRSAux = new RequestSessionToBeSend[numRS+1];
    System.arraycopy(arrayOfdataRS,0,arrayOfdataRSAux,0,numRS);
    arrayOfdataRS = arrayOfdataRSAux;

    ScheduleSlotDescriptionsToBeSend[][] arrayOfdataSSDAux = new ScheduleSlotDescriptionsToBeSend[numRS+1][];
    for(int i=0;i<numRS;i++){
      arrayOfdataSSDAux[i] = new ScheduleSlotDescriptionsToBeSend[arrayOfdataSSD[i].length];
      System.arraycopy(arrayOfdataSSD[i],0,arrayOfdataSSDAux[i],0,arrayOfdataSSD[i].length);
    }
    arrayOfdataSSD = arrayOfdataSSDAux;

    return true;

  }


  /**
   * Having requested one or more test sessions and received affirmative Accept-Session responses, an OWAMP
   * client may start the execution of the requested test sessions by sending a Start-Sessions message to
   * the server.
   * @param saveArrayOfRStoFiles boolean - used to verify if each RS of the arrayOfRS should be saved to a file or not.
   * Each RS of the arrayOfRS should not be saved to file if the ControlClient is used to save the results to a database. Doing
   * this we can prevent a Fetch-Client's user from fetching the results information to be saved in a database when
   * using the ControlClient to save the results to a database (ex. in the web interface).
   * @return boolean - It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean startingTestSessions(boolean saveArrayOfRStoFiles){ // process Starting Test Sessions

    // Test socket connection

    if(!isSocketConnected(socketToServer)){
      System.err.println("Error: You should connect to a server first and request a session, before trying this operation.");

      if(numRS != 0){      // send the SS to Session-Senders. They will free the UDP port allocated to this session.
        StopSessions.stopAllSessions(socketToServer,arrayOfSocketToSender,numRS,false);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
        // Runs the finalization methods of any objects pending finalization. Then, run the garbage colector.
        System.runFinalization();
        System.gc();
      }
      return false;
    }

    if(numRS < 1){  // The client didn't ask for a Request Session yet.
      System.err.println("Error: You should request a session to a server first and wait for an accept message, before trying this operation.");
      System.err.println();
      return false;
    }

    // Send the StartingTestSessions packet to each Sender to start the test sessions. Each request
    // is considered separately in a different socket connection.

    byte accept = 0;
    boolean error = false;

    for(int i = 0; i<numRS;i++){     // For each RS made.

          // Test socket connection

      if(!isSocketConnected(arrayOfSocketToSender[i])){
        System.err.println("Error: Error in Starting Test Session operation.");
        System.err.println("Error: The socket connection to sender n"+(i+1)+" on "+arrayOfSocketToSender[i]+" is closed.");
        System.err.println();
        accept = 1;
        break;
      }

           // Send StartingTestSessions packet to the Sender

      dataSTS.sendSTS(arrayOfSocketToSender[i]);

           // Receive the Start-Ack packet

      if(!dataSA.receiveSA(arrayOfSocketToSender[i])){
        System.err.println("Error: Error in Starting Test Session operation.");
        System.err.println("Error: Error receiving packet Control-Ack in operation Starting Test Session, from Sender n"+i+".");
        System.err.println();
        accept = 1;
        break;
      }

      error = verifyAcceptValue(dataSA.accept,"Session-Sender",arrayOfSocketToSender[i],"The Session-Sender n "+(i+1)+" has rejected the Starting Test Session request.");

      if (error) { // If an error occurred
        System.err.println("Error in Starting Test Session operation.");
        System.err.println();
        accept = 1;
        break;
      }

    }


    // If a Sender didn't accept a request session or an error ocurred, the Client should terminate all sessions.

    if(accept == 1){
      System.err.println("All sessions requested so far will be finished.");
      System.err.println();

      // Stop all sessions started till now.
      // The Client should stop all sessions and stop this OWAMP-Control session.

      StopSessions.stopAllSessions(socketToServer,arrayOfSocketToSender,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)

      // Runs the finalization methods of any objects pending finalization. Then, run the garbage colector.
      System.runFinalization();
      System.gc();

      return false;
    }


    // Send StartingTestSessions packet to the Server
    dataSTS.sendSTS(socketToServer);

    // Receive the Start-Ack packet

    if(!dataSA.receiveSA(socketToServer)){
      System.err.println("Error: Error receiving packet Control-Ack in operation Starting Test Session.");
      System.err.println();
      System.err.println("All sessions requested so far will be finished.");
      System.err.println();

      // Send the Stop All Session to Server. In case of failure,
      // the Client should stop all sessions and stop this OWAMP-Control session.

      StopSessions.stopAllSessions(socketToServer,arrayOfSocketToSender,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)

      // Runs the finalization methods of any objects pending finalization. Then, run the garbage colector.
      System.runFinalization();
      System.gc();

      return false;
    }

    error = false;
    String errorStr = "  Possible errors:\n\n"+
                      "  i)   A Session-Receiver doesn't exists.\n"+
                      "  ii)  A failure occurred in processing Starting Test Sessions between the Server and a Receiver.\n"+
                      "  iii) A Session-Receiver has rejected the Starting Test Session request.\n";
    error = verifyAcceptValue(dataSA.accept,"Server",socketToServer,errorStr);

    if (error) { // If an error occurred, return false.
      // The Start-Sessions request was rejected.
      // Send the Stop All Session to Server. In case of failure,
      // the Client should stop all sessions and stop this OWAMP-Control session.
      StopSessions.stopAllSessions(socketToServer,arrayOfSocketToSender,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
      System.err.println("The Start-Sessions request was rejected.");
      System.err.println("All sessions requested so far will be finished.");
      System.err.println();

      // Runs the finalization methods of any objects pending finalization. Then, run the garbage colector.
      System.runFinalization();
      System.gc();

      return false;
    }


    // Now the Client should wait till the end of all sessions or a request of Stop Session from user.

    long startTime = (long)hiResTimer.milliTime();

    // Calculate the total time to wait
    double totalTimeToWait = 0, timeOutOfLongestSession = 0;
    double ti;

    for(int i = 0; i<numRS;i++){
      ti = totalTimeToWait(arrayOfdataRS[i], arrayOfdataSSD[i], hiResTimer);   // totalTimeToWait(RequestSession arrayOfdataRS, ScheduleSlotDescriptions[] arrayOfdataSSD, HiResTimer hiResTimer)
      if(totalTimeToWait < ti){
        totalTimeToWait = ti;
        timeOutOfLongestSession = arrayOfdataRS[i].timeout.timeInMillisecondsSince1900();
      }
    }

    if(saveArrayOfRStoFiles)
      if(!arrayOfRStoFiles()){  // Record all the RS that have been accepted and started to a file in clientRSdir. This information
                                // will be used in Fetch-Sessions.
        StopSessions.stopAllSessions(socketToServer,arrayOfSocketToSender,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
        System.err.println("Error: Error saving arrayOfRS to file.");
        System.err.println("All sessions requested so far will be finished.");
        System.err.println();

        // Runs the finalization methods of any objects pending finalization. Then, run the garbage colector.
        System.runFinalization();
        System.gc();

        return false;
      }

    int startOption = option;


    // During total time to wait + TIMEOUT (to guarantee that all SS from all sender are correctly received)
    // wait for the end of all test sessions and test if the user didn't request
    // a Stop Session or if a Stop Session haven't been received from Server.

    if(!waitTotalTimeToWait(startTime, totalTimeToWait+TIMEOUT, timeOutOfLongestSession)){ // waitTotalTimeToWait(long startTime, long totalTimeToWait, double timeOutOfLongestSession)
      // Runs the finalization methods of any objects pending finalization. Then, run the garbage colector.
      System.runFinalization();
      System.gc();
      return false;
    }

    System.out.println("End of all Test Sessions started.");
    System.out.println();

    if(startOption==-1){  // Used in Web interface, to start session and end without fecthing.
      // Runs the finalization methods of any objects pending finalization. Then, run the garbage colector.
      System.runFinalization();
      System.gc();
      return true;
    }

    // At the end of all sessions the client should execute none or more Fetch Sessions.

    if(startOption==2){ // Option wasn't Start-session and Fetch.

      for(int i=0;i<numRS;i++){  // Test if a session was started. If yes, process FS. (A session can be stopped before
                                 // its start time)
        if(arrayOfdataRS[i].startTime.timeInMillisecondsSince1970() < hiResTimer.milliTime()){
          OWAMP_FetchClient fetch = new OWAMP_FetchClient(serverAddress,serverTCPportNumber, socketToServer, arrayOfdataRS, arrayOfdataSSD, (short)numRS, systemInReader, hiResTimer);  // OWAMP_FetchClient(InetAddress serverAddress,int serverTCPportNumber, Socket socketToServer, RequestSessionToBeSend[] arrayOfdataRS, ScheduleSlotDescriptionsToBeSend[][] arrayOfdataSSD, short numRS, BufferedReader systemInReader, HiResTimer hiResTimer)
          while(fetch.processFetchSession(false));   /* processFetchSession(boolean readArrayOfRSfromFiles)
                                                      * When this method is used to fetch the test results in the OWAMP_ControlClient all the RS are in the
                                                      * arrayOfdataRS variable, so there is no need to read them from the clientRSdir. Also this directory may have
                                                      * other RS files of others instances of the OWAMP_ControlClient class. When this method is used to fetch the
                                                      * test results in the OWAMP_ControlClientThese sessions shouldn't appear in the Fetch-Session's option list.
                                                      */
          break;
        }
      }

      closeSocket(socketToServer);   // Close socket connection to Server.

      // Runs the finalization methods of any objects pending finalization. Then, run the garbage colector.
      System.runFinalization();
      System.gc();

      return true;
    }

    // Test if the session was started. If yes, process FS. (A session can be stopped before its start time)
    if(arrayOfdataRS[0].startTime.timeInMillisecondsSince1970() < hiResTimer.milliTime()){

      System.out.println(" Option was Start-session and Fetch: ");
      System.out.println();

      // The Client should execute the Fetch Session.

      OWAMP_FetchClient fetch = new OWAMP_FetchClient(serverAddress,serverTCPportNumber, arrayOfdataRS[0], socketToServer, systemInReader, hiResTimer);
      // Fetch all packets
      if(!fetch.fetchSession(0, -1, 0)){  // FetchSession(int beginSeq, int endSeq, int indexOfSession)
        // In case of error, if all results are requested and the test session is finished its RS file should be deleted.
        System.err.println("Error: Error processing Fetch Session.");
        File sessionFileName = new File(clientRSdir.getPath()+clientRSdir.separator+owampHexEncode(arrayOfdataRS[0].sid)+".rs");  // If session file exists, delete it.
        if(sessionFileName.exists())
          sessionFileName.delete();
      }

    }

    closeSocket(socketToServer); // close socket connection to Server

    // Runs the finalization methods of any objects pending finalization. Then, run the garbage colector.
    System.runFinalization();
    System.gc();

    return true;

  }


  /**
   * Method to send the packet Request Session to the Session-Sender.
   * To each Request-Session an OWAMP Session-Sender MUST respond with an Accept-Session message.
   * @param index int - The variable index represents the current Request Session in array arrayOfdataRS.
   * @return boolean - It returns true if everything goes ok, otherwise it returns false.
   */

   public boolean requestSessionToSender(int index){

     // Test socket connection

     if(!isSocketConnected(arrayOfSocketToSender[index])){
       System.err.println("Error: You should connect to a Session-Sender first, before trying this operation.");
       System.err.println();
       return false;
     }

     // Process the complete RequestSession command to sender

     arrayOfdataRS[index].sendCompleteRequestSession(arrayOfdataSSD[index],arrayOfSocketToSender[index]);

     // To each Request-Session message, an Session Sender MUST respond with an Accept-Session message.

     // Receive the AcceptSession packet

     if(!dataASr.receiveAS(arrayOfSocketToSender[index])){
       System.err.println("Error: Error receiving packet AcceptSession in operation Request session.");
       System.err.println();
       closeSocket(arrayOfSocketToSender[index]);
       return false;
     }

     boolean error = false;
     String errorStr = "";
     error = verifyAcceptValue(dataASr.accept,"Session-Sender",arrayOfSocketToSender[index],errorStr);

     if (error) { // If an error occurred, return false.
       System.err.println("The Session-Sender rejected the requested session.");
       System.err.println();
       closeSocket(arrayOfSocketToSender[index]);
       return false;
     }

     return true;

  }


  /**
   * After Starting Test Sessions command, during totalTimeToWait, this method is used wait for the end of all
   * test sessions of this OWAMP control session and to test if the user didn't request a Stop Session or if a
   * Stop Session haven't been received from Server or Session-Sender.
   * Note: At the end of a test session, the Session-Sender involved sends a SS to client with the Session
   * Description Record information.
   * @param startTime long - this variable represents the time that the current session started operating.
   * @param totalTimeToWait long - the time the Client should wait to finish the current session.
   * @param timeOutOfLongestSession double - is the timeOut of the longest Session. It will be used to wait for
   * this timeOut if all Test Session finish before the totalTimeToWait. In this way we guarantee that the last
   * test packet can reach the Receiver before the SS message is processed.
   * @return boolean - It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean waitTotalTimeToWait(long startTime, double totalTimeToWait, double timeOutOfLongestSession){

    // Wait for command Stop Session or end of total time to wait for the end of all sessions.

    try {

      boolean endTotalTimeToWait = false;   // Variable used to test if totalTimeToWait has occurred
      double timeOutToWait = 0;             // timeout to wait a StopSession packet to be received, in ms.
      arrayOfdataSDRr = new SessionDescriptionRecordReceiver[1];      // Used to receive the SDR from Session-Sender
      arrayOfdataSDRs = new SessionDescriptionRecordToBeSend[numRS];  // Array of SDR of all SDR of each Session-Sender
                                                                      // used to send the SS to server.
      numSDR = 0;       // Number of SessionDescriptionRecord.
      int numOfSS = 0;  // Number of stopped sessions.

      do{

        System.out.println("The only commands available now are: ");
        System.out.println();
        System.out.println("3  -  Stop-session ");
        System.out.println("6  -  Finish ");
        System.out.println(" ");
        System.out.print("Choose one of the options: ");

        option = 0;
        timeOutToWait = 3*TIMEOUTTOWAIT;

        do{

          if(systemInReader.ready()){    // If the user choose an option

            try {
              option = Integer.parseInt(systemInReader.readLine().trim());
            }catch(NumberFormatException e){
              System.err.println(e);
              option = 0;
            }

            if ( (option != 3) && (option != 6)){       // Option should be StopSession(3) or Finish(6)
              System.err.println("Error: The only command available is Stop-Sessions or Finish.");
              System.out.println(" ");
              System.out.println("3  -  Stop-session ");
              System.out.println("6  -  Finish ");
              System.out.println(" ");
              System.out.print("Choose one of the options: ");
            }
          }

          if((option == 3) || (option == 6))
            timeOutToWait = TIMEOUTTOWAIT;

          // Test if a Stop Session was sent by the Server or Sender and, if yes, process it.

          if(!testStopSession(timeOutToWait))
            return false;

          if ((numSDR == numRS) || ((startTime + totalTimeToWait) < ((long)hiResTimer.milliTime()))){ // All SS from sender have been received (All Test Session are finished) or Time to wait has finished
            if(numSDR == numRS){ // All Test Session finish before the totalTimeToWait, so we should wait for
                                 // timeOutOfLongestSession. In this way we guarantee that the last test packet
              try{               // can reach the Receiver before the SS message is processed.
                sleep((long)timeOutOfLongestSession);
              }catch(Exception e){}
            }else{
              System.err.println("Error: Total time to wait occurred and not all the Senders sent the Session Description Record.");
              System.err.println("Error: All the results will be invalidated.\n");
              StopSessions.stopAllSessions(socketToServer, arrayOfSocketToSender, numRS, true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
              return false;
            }
            endTotalTimeToWait = true;
            break;
          }

        }while ((option != 3) && (option != 6));


        if(!endTotalTimeToWait){  // If a option has been entered.

          int i = 0, j = 0;
          int[] arrayj = new int[numRS]; // Array of the Index of all session that are still operating.

          if (option == 3) {  // Stop-Session.
                              // Choose the sessions to stop.
            int c;
            String senderStrAddress = "";  // Sender's address string representation.
            int maxLength = 0;             // Length of the biggest Sender's address string representation.

            do {
              System.out.println("Choose which session do you want to stop: ");
              System.out.println();
              maxLength = 0;
              for (i = 0; i < numRS; i++) {
                if (!arrayOfSocketToSender[i].isClosed()) {
                  senderStrAddress = determineSeReAddress(arrayOfdataRS[i].mbz_iPVN, arrayOfdataRS[i].senderAddress).getHostName();
                  maxLength = (senderStrAddress.length() > maxLength)?senderStrAddress.length():maxLength;
                }
              }
              j=0;
              System.out.print("index\tSender Address");
              for(c=0;c<(maxLength - "Sender Address".length());c++)
                System.out.print(" ");
              System.out.print("\t-->\tReceiver Address\n\n");
              for (i = 0; i < numRS; i++) {
                if (!arrayOfSocketToSender[i].isClosed()) {
                  // Determine sender address
                  senderStrAddress = determineSeReAddress(arrayOfdataRS[i].mbz_iPVN, arrayOfdataRS[i].senderAddress).getHostName();
                  System.out.print( (j + 1) + "\t" +senderStrAddress);
                  for(c=0;c<(maxLength - senderStrAddress.length());c++)
                    System.out.print(" ");
                  System.out.print("\t-->\t" +determineSeReAddress(arrayOfdataRS[i].mbz_iPVN, arrayOfdataRS[i].receiverAddress).getHostName()+"\n");

                  arrayj[j] = i;
                  j++;
                }
              }
              System.out.println( (j + 1) + "\tClose all sessions.  ");

              // Waits for the user's option while testing if a SS was received from Server or Sender.

              option = 0;
              do{
                timeOutToWait = 3 * TIMEOUTTOWAIT;
                if (systemInReader.ready()) { // If the user choose an option
                  try {
                    option = Integer.parseInt(systemInReader.readLine().trim());
                  }catch(NumberFormatException e){
                    System.err.println(e);
                    option = -1;
                  }
                  timeOutToWait = TIMEOUTTOWAIT;
                }

                // Test if a Stop Session was sent by the Server or Sender and, if yes, process it.

                if (!testStopSession(timeOutToWait))
                  return false;

                if ((numSDR == numRS) || ((startTime + totalTimeToWait) < ((long)hiResTimer.milliTime()))){ // All SS from sender have been received (All Test Session are finished) or Time to wait has finished
                  if(numSDR == numRS){ // All Test Session finish before the totalTimeToWait, so we should wait for
                                       // timeOutOfLongestSession. In this way we guarantee that the last test packet
                    try{               // can reach the Receiver before the SS message is processed.
                      sleep((long)timeOutOfLongestSession);
                    }catch(Exception e){}
                  }else{
                    System.err.println("Error: Total time to wait occurred and not all the Senders sent the Session Description Record.");
                    System.err.println("Error: All the results will be invalidated.\n");
                    StopSessions.stopAllSessions(socketToServer, arrayOfSocketToSender, numRS, true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
                    return false;
                  }
                  endTotalTimeToWait = true;
                  break;
                }

              }while(option == 0);

              if(endTotalTimeToWait)
                break;

              if((option > (j + 1))||(option < 0)){
                System.err.println();
                System.err.println("Error: The option entered is not valid.");
                System.err.println();
              }

            }while ((option > (j + 1))||(option < 0));

          } else {  // Option = 6 - Finish
            j = 0;
                             // Search for all session that are still operating.
            for (i = 0; i < numRS; i++) {
              if (!arrayOfSocketToSender[i].isClosed()) {
                arrayj[j] = i;
                j++;
              }
            }
            option = j + 1;      // options represents close all sessions.
          }

          if (!endTotalTimeToWait) {

            int begin = 0;
            int end = option;

            if (option != (j + 1)) // If options is different from close all sessions.
              begin = end - 1;
            else
              end = option - 1;

            // Send a Stop Session with accept=0 to each Sender and Receive the SDR of each Sender

            int index = 0; // Number of Sessions running that should be stopped

            short stopVal;

            // For each session we should process the complete SS before requesting the SS to another one. This should
            // be done to prevent errors because the receivers will be waiting for packets that are not sent and they
            // will be marked as lost, but aren't.

            for (i = begin; i < end; i++, index++, numOfSS++) {

              if (!stopSessionToSender(arrayj[i])) // stopSessionToSender(int index)
                return false;

              // Send the stopSessions packet to Server

              dataSSs = new StopSessionsToBeSend( (short) 0, 1);
              dataSSs.sendCompleteSS(arrayOfdataSDRs[numSDR - 1], socketToServer);
              arrayOfdataSDRs[numSDR - 1] = null;  // It should not be send again when all the other session are finished.

              // Receive the StopSession response from Server.

              stopVal = dataSSr.receiveSS(socketToServer, TIMEOUT);

              if ( (stopVal == 1) && (dataSSr.accept != 0)) {
                verifyAcceptValue(dataSSr.accept,"Server",socketToServer,"Error: All results will be invalidated.");
                StopSessions.stopAllSessions(socketToServer, arrayOfSocketToSender, numRS, false); // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
                return false;
              }

              if (stopVal != 1) {  // If an error ocurr in receiving the SS packet all results should be invalidated.
                if(stopVal==0)     // Timeout to wait occurred.
                  System.err.println("Error: The Server didn't response the StopSession.\n");
                System.err.println("Error: All results will be invalidated.");
                System.err.println();
                if(stopVal == -2)  // The connection to socket was reset.
                  StopSessions.stopAllSessions(socketToServer, arrayOfSocketToSender, numRS, false); // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
                else               // An error occurred reading from socket
                  StopSessions.stopAllSessions(socketToServer, arrayOfSocketToSender, numRS, true); // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
                return false;
              }

            }

            // Test if not all session has been finished

            if(numSDR == numRS) { // all session has been finished
              if (index != 0){ // If index != 0 there are sessions stopped.
                sleepTime( (int) arrayOfdataRS[arrayj[begin]].timeout.
                          timeInMillisecondsSince1900()); // Wait the receiver to receive the SS packet.
                if(numOfSS == numRS){  // All session were already stopped.
                  return true;
                }
              }
              endTotalTimeToWait = true;
              break;
            }

            if (index != 0) { // If index != 0 there are sessions stopped.
              // When a session is stopped, we must test total time to wait again because the current
              // TotalTimeToWait maybe has respect to a stopped RS.

              double ti = 0;
              totalTimeToWait = 0;
              for (i = 0; i < numRS; i++) {
                if (!arrayOfSocketToSender[i].isClosed()) {
                  ti = totalTimeToWait(arrayOfdataRS[i], arrayOfdataSSD[i], hiResTimer); // totalTimeToWait(RequestSession dataRS, ScheduleSlotDescriptions[] arrayOfdataSSD, HiResTimer hiResTimer)
                  if (totalTimeToWait < ti)
                    totalTimeToWait = ti;
                }
              }
              totalTimeToWait = totalTimeToWait + TIMEOUT;  // total time to wait + TIMEOUT (to guarantee that all SS from all sender are correctly received)
            }
          }
        }
      }while(!endTotalTimeToWait);

      // If all Test Session finished
      // Send a StopSession to Server

      System.out.println("All Test Session finished");

      dataSSs = new StopSessionsToBeSend( (short)0, (numRS-numOfSS));  // NumberOfSession = numRS - numOfSS(num of stopped sessions)
      dataSSs.sendCompleteSS(arrayOfdataSDRs,socketToServer);

                    // Receive the StopSession response from Server.

      short stopVal = dataSSr.receiveSS(socketToServer, (numRS-numOfSS)*TIMEOUT);

      if ( (stopVal == 1) && (dataSSr.accept != 0)) {
        verifyAcceptValue(dataSSr.accept,"Server",socketToServer,"Error: All results will be invalidated.");
        closeSocket(socketToServer);
        return false;
      }

      if (stopVal != 1) {  // If an error ocurr in receiving the SS packet all results should be invalidated.
        if(stopVal==0)     // Timeout to wait occurred.
          System.err.println("Error: The Server didn't response the StopSession.\n");
        closeSocket(socketToServer);
        return false;
      }

      return true;

    }
    catch (IOException e) {
      System.err.println(e);
      System.err.println();
      System.err.println("All sessions requested so far will be finished.");
      System.err.println();
      StopSessions.stopAllSessions(socketToServer, arrayOfSocketToSender, numRS, true); // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
      return false;
    }

  }


  /**
   * This method is used to test if a SS was received from Server or from any Sender. If a SS is received
   * it is processed acording to the element that sent it. If a SS is received from a Sender and accept is 0
   * we should record the SDR in the arrayOfdataSDRs. This information will be send to the Server at the end
   * of all test sessions.
   * Note: At the end of a test session, the Session-Sender involved sends a SS to client with the Session
   * Description Record information.
   * @param timeOutToWait double - timeOut to wait for the SS.
   * @return boolean - It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean testStopSession(double timeOutToWait){

    // Verify the socketToServer. If it is closed before totalTimeToWait
    // It means that a failure has occurred, possibly a loss of connection
    // to Server. The session must be finished.

    if (!isSocketConnected(socketToServer)){
      System.err.println("Error: The socket connection to Server is no longer available. All sessions will be finished.");
      System.err.println();
      StopSessions.stopAllSessions(socketToServer, arrayOfSocketToSender, numRS, false);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
      return false;
    }

    short stopVal = 0;  // Auxiliary variable used to verify if a session was stopped.

    // Test if a Stop Session was sent by the Server and, if yes, process it.

    stopVal = dataSSr.receiveSS(socketToServer, timeOutToWait);

    if (stopVal == -2){ // The socket connections was reset.
      StopSessions.stopAllSessions(socketToServer, arrayOfSocketToSender, numRS, false);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
      System.err.println("All sessions will be finished.");
      System.err.println();
      return false;
    }

    if ( (stopVal == -1) || ( (stopVal == 1) && (dataSSr.accept != 0))){ // If stopSession's accept val = 1 or an error ocurred stop operation.
      if (stopVal != -1) {
        System.out.println("StopSession from Server on " + socketToServer);
        verifyAcceptValue(dataSSr.accept, "Server",socketToServer, "");
        StopSessions.stopAllSessions(socketToServer, arrayOfSocketToSender, numRS, false);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
      }else
        StopSessions.stopAllSessions(socketToServer, arrayOfSocketToSender, numRS, true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
      System.err.println("All sessions will be finished.");
      System.err.println();
      return false;
    }

    if (stopVal == 1){ // If there is a StopSession and accept = 0, the Server sent a bad SS. Only
                       // CC can send this kind of SS.
      System.out.println("StopSession from Server on " + socketToServer);
      System.err.println("Error: Error in StopSession operation from Server.");
      System.err.println();
      System.err.println("  Possible errors:");
      System.err.println(
          "  i)   The Server is trying to stop sessions that it didn't initiate.");
      System.err.println();
      System.err.println("All sessions will be finished.");
      System.err.println();
      StopSessions.stopAllSessions(socketToServer, arrayOfSocketToSender, numRS, true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
      return false;
    }

    // Test if stop session was sent from each Sender. A Sender can send a SS to Control-Client
    // if it can not bind the datagram socket to the port especified in RS datagram, in case this
    // port is already taken or when the test session is finished. In this last case we should record
    // the Session Description Record packet received in the array of Session Description Records (arrayOfdataSDRs).

    for (int i = 0; i < numRS; i++){
      if (!arrayOfSocketToSender[i].isClosed()){
        stopVal = dataSSr.receiveSS(arrayOfSocketToSender[i], TIMEOUTTOWAIT);

        if (stopVal == -2){ // The socket connections was reset.
          closeSocket(arrayOfSocketToSender[i]); // Close socket connection to Sender
          StopSessions.stopAllSessions(socketToServer, arrayOfSocketToSender, numRS, true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
          System.err.println("All sessions will be finished.");
          System.err.println();
          return false;
        }

        if ( (stopVal == -1) || ( (stopVal == 1) && (dataSSr.accept != 0))) { // If stopSession's accept val = 1 or an error ocurred stop operation.
          if (stopVal != -1) {
            System.out.println("StopSession from Session-sender on " +
                               arrayOfSocketToSender[i]);
            System.out.println();
            String error = "  Possible errors:\n\n" +
                "  i)   The Sender can not bind the datagram socket to the port especified in RS datagram because this port is already taken.\n";
            verifyAcceptValue(dataSSr.accept, "Session-sender",arrayOfSocketToSender[i], error);
            closeSocket(arrayOfSocketToSender[i]); // Close socket connection to Sender
          }
          System.err.println("Error: Error in StopSession operation from Session-Sender.");
          System.err.println("All sessions will be finished.");
          System.err.println();
          StopSessions.stopAllSessions(socketToServer, arrayOfSocketToSender,
                                       numRS, true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
          return false;
        }

        if (stopVal == 1){ // A SS was received and accept is zero. In this case we should receive the complete SS and
          // record the Session Description Record packet received in the arrayOfSessionDescriptionRecords.
          arrayOfdataSDRr = dataSSr.receiveCompleteSS(arrayOfSocketToSender[i]);

          if (arrayOfdataSDRr == null) {
            System.err.println("Error: Error in StopSession operation from Sender "+arrayOfSocketToSender[i]);
            System.err.println("All sessions will be finished.");
            System.err.println();
            StopSessions.stopAllSessions(socketToServer, arrayOfSocketToSender, numRS, true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
            return false;
          }

          arrayOfdataSDRs[numSDR] = new SessionDescriptionRecordToBeSend(
              arrayOfdataSDRr[0].sid, arrayOfdataSDRr[0].nextSeqno,
              arrayOfdataSDRr[0].numberOfSkipRanges,
              arrayOfdataSDRr[0].skipRangeDescriptions);

          numSDR++;

          // Send the SS response message to Session-Sender
          dataSSs = new StopSessionsToBeSend( (short) 0, 0);
          dataSSs.sendSS(arrayOfSocketToSender[i]);

          closeSocket(arrayOfSocketToSender[i]);
        }
      }
    }
    return true;
  }


  /**
   * This method is used to test if not all session has been finished.
   * @return boolean - true if all sessions are finished, otherwise false
   */

  public boolean isAllSessionsFinished(){

    // Test if not all session has been finished

    int j = 0;
    for (int i = 0; i < numRS; i++)
      if (!arrayOfSocketToSender[i].isClosed()){
        j++;
        break;
      }

    if (j == 0){ // all session has been finished (Stop operation)
      closeSocket(socketToServer);
      return true;
    }

    return false;

  }


  /**
   * This method is used to request a Stop Session to a sender giving the index of the RS that correspond to
   * this sender. The client sends a StopSession packet to the sender with accept field with zero value. The
   * sender should respond with another StopSession packet with the SDR information of this Test-Session.
   * @param index int - index represents the index of the RS in array arrayOfdataRS.
   * @return boolean - It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean stopSessionToSender(int index){

    // Test socket connection

    if(!isSocketConnected(arrayOfSocketToSender[index])){
      System.err.println("Error: The socket connection to the sender "+arrayOfSocketToSender[index]+" is no longer available. This session is finished.");
      System.err.println("All sessions will be finished.");
      System.err.println();
      StopSessions.stopAllSessions(socketToServer, arrayOfSocketToSender,
                                   numRS, true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
      return false;
    }

    // Send the stopSessions packet

    dataSSs = new StopSessionsToBeSend((short)0, 0);
    dataSSs.sendSS(arrayOfSocketToSender[index]);

    // Receive the stopSessions packet from Sender with the SDR information

    short stopVal = dataSSr.receiveSS(arrayOfSocketToSender[index], TIMEOUT);

    if (stopVal == -2){ // The socket connections was reset.
      closeSocket(arrayOfSocketToSender[index]); // Close socket connection to Sender
      StopSessions.stopAllSessions(socketToServer, arrayOfSocketToSender, numRS, true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
      System.err.println("All sessions will be finished.");
      System.err.println();
      return false;
    }

    if ( (stopVal == -1) || ( (stopVal == 1) && (dataSSr.accept != 0))) { // If stopSession's accept val = 1 or an error ocurred stop operation.
      if (stopVal != -1) {
        System.out.println("StopSession from Session-sender on " +
                           arrayOfSocketToSender[index]);
        String error = "  Possible errors:\n\n" +
            "  i)   The Sender can not bind the datagram socket to the port especified in RS datagram because this port is already taken.\n";
        verifyAcceptValue(dataSSr.accept, "Session-sender",arrayOfSocketToSender[index], error);
        closeSocket(arrayOfSocketToSender[index]); // Close socket connection to Sender
      }
      System.err.println("All sessions will be finished.");
      System.err.println();
      StopSessions.stopAllSessions(socketToServer, arrayOfSocketToSender,
                                   numRS, true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
      return false;
    }

    if (stopVal == 1) { // A SS was received and accept is zero. In this case we should receive the complete SS and
      // record the Session Description Record packet received in the arrayOfSessionDescriptionRecords.
      arrayOfdataSDRr = dataSSr.receiveCompleteSS(arrayOfSocketToSender[index]);
      arrayOfdataSDRs[numSDR] = new SessionDescriptionRecordToBeSend(
          arrayOfdataSDRr[0].sid, arrayOfdataSDRr[0].nextSeqno,
          arrayOfdataSDRr[0].numberOfSkipRanges,
          arrayOfdataSDRr[0].skipRangeDescriptions);
      numSDR++;

      closeSocket(arrayOfSocketToSender[index]);
      return true;
    }else{  // A SS with the SDR information wasn't received from Sender
      System.err.println(
            "Error: Error in StopSession operation from Session-Sender. The Sender didn't answer to the StopSession message.");
      System.err.println("All sessions will be finished.");
      System.err.println();
      StopSessions.stopAllSessions(socketToServer, arrayOfSocketToSender,
                                   numRS, true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
      return false;
    }

  }


  /**
   * This method generates the TypeP_Descriptor value to be used in packet Request Session
   * @param typePDescriptor byte - This variable specifies the Type-P Descriptor to be used. Type-P Descriptor
   * covers only a subset of (very large) Type-P space.
   * If the first two bits of Type-P Descriptor are 00 (typePDescriptor = 0), then subsequent 6 bits specify the requested
   * Differentiated Services Codepoint (DSCP) value of sent OWAMP-Test packets as defined in RFC 2474.
   * If the first two bits of Type-P descriptor are 01 (typePDescriptor = 1), then subsequent 16 bits specify the requested Per
   * Hop Behavior Identification Code (PHB ID) as defined in RFC 2836.
   * Therefore, the value of all zeros specifies the default best-effort service.
   * If Conf-Sender is set, Type-P Descriptor is to be used to configure the sender to send packets
   * according to its value.  If Conf-Sender is not set, Type-P Descriptor is a declaration of how the
   * sender will be configured. If Conf-Sender is set and the server doesn't recognize Type-P Descriptor,
   * cannot or does not wish to set the corresponding attributes on OWAMP-Test packets, it SHOULD reject
   * the session request.  If Conf-Sender is not set, the server SHOULD accept the session regardless of
   * the value of Type-P Descriptor.
   * @param typePvalue int
   * @return byte[] - byte array representing the TypeP_Descriptor.
   */

  public byte[] generateTypeP_Descriptor(byte typePDescriptor,int typePvalue){

    byte[] typeP_Descriptor = new byte[4];

    switch(typePDescriptor){
      case 0:                       // If the first two bits of Type-P Descriptor are 00, then
        typeP_Descriptor[0] = 0;    // subsequent 6 bits specify the requested Differentiated
        typeP_Descriptor[1] = 0;    // Services Codepoint (DSCP) value of sent OWAMP-Test packets
                                    // as defined in RFC 2474.
        typeP_Descriptor[2] = (byte)(typePvalue << 2);
        break;
      case 1:                       // If the first two bits of Type-P descriptor are 01, then
        typeP_Descriptor[0] = 0;    // subsequent 16 bits specify the requested Per Hop Behavior
        typeP_Descriptor[1] = 1;    // Identification Code (PHB ID) as defined in RFC 2836.
        typeP_Descriptor[2] = (byte)((typePvalue >> 8) & 0xFF);
        typeP_Descriptor[3] = (byte)(typePvalue  & 0xFF);
        break;
      default:       // Therefore, the value of all zeros specifies the default best-effort service.
        for(int i = 0;i<typeP_Descriptor.length;i++)
          typeP_Descriptor[i] = 0;
    }

    return typeP_Descriptor;
  }


  /**
   * Method used to record the arrayOfdataRS and the arrayOfdataSSD, for each RS, in a file to be used in Fetch Session. For each RS we have a
   * different file, which name is the session SID as a string.
   * In this case we can not have the same file to all sessions because, if more than one Control-Client
   * are being used at the same time, they would use the same file. Which can not be possible. The
   * only way to resolve this problem is to have different files, with different names.
   * @return boolean -  This method returns true if everything goes ok, otherwise it returns false.
   */

  public boolean arrayOfRStoFiles(){

    try{
      // Test clientRSdir directory, to prevent errors if a user deletes it. If it doesn't exist it should be created.
      if (!clientRSdir.exists())
        clientRSdir.mkdir();

      for(int i=0;i<numRS;i++){
        if (!arrayOfSocketToSender[i].isClosed()) {  // It means that this session was started
                                                     // record the arrayOfdataRS
          File outputFileName = new File(clientRSdir.getPath()+clientRSdir.separator+owampHexEncode(arrayOfdataRS[i].sid)+".rs");  // If outputFile already exists, delete it
          if(outputFileName.exists())
            outputFileName.delete();

          RandomAccessFile outputFile = new RandomAccessFile(outputFileName,"rw");  // Create a new outputFile
          if(!saveRStoFile(arrayOfdataRS[i],arrayOfdataSSD[i],outputFile)){
            outputFile.close();
            return false;
          }

          outputFile.close();
        }
      }

    }
    catch(IOException ex){
      System.err.println(ex);
      return false;
    }

    return true;
  }


  /**
   * A generator of random numbers with exponential distribution.
   */

  public class RandomExponentialDistribution{

    private Random rand;
    private double mean;

    /**
     * Create a new generator of random numbers
     * with exponential distribution of mean 1.
     */

    public RandomExponentialDistribution(){
      this(1.0, new Random());
    }

    /**
     * Create a new generator of random numbers
     * with exponential distribution of specified mean.
     * @param mean the mean of the exponential distribution
     */

    public RandomExponentialDistribution(double mean){
      this(mean, new Random((long)hiResTimer.nanoTime() + (seed++)));
    }

    /**
     * Create a new generator of random numbers with exponential
     * distribution of specified mean that will use randgen
     * as its source of random numbers.
     * @param mean the mean of the exponential distribution
     * @param randgen a Random object to be used by the generator.
     */

    public RandomExponentialDistribution(double mean, Random randgen){
      this.mean = mean;
      rand = randgen;
    }

    /**
     * Return a random number with exponential distribution.
     */

    public double nextDouble(){
      double val;

      do
        val = rand.nextDouble();
      while (val == 0.0);

      return mean * (-Math.log(val));
    }

  }


  /**
   * This method is used to generate a Timestamp. This Timestamp is used to represents a SlotParameter.
   * It's value is exponentially distributed with mean 'mean'. This SlotParameter will
   * be used in RequestSession, in ScheduleSlotDescriptions packets.
   * @param mean double - the mean value to be used in seconds.
   * @return Timestamp -  the Timestamp to be returned.
   */

  public Timestamp generateSlotParameter(double mean){

    RandomExponentialDistribution exponential = new RandomExponentialDistribution(mean);
    Timestamp slotParameter = new Timestamp(exponential.nextDouble(),(short)0); // Time in seconds since 1Jan1900

    return slotParameter;
  }


  /**
   * This method is used to read the configuration file to the position index of arrayOfdataRS, used to process
   * a new RS. In this file we have all the fields necessary to configure a RS command. In this file all
   * lines beginning by character '#' are ignored. All configurations parameters should be read always
   * following the same format present in the RSconfigFile, and it should not be changed. Their values
   * are placed after the signal '='. Any spaces after this signal is ignored, except in the case of the
   * Sender's and receiver's addresses (no space should exist between signal '=' and the value).
   * @param index int - index of arrayOfdataRS where the configuration of the new RS is read to.
   * @return boolean - This method returns true if everything goes ok, otherwise it returns false.
   */

  public boolean readConfigFile(int index){

    try{
      RandomAccessFile configFile = new RandomAccessFile("RSconfigFile.txt","r");
      arrayOfdataRS[index] = new RequestSessionToBeSend();

      while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
        configFile.readLine();
      try {
        arrayOfdataRS[index].confSender = (byte) Integer.parseInt(configFile.readLine().trim());
      }catch(NumberFormatException e){
        System.err.println(e);
        arrayOfdataRS[index].confSender = -1;
      }
      if((arrayOfdataRS[index].confSender != 0)&&(arrayOfdataRS[index].confSender != 1)){
        System.err.println("Error: The value of confSender read is not valid.");
        System.err.println();
        return false;
      }


      while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
        configFile.readLine();
      try {
        arrayOfdataRS[index].confReceiver = (byte) Integer.parseInt(configFile.readLine().trim());
      }catch(NumberFormatException e){
        System.err.println(e);
        arrayOfdataRS[index].confReceiver = -1;
      }
      if((arrayOfdataRS[index].confReceiver != 0)&&(arrayOfdataRS[index].confReceiver != 1)){
        System.err.println("Error: The value of confReceiver read is not valid.");
        System.err.println();
        return false;
      }

      while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
        configFile.readLine();
      try {
        arrayOfdataRS[index].numberOfScheduleSlots = Integer.parseInt(configFile.readLine().trim());
      }catch(NumberFormatException e){
        System.err.println(e);
        arrayOfdataRS[index].numberOfScheduleSlots = -1;
      }
      if(arrayOfdataRS[index].numberOfScheduleSlots == -1){
        System.err.println("Error: The value of numberOfScheduleSlots read is not valid.");
        System.err.println();
        return false;
      }

      while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
        configFile.readLine();
      try {
        arrayOfdataRS[index].numberOfPackets = Integer.parseInt(configFile.readLine().trim());
      }catch(NumberFormatException e){
        System.err.println(e);
        arrayOfdataRS[index].numberOfPackets = -1;
      }
      if(arrayOfdataRS[index].numberOfPackets == -1){
        System.err.println("Error: The value of numberOfPackets read is not valid.");
        System.err.println();
        return false;
      }

      while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
        configFile.readLine();
      try {
        arrayOfdataRS[index].senderPort = Integer.parseInt(configFile.readLine().trim());
      }catch(NumberFormatException e){
        System.err.println(e);
        System.err.println("Error: Error reading Sender port number.");
        System.err.println();
        return false;
      }
      if (arrayOfdataRS[index].senderPort < 0 || arrayOfdataRS[index].senderPort >= 65536) {
        System.err.println("Error: Sender port number out of range (" +arrayOfdataRS[index].senderPort+").");
        System.err.println();
        return false;
      }

      while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
        configFile.readLine();
      try {
        arrayOfdataRS[index].receiverPort = Integer.parseInt(configFile.readLine().trim());
      }catch(NumberFormatException e){
        System.err.println(e);
        System.err.println("Error: Error reading Receiver port number.");
        System.err.println();
        return false;
      }
      if (arrayOfdataRS[index].receiverPort < 0 || arrayOfdataRS[index].receiverPort >= 65536) {
        System.err.println("Error: Receiver port number out of range (" +arrayOfdataRS[index].receiverPort+").");
        System.err.println();
        return false;
      }


      // read the sender address

      while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
        configFile.readLine();

      /* The sender address should be in the following format:

            IPv4_address:TCPportNumber    // for IPv4 address
            [IPv6_address]:TCPportNumber  // for IPv6 address

         The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the
         Control-Client and the respective OWAMP-Sender. If the user wants to use the default TCPportNumber
         the sender address can be in the following format:

            IPv4_address    // for IPv4 address
            [IPv6_address]  // for IPv6 address
      */

      String address =  configFile.readLine().trim();
      String addressAux = address;
      int senderTCPportNumber = DEFAULT_SENDER_PORT;  // TCPportNumber to use in OWAMP-Control with sender.
      boolean IPv6 = false;
      int i=0;

      try{

        if(address.indexOf("[") != -1){  // IPv6 address
          IPv6 = true;
          if(address.indexOf("]:") != -1){  // TCPportNumber to use in OWAMP-Control with sender specified.
            senderTCPportNumber = Integer.parseInt(address.substring(address.indexOf("]:")+2));
          }
          address = address.substring(address.indexOf("[")+1,address.indexOf("]"));
        }else{  // IPv4 address
          if(address.indexOf(":") != -1){  // TCPportNumber to use in OWAMP-Control with sender specified.
            senderTCPportNumber = Integer.parseInt(address.substring(address.indexOf(":")+1));
            address = address.substring(0,address.indexOf(":"));
          }
        }

        if (senderTCPportNumber < 0 || senderTCPportNumber >= 65536) {
          System.err.println("Error: Sender TCP port number out of range (" +senderTCPportNumber+").");
          System.err.println();
          return false;
        }
        arrayOfSenderTCPportNumber[index] = senderTCPportNumber;

        InetAddress inetAddress[] = InetAddress.getAllByName(address);     // The array of all the IP addresses for a given host name.

        for(i=0;i<inetAddress.length;i++){
          if(IPv6 && (inetAddress[i].getAddress().length == 16)){  // IPv6 address
            senderAddress = inetAddress[i].getAddress();
            break;
          }

          if(!IPv6 && (inetAddress[i].getAddress().length == 4)){  // IPv4 address
            senderAddress = inetAddress[i].getAddress();
            break;
          }

          if(i == (inetAddress.length-1)){  // last cicle.
            System.err.println("Error: There isn't a valid IPv"+(IPv6?6:4)+" address for the sender address introduced ("+address+").");
            System.err.println();
            return false;
          }
        }

        // The sender and receiver addresses can only be both loopback addresses if the Control-Client
        // is on the same machine were the Server is running.
        if(InetAddress.getByAddress(senderAddress).isLoopbackAddress() && !isAnyLocalAddress(serverAddress)){
          System.err.println("Error: The sender address can only be loopback address if the Control-Client is on the same machine were the Server is running.");
          System.err.println();
          return false;
        }

        System.arraycopy(senderAddress,0,arrayOfdataRS[index].senderAddress,0,senderAddress.length);

      }catch(NumberFormatException e){
        System.err.println("Error: The the sender address read is not a valid address ("+addressAux+").");
        System.err.println("The sender address should be in the following format:\n\n"+
                           "    IPv4_address:TCPportNumber    // for IPv4 address\n"+
                           "    [IPv6_address]:TCPportNumber  // for IPv6 address\n\n"+
                           "The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the \n"+
                           "Control-Client and the respective OWAMP-Sender. If the user wants to use the default TCPportNumber \n"+
                           "the sender address can be in the following format:\n\n"+
                           "    IPv4_address    // for IPv4 address\n"+
                           "    [IPv6_address]  // for IPv6 address\n");
        System.err.println();
        return false;
      }catch(Exception e2){
        System.err.println("Error: The the sender address read is not a valid address ("+addressAux+").");
        System.err.println("The sender address should be in the following format:\n\n"+
                           "    IPv4_address:TCPportNumber    // for IPv4 address\n"+
                           "    [IPv6_address]:TCPportNumber  // for IPv6 address\n\n"+
                           "The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the \n"+
                           "Control-Client and the respective OWAMP-Sender. If the user wants to use the default TCPportNumber \n"+
                           "the sender address can be in the following format:\n\n"+
                           "    IPv4_address    // for IPv4 address\n"+
                           "    [IPv6_address]  // for IPv6 address\n");
        System.err.println();
        return false;
      }


      // read the receiver address

      while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
        configFile.readLine();

      /* The receiver address should be in the following format:

             IPv4_address     // for IPv4 address
             [IPv6_address]   // for IPv6 address

         The TCP port to be used in OWAMP-Control communication between the Server and receiver should be defined
         when running OWAMP_Server, or the default TCP port will be used.
      */

      address =  configFile.readLine().trim();
      addressAux = address;
      IPv6 = false;

      try{

        if(address.indexOf("[") != -1){  // IPv6 address
          IPv6 = true;
          arrayOfdataRS[index].mbz_iPVN = (byte)6;
          if(address.indexOf("]:") != -1){  // TCPportNumber to use in OWAMP-Control with receiver specified.
            System.err.println("Error: The TCP port to be used in OWAMP-Control communication between the Server and receiver should be defined when running OWAMP_Server.");
            System.err.println();
            return false;
          }
          address = address.substring(address.indexOf("[")+1,address.indexOf("]"));
        }else{  // IPv4 address
          arrayOfdataRS[index].mbz_iPVN = (byte)4;
          if(address.indexOf(":") != -1){  // TCPportNumber to use in OWAMP-Control with receiver specified.
            System.err.println("Error: The TCP port to be used in OWAMP-Control communication between the Server and receiver should be defined when running OWAMP_Server.");
            System.err.println();
            return false;
          }
        }

        InetAddress inetAddress[] = InetAddress.getAllByName(address);     // The array of all the IP addresses for a given host name.

        for(i=0;i<inetAddress.length;i++){
          if(IPv6 && (inetAddress[i].getAddress().length == 16)){  // IPv6 address
            receiverAddress = inetAddress[i].getAddress();
            break;
          }

          if(!IPv6 && (inetAddress[i].getAddress().length == 4)){  // IPv4 address
            receiverAddress = inetAddress[i].getAddress();
            break;
          }

          if(i == (inetAddress.length-1)){  // last cicle.
            System.err.println("Error: There isn't a valid IPv"+(IPv6?6:4)+" address for the receiver address introduced ("+address+").");
            System.err.println();
            return false;
          }
        }

        // If sender address is loopback address receiver should be loopback address too. If receiver address is loopback address sender should be loopback address too.
        if((InetAddress.getByAddress(senderAddress).isLoopbackAddress()&&!InetAddress.getByAddress(receiverAddress).isLoopbackAddress())||(!InetAddress.getByAddress(senderAddress).isLoopbackAddress()&&InetAddress.getByAddress(receiverAddress).isLoopbackAddress())){
          if(InetAddress.getByAddress(senderAddress).isLoopbackAddress())
            System.err.println("Error: The sender address is loopback address and receiver address isn't.");
          if(InetAddress.getByAddress(receiverAddress).isLoopbackAddress())
            System.err.println("Error: The receiver address is loopback address and sender address isn't.");
          System.err.println();
          return false;
        }

        if(senderAddress.length != receiverAddress.length){  // The sender and receiver address should have the same IP version.
          System.err.println("Error: The sender and receiver address should have the same IP version.");
          System.err.println();
          return false;
        }

        System.arraycopy(receiverAddress,0,arrayOfdataRS[index].receiverAddress,0,receiverAddress.length);

      }catch(NumberFormatException e){
        System.err.println("Error: The the receiver address read is not a valid address ("+addressAux+").");
        System.err.println("The receiver address should be in the following format:\n\n"+
                           "    IPv4_address    // for IPv4 address\n"+
                           "    [IPv6_address]  // for IPv6 address\n");
        System.err.println();
        return false;
      }catch(Exception e2){
        System.err.println("Error: The the receiver address read is not a valid address ("+addressAux+").");
        System.err.println("The receiver address should be in the following format:\n\n"+
                           "    IPv4_address    // for IPv4 address\n"+
                           "    [IPv6_address]  // for IPv6 address\n");
        System.err.println();
        return false;
      }


      while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
        configFile.readLine();

      try {
        arrayOfdataRS[index].paddingLenght = Integer.parseInt(configFile.readLine().trim());
      }catch(NumberFormatException e){
        System.err.println(e);
        arrayOfdataRS[index].paddingLenght = -1;
      }
      if(arrayOfdataRS[index].paddingLenght == -1){
        System.err.println("Error: The value of paddingLenght read is not valid.");
        System.err.println();
        return false;
      }

      while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
        configFile.readLine();
      int year = 0;
      try {
        year = Integer.parseInt(configFile.readLine().trim());
      }catch(NumberFormatException e){
        System.err.println(e);
        System.err.println("Error: Error reading the year value.");
        System.err.println();
        return false;
      }
      if((year < 1970)||(year > 2036)){
        System.err.println("Error: The value of year read is not valid.");
        System.err.println();
        return false;
      }

      while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
        configFile.readLine();
      int month = 0;
      try {
        month = (Integer.parseInt(configFile.readLine().trim())-1);  // 0 represents January and 11 December.
      }catch(NumberFormatException e){
        System.err.println(e);
        System.err.println("Error: Error reading the month value.");
        System.err.println();
        return false;
      }
      if((month < 0)||(month > 11)){
        System.err.println("Error: The value of month read is not valid.");
        System.err.println();
        return false;
      }

      while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
        configFile.readLine();
      int date = 0;
      try {
        date = Integer.parseInt(configFile.readLine().trim());
      }catch(NumberFormatException e){
        System.err.println(e);
        System.err.println("Error: Error reading the date value.");
        System.err.println();
        return false;
      }
      if((date < 1)||(date > 31)){
        System.err.println("Error: The value of date read is not valid.");
        System.err.println();
        return false;
      }

      while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
        configFile.readLine();
      int hour = 0;
      try {
        hour = Integer.parseInt(configFile.readLine().trim());
      }catch(NumberFormatException e){
        System.err.println(e);
        System.err.println("Error: Error reading the hour value.");
        System.err.println();
        return false;
      }
      if((hour < 0)||(hour > 23)){
        System.err.println("Error: The value of hour read is not valid.");
        System.err.println();
        return false;
      }

      while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
        configFile.readLine();
      int minute = 0;
      try {
        minute = Integer.parseInt(configFile.readLine().trim());
      }catch(NumberFormatException e){
        System.err.println(e);
        System.err.println("Error: Error reading the minute value.");
        System.err.println();
        return false;
      }
      if((minute < 0)||(minute > 59)){
        System.err.println("Error: The value of minute read is not valid.");
        System.err.println();
        return false;
      }

      while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
        configFile.readLine();
      int second = 0;
      try {
        second = Integer.parseInt(configFile.readLine().trim());
      }catch(NumberFormatException e){
        System.err.println(e);
        System.err.println("Error: Error reading the second value.");
        System.err.println();
        return false;
      }
      if((second < 0)||(second > 59)){
        System.err.println("Error: The value of second read is not valid.");
        System.err.println();
        return false;
      }

      Calendar startTime = Calendar.getInstance();
      startTime.set(year, month, date, hour, minute, second);
      startTime.set(startTime.MILLISECOND, 0);

      arrayOfdataRS[index].startTime = new Timestamp(startTime.getTimeInMillis(), (short)1); // Time in milliseconds since 1Jan1970

      while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
        configFile.readLine();
      double timeOut = 0;
      try {
        timeOut = Double.parseDouble(configFile.readLine().trim());
      }catch(NumberFormatException e){
        System.err.println(e);
        timeOut = -1;
      }
      if(timeOut < 0){
        System.err.println("Error: The value of timeOut read is not valid.");
        System.err.println();
        return false;
      }

      arrayOfdataRS[index].timeout = new Timestamp(timeOut, (short)0);    // Timeout in seconds since 1Jan1900 = timeout seconds * 1000


      byte typePDescriptor_Type = 0;
      while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
        configFile.readLine();
      try {
        typePDescriptor_Type = (byte)Integer.parseInt(configFile.readLine().trim());
      }catch(NumberFormatException e){
        System.err.println(e);
        typePDescriptor_Type = -1;
      }
      if((typePDescriptor_Type == -1)||(typePDescriptor_Type != 0)&&(typePDescriptor_Type != 1)){
        System.err.println("Error: The type of typePDescriptor read is not valid.");
        System.err.println();
        return false;
      }

      while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
        configFile.readLine();
      int typePDescriptor_Value = 0;
      try {
        typePDescriptor_Value = Integer.parseInt(configFile.readLine().trim());
      }catch(NumberFormatException e){
        System.err.println(e);
        typePDescriptor_Value = -1;
      }
      if(typePDescriptor_Value == -1){
        System.err.println("Error: The value of typePDescriptor read is not valid.");
        System.err.println();
        return false;
      }

      arrayOfdataRS[index].typeP_Descriptor = generateTypeP_Descriptor(typePDescriptor_Type,typePDescriptor_Value);


      arrayOfdataSSD[index] = new ScheduleSlotDescriptionsToBeSend[arrayOfdataRS[index].numberOfScheduleSlots];

      for(i=0;i<arrayOfdataRS[index].numberOfScheduleSlots;i++){

        arrayOfdataSSD[index][i] = new ScheduleSlotDescriptionsToBeSend();

        while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
          configFile.readLine();
        try {
          arrayOfdataSSD[index][i].slotType = (byte)Integer.parseInt(configFile.readLine().trim());
        }catch(NumberFormatException e){
          System.err.println(e);
          arrayOfdataSSD[index][i].slotType = -1;
        }
        if((arrayOfdataSSD[index][i].slotType != 0)&&(arrayOfdataSSD[index][i].slotType != 1)){
          System.err.println("Error: The value of slotType "+i+" read is not valid.");
          System.err.println();
          return false;
        }

        while(configFile.readByte() == 35)  // 35 code ASCII of character '#'
          configFile.readLine();
        double parameter = 0;
        try {
          parameter = Double.parseDouble(configFile.readLine().trim());
        }catch(NumberFormatException e){
          System.err.println(e);
          parameter = -1;
        }
        if(parameter < 0){
          System.err.println("Error: The value of slot parameter "+i+" read is not valid.");
          System.err.println();
          return false;
        }

        if(arrayOfdataSSD[index][i].slotType == 0)
          arrayOfdataSSD[index][i].slotParameter = generateSlotParameter(parameter);
        else
          arrayOfdataSSD[index][i].slotParameter = new Timestamp(parameter, (short)0);

      }

      return true;
    }
    catch(IOException ex){
      System.err.println(ex);
      System.err.println("Error: Error reading from configurations file.");
      System.err.println();
      System.err.println("Possible error:");
      System.err.println("i)  The format required is not followed.");
      System.err.println("ii) Error reading from file");
      System.err.println();
    }

    return false;

  }


  /**
   * The user can execute this module with all the configurations parameters in the command line. In
   * this case the user must instroduce 28 parameters. This mode of execution is also used in the
   * Web interface. The arrayOfRequestSession is saved to a database.
   *
   * @param userID String - is the string representation of the user id.
   * @param sqlServerAddress String - it's the data base server's address.
   * @param dataBase String - it's the database used to put the table with all Request Sessions.
   * @param userName String - user name.
   * @param passWord String - user password.
   * @param table String - it's the table which have all RS.
   * @param _senderAddress String - the sender address entered
   * @param _receiverAddress String - the receiver address entered
   * @param slotParameter double - to keep its original value, in case of slotType 0, it should be passed as an argument.
   * If slotType is 0 arrayOfdataSSD[][].slotParameter doesn't have the original slotParameter.
   * @return boolean  -  It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean processSingleSession(String userID, String sqlServerAddress, String dataBase, String userName,
                                      String passWord, String table, String _senderAddress, String _receiverAddress,
                                      double slotParameter){

    // Test socket connection

    if(isSocketConnected(socketToServer)){
      System.err.println("Error: The client is already connected to this server. You can not process this operation again.");
      System.err.println();
      return false;
    }

    // Save the arrayOfDataRS to the data base (to the jowampRS table). Even if the startTime is in the future
    // we should save the configuration details of this test session in the data base because this information
    // will be used in the JOWAMP web interface, to show all the sessions configured in the "Session Results" section.

    // Establish a connection to the data base.

    MySQLcom MySQLclient= null;
    Connection con = null;
    try{
      MySQLclient = new MySQLcom (sqlServerAddress, dataBase, userName, passWord);
      con = MySQLclient.getConnection();
    }catch(Exception e){
      System.out.println(e);
      System.err.println("Error: This session will not be started.");
      System.err.println();
      return false;
    }

    // Test if the table of jowampRS (table of all RS made) exists in the dataBase. If it doesn't exist
    // it should be created.

    try{
      MySQLclient.executeUpdate ( con, "CREATE TABLE `"+table+"` (id INT NOT NULL AUTO_INCREMENT, userid VARCHAR(39), mbzIPVN VARCHAR(1) /* IPVN */, confSender BIT, confReceiver BIT, numberOfScheduleSlots INT, numberOfPackets INT, senderPort INT, receiverPort INT, senderAddress VARCHAR(47), receiverAddress VARCHAR(41), SID VARCHAR(32), paddingLength INT, startHour INT(2) /*start Hour*/, startMinute INT(2) /*start Minute*/, startSecond INT(2) /*start Second*/, startDate INT(2) /*start Date*/, startMonth INT(2) /*start Month*/, startYear INT(4) /*start Year*/, timeOut DOUBLE /*Timeout in seconds*/, typeP_Descriptor VARCHAR(4), slotType INT(1), slotParameter DOUBLE /*Slot parameter in seconds*/, totalTimeToWait DOUBLE /*Total test duration in milliseconds*/, realStartTime BIGINT /*Time when session started in milliseconds since 0h 1 Jan 1970*/, PRIMARY KEY (id))" );
    }catch(Exception e){  // If the table exists, an exception is thrown.
      try{ // Verify if the exception occurred because the table of jowampRS already exits in the database. If not, an error ocurred
           // and this method should return false.
        MySQLclient.executeQuery(con, "SELECT COUNT(*) FROM `"+table+"`"); // If the table doesn't exist, an exception is thrown
      }catch(Exception e2){
        System.out.println(e);
        MySQLclient.close(con);
        System.err.println("Error: This session will not be started.");
        System.err.println();
        return false;
      }
    }

    // Generate a SID value to be used to identify this session in the jowampRS table after the RS is made, to update
    // the entry corresponding to this session. This value doesn't correspond to the session SID. It will be changed
    // to the session SID when the RS is made.
    String SIDAux = owampHexEncode(generateSid());   // generates the value for the variable sid
    double totalTimeToWait = 0;

    try{
      Calendar startTime = Calendar.getInstance();
      startTime.setTimeInMillis((long)arrayOfdataRS[0].startTime.timeInMillisecondsSince1970());

      // Calculate the total time to wait

      for(int i=0, j = 0;i<(arrayOfdataRS[0].numberOfPackets-1);i++,j++){
        totalTimeToWait = totalTimeToWait + arrayOfdataSSD[0][j].slotParameter.timeInMillisecondsSince1900();
        if(j == (arrayOfdataSSD[0].length -1))  // If we are at the end of the array of SSD we have to
          j = -1;                               // turn around.  j = -1 -> when incremented by the cycle
                                                // for, it will have the correct value (zero).
      }
      totalTimeToWait = totalTimeToWait + arrayOfdataRS[0].timeout.timeInMillisecondsSince1900();

      String typeP_Descriptor = "";
      for(int i = 0;i<arrayOfdataRS[0].typeP_Descriptor.length;i++)
        typeP_Descriptor = typeP_Descriptor+""+(0xFF & arrayOfdataRS[0].typeP_Descriptor[i]);

      MySQLclient.executeUpdate(con,
                                "INSERT INTO `" + table + "` (userid, mbzIPVN, confSender, confReceiver, numberOfScheduleSlots, numberOfPackets, senderAddress, receiverAddress, SID, paddingLength, startHour /*start Hour*/, startMinute /*start Minute*/, startSecond /*start Second*/, startDate /*start Date*/, startMonth /*start Month*/, startYear /*start Year*/, timeOut, typeP_Descriptor, slotType, slotParameter, totalTimeToWait) values ('" + userID +"', '" +
                                arrayOfdataRS[0].mbz_iPVN + "', '" +
                                arrayOfdataRS[0].confSender + "', '" +
                                arrayOfdataRS[0].confReceiver + "', '" +
                                arrayOfdataRS[0].numberOfScheduleSlots + "', '" +
                                arrayOfdataRS[0].numberOfPackets + "', '" +
                                _senderAddress + "', '" +
                                _receiverAddress + "', '" +
                                SIDAux + "', '" +
                                arrayOfdataRS[0].paddingLenght + "', '" +
                                startTime.get(startTime.HOUR_OF_DAY) + "', '" +
                                startTime.get(startTime.MINUTE) + "', '" +
                                startTime.get(startTime.SECOND) + "', '" +
                                startTime.get(startTime.DAY_OF_MONTH) + "', '" +
                                (startTime.get(startTime.MONTH) + 1) + "', '" +
                                startTime.get(startTime.YEAR) + "', '" +
                                (arrayOfdataRS[0].timeout.timeInMillisecondsSince1900()/1000)+ "', '" + /*timeOut in seconds*/
                                typeP_Descriptor+ "', '" +
                                arrayOfdataSSD[0][0].slotType + "', '" + slotParameter + "', '" +  /*slotParameter in seconds*/
                                totalTimeToWait+"')"); /*in milliseconds*/
    }
    catch (Exception e) {
      MySQLclient.close(con);
      System.out.println(e);
      System.err.println("Error: This session will not be started.");
      System.err.println();
      return false;
    }
    MySQLclient.close(con);

    // Wait for the startTime before processing this session. We just should start processing this session 2 minutes
    // (its a reasonable interval of time) before its startTime less timeout if it is in the future, otherwise the session should
    // be started imediatly. Note: If The startTime is in the past and separated from the present by more than Timeout, the
    // session will not be started. So we should compare the current time with (startTime - timeout) less 2 minutes.
    // Using this we can prevent from openning a socket connection to Server too early when the startTime is in the future.

    long startTimeLessTwoMinutes =((long)arrayOfdataRS[0].startTime.timeInMillisecondsSince1970() - (long)arrayOfdataRS[0].timeout.timeInMillisecondsSince1900()) - 120*1000;

    if(startTimeLessTwoMinutes > (long)hiResTimer.milliTime())
      try{
        sleep(startTimeLessTwoMinutes-(long)hiResTimer.milliTime());
      }catch(Exception e){}


    // process connectionSetup to Server

    // We should set the socket to server's timeout to 2 * TIMEOUT because the server waits
    // for TIMEOUT ms to receive any packet from Receiver, and then it sends it to Control-Client.
                                // connectionSetup(int mode, InetAddress ia, int port, int timeout)
    socketToServer = connectionSetupRequest(mode,serverAddress,serverTCPportNumber,2*TIMEOUT);

    if(socketToServer == null){
      System.err.println("Error: The Server specified doesn't exist.\n");
      System.err.println("Error: Error in processing Connection Setup.");
      System.err.println();
      return false;
    }

    // process requestSession

    boolean closeConnection = false;

    if(!requestSession(closeConnection)){
      System.err.println("Error: Error in processing Request Session.");
      System.err.println();
      return false;
    }

    // Establish a connection to the data base.

    try{
      MySQLclient = new MySQLcom (sqlServerAddress, dataBase, userName, passWord);
      con = MySQLclient.getConnection();
    }catch(Exception e){
      System.out.println(e);
      System.err.println("Error: All sessions will be finished.");
      System.err.println();
      StopSessions.stopAllSessions(socketToServer,arrayOfSocketToSender,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
      return false;
    }

    // Save the arrayOfDataRS to the data base (to the jowampRS table) (Update configuration information).
    long realStartTime = 0;
    try{

      // Calculate the real startTime
      realStartTime = (arrayOfdataRS[0].startTime.timeInMillisecondsSince1970() < (long)hiResTimer.milliTime())?(long)hiResTimer.milliTime():(long)arrayOfdataRS[0].startTime.timeInMillisecondsSince1970();

      MySQLclient.executeUpdate(con,
                                  "UPDATE `" + table + "` SET senderPort = "+arrayOfdataRS[0].senderPort + ", receiverPort = "+arrayOfdataRS[0].receiverPort + ", SID = '"+owampHexEncode(arrayOfdataRS[0].sid) +
                                  "', realStartTime = "+realStartTime+   /*in milliseconds*/
                                  " WHERE SID = '"+SIDAux +"'");
    }
    catch (Exception e) {
      MySQLclient.close(con);
      System.out.println(e);
      System.err.println("Error: All sessions will be finished.");
      System.err.println();
      StopSessions.stopAllSessions(socketToServer,arrayOfSocketToSender,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
      return false;
    }
    MySQLclient.close(con);

    // process Starting Test Sessions

    option = -1;  // Start Test Sessions
    if(!startingTestSessions(false)){  // startingTestSessions(boolean saveArrayOfRStoFiles)
      System.err.println("Error: Error in processing Starting Test Session.");
      System.err.println();
      MySQLclient.close(con);
      closeSocket(socketToServer); // close socket connection to Server
      return false;
    }

    // The Client should execute the Fetch Session at the end of the session.

    OWAMP_FetchClient fetch = new OWAMP_FetchClient(serverAddress, serverTCPportNumber, socketToServer, systemInReader, hiResTimer);

    // processFetchSessionToDataBase(String sqlServerAddress, String dataBase, String userName, String passWord,
    //                               String strSid, int numberOfPackets, double slotParameter /*in seconds*/,
    //                               long realStartTime /*in milliseconds*/, double totalTimeToWait /*in milliseconds*/){ // process Fetch-Session

    if(!fetch.processFetchSessionToDataBase(sqlServerAddress, dataBase, userName, passWord, owampHexEncode(arrayOfdataRS[0].sid), arrayOfdataRS[0].numberOfPackets, slotParameter, realStartTime, totalTimeToWait)){
      System.err.println("Error: Error processing Fetch Session.");
      System.err.println();
      closeSocket(socketToServer); // close socket connection to Server
      return false;
    }

    closeSocket(socketToServer); // close socket connection to Server

    return true;

  }


  /**
   * This method is used to process a Confidence Interval session and save the session's configuration and
   * fetched results to a database.
   *
   * @param userID String - is the string representation of the user id.
   * @param sqlServerAddress String - it's the data base server's address.
   * @param dataBase String - it's the database used to save the session's configuration and fetched results.
   * @param userName String - user name.
   * @param passWord String - user password.
   * @param CIconfigTable String - table used to save the CI session's configuration.
   * @param CIsidTable String - table used to save the session's SID, senderPort, receiverPort, real start time and total test duration for
   * each of the individual single session of each interval.
   * Note: If senderPort = 0 a free UDP port will be used to send test packets.
   * @param _senderAddress String - the sender address entered
   * @param _receiverAddress String - the receiver address entered
   * @param timeBetweenSetOfTests int - it's the time duration in milliseconds between the start time of each
   * consecutive intervals.
   * @param timeBetweenTests int - it's the time duration in milliseconds between the start time of each consecutive
   * single session of an interval.
   * @param numInterval int - it's the total number of intervals.
   * @param numOfTestInInterval int - it's the total number of single tests to be used in each interval.
   * @param slotParameter double - to keep its original value, in case of slotType 0, it should be passed as an argument.
   * If slotType is 0 arrayOfdataSSD[][].slotParameter doesn't have the original slotParameter. The should be in seconds.
   * @param typePDescriptor_Type String - typePDescriptor_Type
   * @param typePDescriptor_Value String - typePDescriptor_Value
   * @return boolean  -  It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean processCIsession(String userID, String sqlServerAddress, String dataBase, String userName,
                                  String passWord, String CIconfigTable, String CIsidTable, String _senderAddress, String _receiverAddress,
                                  int timeBetweenSetOfTests /*in milliseconds*/, int timeBetweenTests /*in milliseconds*/,
                                  int numIntervals, int numOfTestsInInterval, double slotParameter /*in seconds*/,
                                  String typePDescriptor_Type, String typePDescriptor_Value){

    // Test socket connection

    if (isSocketConnected(socketToServer)) {
      System.err.println("Error: The client is already connected to this server. You can not process this operation again.");
      System.err.println();
      return false;
    }

    // Establish a connection to the data base.

    MySQLcom MySQLclient = null;
    Connection con = null;
    try { // String sqlServerAddress, String dataBase, String userName, String passWord
      MySQLclient = new MySQLcom(sqlServerAddress, dataBase, userName, passWord);
      con = MySQLclient.getConnection();
    }catch (Exception e) {
      System.out.println(e);
      System.err.println("Error: This session will not be started.");
      System.err.println();
      System.exit(1);
    }

    // Test if the table CIconfigTable (table with the CI configurations of all CI sessions made) exists in
    // the dataBase. If it doesn't exist it should be created.

    try { // String CIconfigTable
      MySQLclient.executeUpdate(con, "CREATE TABLE `" + CIconfigTable + "` (id INT NOT NULL AUTO_INCREMENT, userid VARCHAR(39), mbzIPVN VARCHAR(1) /* IPVN */, confSender BIT, confReceiver BIT, numberOfScheduleSlots INT, numberOfPackets INT, senderPort INT, receiverPort INT, senderAddress VARCHAR(47), receiverAddress VARCHAR(41), paddingLength INT, startHour INT(2) /*start Hour*/, startMinute INT(2) /*start Minute*/, startSecond INT(2) /*start Second*/, startDate INT(2) /*start Date*/, startMonth INT(2) /*start Month*/, startYear INT(4) /*start Year*/, timeOut DOUBLE /*Timeout in seconds*/, typePDescriptor_Type VARCHAR(4), typePDescriptor_Value VARCHAR(4), slotType INT(1), slotParameter DOUBLE /*Slot parameter in seconds*/, timeBetweenSetOfTests DOUBLE, timeBetweenTests DOUBLE, numIntervals INT, numOfTestsInInterval INT, PRIMARY KEY (id))");
    }
    catch (Exception e) { // If the table exists, an exception is thrown.
      try { // Verify if the exception occurred because the table CIconfigTable already exits in the database. If not, an error ocurred
            // and this method should return false.
        MySQLclient.executeQuery(con, "SELECT COUNT(*) FROM `" + CIconfigTable+"`"); // If the table doesn't exist, an exception is thrown
      }
      catch (Exception e2) {
        System.out.println(e);
        MySQLclient.close(con);
        System.err.println("Error: This session will not be started.");
        System.err.println();
        System.exit(1);
      }
    }

    // Test if the table of CIsidTable (table used to save the session's SID, senderPort, receiverPort, real start time and total test
    // duration (totalTimeToWait) for each of the individual single session of each interval.) exists in the dataBase.
    // If it doesn't exist it should be created.

    try { // String CIsidTable
      MySQLclient.executeUpdate(con, "CREATE TABLE `" + CIsidTable + "` (id INT NOT NULL AUTO_INCREMENT, SID VARCHAR(32) NOT NULL, senderPort INT, receiverPort INT, totalTimeToWait DOUBLE /*Total test duration in milliseconds*/, realStartTime BIGINT /*Time when session started in milliseconds since 0h 1 Jan 1970*/, PRIMARY KEY (id))");
    }
    catch (Exception e) { // If the table exists, an exception is thrown.
      try { // Verify if the exception occurred because the table CIsidTable already exits in the database. If not, an error ocurred
            // and this method should return false.
        MySQLclient.executeQuery(con, "SELECT COUNT(*) FROM `" + CIsidTable+"`"); // If the table doesn't exist, an exception is thrown
      }
      catch (Exception e2) {
        System.out.println(e);
        MySQLclient.close(con);
        System.err.println("Error: This session will not be started.");
        System.err.println();
        System.exit(1);
      }
    }

    // Save the CI session's configurations to CIconfigTable table.

    Calendar sessionStartTime = Calendar.getInstance();
    sessionStartTime.setTimeInMillis((long)arrayOfdataRS[0].startTime.timeInMillisecondsSince1970());

    try {
      MySQLclient.executeUpdate(con,
                                "INSERT INTO `" + CIconfigTable + "` (userid, mbzIPVN, confSender, confReceiver, numberOfScheduleSlots, numberOfPackets, senderPort, receiverPort, senderAddress, receiverAddress, paddingLength, startHour /*start Hour*/, startMinute /*start Minute*/, startSecond /*start Second*/, startDate /*start Date*/, startMonth /*start Month*/, startYear /*start Year*/, timeOut, typePDescriptor_Type, typePDescriptor_Value, slotType, slotParameter, timeBetweenSetOfTests, timeBetweenTests, numIntervals, numOfTestsInInterval) values ('" +
                                userID + "', '" +
                                arrayOfdataRS[0].mbz_iPVN + "', '" +
                                arrayOfdataRS[0].confSender + "', '" +
                                arrayOfdataRS[0].confReceiver + "', '" +
                                arrayOfdataRS[0].numberOfScheduleSlots + "', '" +
                                arrayOfdataRS[0].numberOfPackets + "', '" +
                                arrayOfdataRS[0].senderPort + "', '" +
                                arrayOfdataRS[0].receiverPort + "', '" +
                                _senderAddress + "', '" +
                                _receiverAddress + "', '" +
                                arrayOfdataRS[0].paddingLenght + "', '" +
                                sessionStartTime.get(sessionStartTime.HOUR_OF_DAY) + "', '" +
                                sessionStartTime.get(sessionStartTime.MINUTE) + "', '" +
                                sessionStartTime.get(sessionStartTime.SECOND) + "', '" +
                                sessionStartTime.get(sessionStartTime.DAY_OF_MONTH) + "', '" +
                                (sessionStartTime.get(sessionStartTime.MONTH) + 1) + "', '" +
                                sessionStartTime.get(sessionStartTime.YEAR) + "', '" +
                                (arrayOfdataRS[0].timeout.timeInMillisecondsSince1900() / 1000) +"', '" + /*timeOut in seconds*/
                                typePDescriptor_Type+ "', '" +typePDescriptor_Value + "', '" +
                                arrayOfdataSSD[0][0].slotType + "', '" + slotParameter +"', '" + /*slotParameter in seconds*/
                                (timeBetweenSetOfTests/1000)/60 /*in minutes*/+ "', '" +
                                timeBetweenTests/1000 /*in seconds*/+ "', '" +
                                numIntervals+ "', '" +
                                numOfTestsInInterval + "')");
    }
    catch (Exception e) {
      MySQLclient.close(con);
      System.out.println(e);
      System.err.println("Error: This session will not be started.");
      System.err.println();
      return false;
    }
    MySQLclient.close(con);

    // Save the inicial values of senderPort and receiverPort because if senderPort = 0 a free UDP port will be used,
    // so it should be zero for all single sessions started.

    int senderPort = arrayOfdataRS[0].senderPort,
        receiverPort = arrayOfdataRS[0].receiverPort;

    // Calculate the total time to wait for the first single test session. If slot type is 1, this value will be
    // the same for all single sessions of this CI session.

    double totalTimeToWait = 0;
    for(int i=0, j = 0;i<(arrayOfdataRS[0].numberOfPackets-1);i++,j++){
      totalTimeToWait = totalTimeToWait + arrayOfdataSSD[0][j].slotParameter.timeInMillisecondsSince1900();
      if(j == (arrayOfdataSSD[0].length -1))  // If we are at the end of the array of SSD we have to
        j = -1;                               // turn around.  j = -1 -> when incremented by the cycle
                                              // for, it will have the correct value (zero).
    }
    totalTimeToWait = totalTimeToWait + arrayOfdataRS[0].timeout.timeInMillisecondsSince1900();


    // Process each Single Session of each interval

    int c = 0, num = 0;
    long startTime = sessionStartTime.getTimeInMillis() /*Start time for this CI session is the startTime of the first SS*/,
         testStartTime = 0, time = 0, startTimeLessTwoMinutes = 0, realStartTime = 0;
     boolean closeConnection = false;

    for (int i = 0; i < numIntervals; i++) { // for each interval
      testStartTime = startTime + i * timeBetweenSetOfTests; // Test start time for this interval in milliseconds
      for (int j = 0; j < numOfTestsInInterval; j++) { // for each test in this interval

        time = testStartTime + j * (timeBetweenTests); // StartTime for this session in milliseconds.

        arrayOfdataRS[0].senderPort = senderPort;
        arrayOfdataRS[0].receiverPort = receiverPort;
        arrayOfdataRS[0].startTime = new Timestamp(time, (short)1); // Time in milliseconds since 1Jan1970
        numRS = 0;    // Number of Request Session made
        seed = 0;     // seed used to generate random values.
        numSDR = 0;   // Number of SDR received from Sender.

        // If slot type is 0 the parameter and totalTimeToWait values may be different for each session. In this
        // case we have to calculate slot parameter and totalTimeToWait again.
        if((arrayOfdataSSD[0][0].slotType == 0)&&(time != startTime)){

          for (c = 0; c < arrayOfdataRS[0].numberOfScheduleSlots; c++)
            arrayOfdataSSD[0][c].slotParameter = generateSlotParameter(slotParameter);

          totalTimeToWait = 0;
          for(c=0, num = 0;c<(arrayOfdataRS[0].numberOfPackets-1);c++,num++){
            totalTimeToWait = totalTimeToWait + arrayOfdataSSD[0][num].slotParameter.timeInMillisecondsSince1900();
            if(num == (arrayOfdataSSD[0].length -1))  // If we are at the end of the array of SSD we have to
              num = -1;                               // turn around.  num = -1 -> when incremented by the cycle
                                                      // for, it will have the correct value (zero).
          }
          totalTimeToWait = totalTimeToWait + arrayOfdataRS[0].timeout.timeInMillisecondsSince1900();
        }

        // Wait for the startTime before processing this session. We just should start processing this session 2 minutes
        // (its a reasonable interval of time) before its startTime less timeout if it is in the future, otherwise the session should
        // be started imediatly. Note: If The startTime is in the past and separated from the present by more than Timeout, the
        // session will not be started. So we should compare the current time with (startTime - timeout) less 2 minutes.
        // Using this we can prevent from openning a socket connection to Server too early when the startTime is in the future.

        startTimeLessTwoMinutes =((long)arrayOfdataRS[0].startTime.timeInMillisecondsSince1970() - (long)arrayOfdataRS[0].timeout.timeInMillisecondsSince1900()) - 120*1000;

        if(startTimeLessTwoMinutes > (long)hiResTimer.milliTime())
          try{
            sleep(startTimeLessTwoMinutes-(long)hiResTimer.milliTime());
          }catch(Exception e){}


        // process connectionSetup to Server

        // We should set the socket to server's timeout to 2 * TIMEOUT because the server waits
        // for TIMEOUT ms to receive any packet from Receiver, and then it sends it to Control-Client.
                                    // connectionSetup(int mode, InetAddress ia, int port, int timeout)
        socketToServer = connectionSetupRequest(mode,serverAddress,serverTCPportNumber,2*TIMEOUT);

        if(socketToServer == null){
          System.err.println("Error: The Server specified doesn't exist.\n");
          System.err.println("Error: Error in processing Connection Setup.");
          System.err.println();
          return false;
        }

        // process requestSession

        closeConnection = false;

        if(!requestSession(closeConnection)){
          System.err.println("Error: Error in processing Request Session.");
          System.err.println();
          return false;
        }

        // Calculate the real startTime
        realStartTime = (arrayOfdataRS[0].startTime.timeInMillisecondsSince1970() < (long)hiResTimer.milliTime())?(long)hiResTimer.milliTime():(long)arrayOfdataRS[0].startTime.timeInMillisecondsSince1970();

        // Establish a connection to the data base.

        try{
          con = MySQLclient.getConnection();
        }catch(Exception e){
          System.out.println(e);
          System.err.println("Error: All sessions will be finished.");
          System.err.println();
          StopSessions.stopAllSessions(socketToServer,arrayOfSocketToSender,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
          return false;
        }

        // Save the SID, senderPort, receiverPort, realStartTime and totalTimeToWait of this individual single session of this interval
        // in the table CIsidTable.
        try{

          MySQLclient.executeUpdate(con,
                                "INSERT INTO `" + CIsidTable + "` (sid, senderPort, receiverPort, realStartTime, totalTimeToWait) values ('" +
                                owampHexEncode(arrayOfdataRS[0].sid) + "', '" +
                                arrayOfdataRS[0].senderPort + "', '" +
                                arrayOfdataRS[0].receiverPort + "', '" +
                                realStartTime /*in milliseconds*/+ "', '" +
                                totalTimeToWait /*in milliseconds*/+ "')");

        }catch (Exception e) {
          MySQLclient.close(con);
          System.out.println(e);
          System.err.println("Error: All sessions will be finished.");
          System.err.println();
          StopSessions.stopAllSessions(socketToServer,arrayOfSocketToSender,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
          return false;
        }
        MySQLclient.close(con);

        // process Starting Test Sessions

        option = -1;  // Start Test Sessions
        if(!startingTestSessions(false)){  // startingTestSessions(boolean saveArrayOfRStoFiles)
          System.err.println("Error: Error in processing Starting Test Session.");
          System.err.println();
          MySQLclient.close(con);
          closeSocket(socketToServer); // close socket connection to Server
          return false;
        }

        // The Client should execute the Fetch Session at the end of the session.

        OWAMP_FetchClient fetch = new OWAMP_FetchClient(serverAddress, serverTCPportNumber, socketToServer, systemInReader, hiResTimer);

        // processFetchSessionToDataBase(String sqlServerAddress, String dataBase, String userName, String passWord,
        //                               String strSid, int numberOfPackets, double slotParameter /*in seconds*/,
        //                               long realStartTime /*in milliseconds*/, double totalTimeToWait /*in milliseconds*/){ // process Fetch-Session

        if(!fetch.processFetchSessionToDataBase(sqlServerAddress, dataBase, userName, passWord, owampHexEncode(arrayOfdataRS[0].sid), arrayOfdataRS[0].numberOfPackets, slotParameter, realStartTime, totalTimeToWait)){
          System.err.println("Error: Error processing Fetch Session.");
          System.err.println();
          closeSocket(socketToServer); // close socket connection to Server
          return false;
        }

        closeSocket(socketToServer); // close socket connection to Server

      }
    }

    return true;

  }


  // Main method

  public static void main(String[] args){

    BufferedReader systemInReader = new BufferedReader(new InputStreamReader(System.in));

    // The user can execute this module with all the configurations parameters in the command line. In
    // this case the user must instroduce 27 or 29 parameters. In this case he wants to save the results
    // to a data base. Parameters (String userID,String serverAddress, String _confSender,
    // String _confReceiver, String _numberOfScheduleSlots, String _numberOfPackets, String _senderPort,
    // String _receiverPort, String _senderAddress, String _receiverAddress, String _paddingLenght,
    // String _year, String _month, String _date, String _hour, String _minute, String _second,
    // String _timeOut, String _typePDescriptor_Type, String _typePDescriptor_Value, String _slotType,
    // String _slotParameter, String sqlServerAddress, String dataBase, String userName, String passWord,
    // String SSconfigTable, String ntpServerAddress, String updateInterval). The ntpServerAddress and
    // updateInteval are optional. If not defined the system time will be used to get the time reference.
    // This mode of execution is also used in the Web interface.

    if ((args.length == 27)||((args.length == 29))) {

      try {

        // Test if if the startTime is in the past, and separated from the present by more than
        // Timeout time. In this case, this session should be rejected.

        int year = Integer.parseInt(args[11].trim());
        if ( (year < 1970) || (year > 2036)) {
          System.err.println("Error: The value of year read is not valid.");
          System.err.println();
          System.exit(1);
        }

        int month = (Integer.parseInt(args[12].trim()) - 1); // 0 represents January and 11 December.
        if ( (month < 0) || (month > 11)) {
          System.err.println("Error: The value of month read is not valid.");
          System.err.println();
          System.exit(1);
        }

        int date = Integer.parseInt(args[13].trim());
        if ( (date < 1) || (date > 31)) {
          System.err.println("Error: The value of date read is not valid.");
          System.err.println();
          System.exit(1);
        }

        int hour = Integer.parseInt(args[14].trim());
        if ( (hour < 0) || (hour > 23)) {
          System.err.println("Error: The value of hour read is not valid.");
          System.err.println();
          System.exit(1);
        }

        int minute = Integer.parseInt(args[15].trim());
        if ( (minute < 0) || (minute > 59)) {
          System.err.println("Error: The value of minute read is not valid.");
          System.err.println();
          System.exit(1);
        }

        int second = Integer.parseInt(args[16].trim());
        if ( (second < 0) || (second > 59)) {
          System.err.println("Error: The value of second read is not valid.");
          System.err.println();
          System.exit(1);
        }

        Calendar startTime = Calendar.getInstance();
        // startTime.set(year, month, date, hour, minute, second)
        startTime.set(year, month, date, hour, minute, second);
        startTime.set(startTime.MILLISECOND, 0);

        double timeOut = Double.parseDouble(args[17]); // Timeout in seconds.
        if (timeOut < 0) {
          System.err.println("Error: The value of timeOut read is not valid.");
          System.exit(1);
        }

        // Verify if NTP server address was introduced.
        String  ntpServerAddress = null;  // The address of the NTP server to be used to get the time reference.
        int updateInterval = -1;          // The interval duration (in minutes) between two consecutives updates to time reference.

        if(args.length == 29){  // Read the NTP server's address

          ntpServerAddress = args[27].trim();
          try { // Test if the NTP server address introduced is a valid address.
            if(InetAddress.getByName(ntpServerAddress).isLoopbackAddress()){ // The IP addresse for a given host name.
              System.err.println(
                  "Error: The NTP server's address read is not a valid address (" +
                  InetAddress.getByName(ntpServerAddress) + ").");
              System.exit(1);
            }
          }catch (Exception e) {
            System.err.println(
                "Error: The NTP server's address read is not a valid address (" +
                ntpServerAddress + ").");
            System.err.println();
            System.exit(1);
          }

          updateInterval = Integer.parseInt(args[28].trim());
          if (updateInterval < 0) {
            System.err.println("Error: The value of updateInterval can not be a negative number (" +
                               updateInterval + ").");
            System.err.println();
            System.exit(1);
          }

        }

        // Define the object hiResTimer used to get the system time.
        HiResTimer hiResTimer = null;
        if(updateInterval == -1){  // Use default updateInterval defined on HiResTimer class.
          hiResTimer = new HiResTimer(ntpServerAddress);
        }else
          hiResTimer = new HiResTimer(ntpServerAddress,updateInterval);

        // Verify start time
        if ( (startTime.getTimeInMillis() + Math.round(timeOut * 1000)) <
            hiResTimer.milliTime()) {
          System.err.println("Error: The startTime is in the past, and separated from the present by more than Timeout.");
          System.err.println();
          System.exit(1);
        }


        // Process Session

        /* OWAMP_ControlClient(String serverAddress,String _confSender,
                               String _confReceiver, String _numberOfScheduleSlots, String _numberOfPackets,
                               String _senderPort, String _receiverPort, String _senderAddress,
                               String _receiverAddress, String _paddingLenght, long _startTime, String _timeOut,
                               String _typePDescriptor_Type, String _typePDescriptor_Value, String _slotType,
                               String _slotParameter , BufferedReader systemInReader, HiResTimer hiResTimer) */
        OWAMP_ControlClient session = new OWAMP_ControlClient(args[1], args[2],
            args[3], args[4], args[5], args[6], args[7], args[8], args[9],
            args[10], startTime.getTimeInMillis(), args[17],
            args[18], args[19], args[20], args[21], systemInReader, hiResTimer);

        double slotParameter = Double.parseDouble(args[21]);  // Slot parameter in seconds
        if(slotParameter < 0){
          System.err.println("Error: The value of slot parameter read is not valid.");
          System.exit(1);
        }

        // processSingleSession(String userID, String sqlServerAddress, String dataBase,
        //                      String userName, String passWord, String SSconfigTable, String _senderAddress,
        //                      String _receiverAddress, double slotParameter)
        if (!session.processSingleSession(args[0], args[22], args[23], args[24], args[25],
                                          args[26], args[8], args[9], slotParameter /*in seconds*/)) {
          System.err.println("Error: Error processing session.");
          System.exit(1);
        }

      }catch (NumberFormatException e) {
        System.err.println(e);
        System.err.println("Error: One argument value read is not valid.");
        System.err.println("Error: Error processing session.");
        System.exit(1);
      }

      System.exit(0);  // In this case we must terminate the execution at the end of the processing of
                       // the requested session.
    }


    // The user can execute this module with all the configurations parameters in the command line, to process
    // multisession requests to calculate confidence intervals.
    // In this case the user must instroduce 32 or 34 parameters. In this case he wants to save the results
    // to a data base. Parameters (String userID,String serverAddress, String _confSender,
    // String _confReceiver, String _numberOfScheduleSlots, String _numberOfPackets, String _senderPort,
    // String _receiverPort, String _senderAddress, String _receiverAddress, String _paddingLenght,
    // String StartYear, String StartMonth, String StartDate, String StartHour, String StartMinute, String StartSecond,
    // String _timeOut, String _typePDescriptor_Type, String _typePDescriptor_Value, String _slotType,
    // String _slotParameter, String sqlServerAddress, String dataBase, String userName, String passWord,
    // String CIconfigTable, String CIsidTable, String timeBetweenSetOfTests /*in minutes*/,
    // String timeBetweenTests /*in seconds*/, String numIntervals, String numOfTestsInInterval,
    // String ntpServerAddress, String updateInterval). The ntpServerAddress and updateInteval are optional.
    // If not defined the system time will be used to get the time reference. This mode of execution is also used
    // in the Web interface.

    if((args.length == 32)||(args.length == 34)){
      try{

        int year = Integer.parseInt(args[11].trim());
        if ( (year < 1970) || (year > 2036)) {
          System.err.println("Error: The value of year read is not valid.");
          System.err.println();
          System.exit(1);
        }

        int month = (Integer.parseInt(args[12].trim()) - 1); // 0 represents January and 11 December.
        if ( (month < 0) || (month > 11)) {
          System.err.println("Error: The value of month read is not valid.");
          System.err.println();
          System.exit(1);
        }

        int date = Integer.parseInt(args[13].trim());
        if ( (date < 1) || (date > 31)) {
          System.err.println("Error: The value of date read is not valid.");
          System.err.println();
          System.exit(1);
        }

        int hour = Integer.parseInt(args[14].trim());
        if ( (hour < 0) || (hour > 23)) {
          System.err.println("Error: The value of hour read is not valid.");
          System.err.println();
          System.exit(1);
        }

        int minute = Integer.parseInt(args[15].trim());
        if ( (minute < 0) || (minute > 59)) {
          System.err.println("Error: The value of minute read is not valid.");
          System.err.println();
          System.exit(1);
        }

        int second = Integer.parseInt(args[16].trim());
        if ( (second < 0) || (second > 59)) {
          System.err.println("Error: The value of second read is not valid.");
          System.err.println();
          System.exit(1);
        }

        Calendar startTime = Calendar.getInstance();
        // startTime.set(year, month, date, hour, minute, second)
        startTime.set(year, month, date, hour, minute, second);
        startTime.set(startTime.MILLISECOND, 0);

        double timeOut = Double.parseDouble(args[17]); // Timeout in seconds.
        if (timeOut < 0) {
          System.err.println("Error: The value of timeOut read is not valid.");
          System.exit(1);
        }


        // Verify if NTP server address was introduced.
        String  ntpServerAddress = null;  // The address of the NTP server to be used to get the time reference.
        int updateInterval = -1;          // The interval duration (in minutes) between two consecutives updates to time reference.

        if(args.length == 34){  // Read the NTP server's address

          ntpServerAddress = args[32].trim();
          try { // Test if the NTP server address introduced is a valid address.
            if(InetAddress.getByName(ntpServerAddress).isLoopbackAddress()){ // The IP addresse for a given host name.
              System.err.println(
                  "Error: The NTP server's address read is not a valid address (" +
                  InetAddress.getByName(ntpServerAddress) + ").");
              System.exit(1);
            }
          }catch (Exception e) {
            System.err.println(
                "Error: The NTP server's address read is not a valid address (" +
                ntpServerAddress + ").");
            System.err.println();
            System.exit(1);
          }

          updateInterval = Integer.parseInt(args[33].trim());
          if (updateInterval < 0) {
            System.err.println("Error: The value of updateInterval can not be a negative number (" +
                               updateInterval + ").");
            System.err.println();
            System.exit(1);
          }

        }

        // Define the object hiResTimer used to get the system time.
        HiResTimer hiResTimer = null;
        if(updateInterval == -1){  // Use default updateInterval defined on HiResTimer class.
          hiResTimer = new HiResTimer(ntpServerAddress);
        }else
          hiResTimer = new HiResTimer(ntpServerAddress,updateInterval);

        // Verify start time.
        if ( (startTime.getTimeInMillis() + Math.round(timeOut * 1000)) <
            hiResTimer.milliTime()) {
          System.err.println("Error: The startTime is in the past, and separated from the present by more than Timeout.");
          System.err.println();
          System.exit(1);
        }


        // Process Session

        /* OWAMP_ControlClient(String serverAddress,String _confSender,
                               String _confReceiver, String _numberOfScheduleSlots, String _numberOfPackets,
                               String _senderPort, String _receiverPort, String _senderAddress,
                               String _receiverAddress, String _paddingLenght, long _startTime, String _timeOut,
                               String _typePDescriptor_Type, String _typePDescriptor_Value, String _slotType,
                               String _slotParameter , BufferedReader systemInReader, HiResTimer hiResTimer) */
        OWAMP_ControlClient session = new OWAMP_ControlClient(args[1], args[2],
            args[3], args[4], args[5], args[6], args[7], args[8], args[9],
            args[10], startTime.getTimeInMillis(), args[17],
            args[18], args[19], args[20], args[21], systemInReader, hiResTimer);

        double slotParameter = Double.parseDouble(args[21]);  // Slot parameter in seconds
        if(slotParameter < 0){
          System.err.println("Error: The value of slot parameter read is not valid.");
          System.exit(1);
        }

        // processCIsession(String userID, String sqlServerAddress, String dataBase, String userName,
        //                  String passWord, String CIconfigTable, String CIsidTable,String _senderAddress,
        //                  String _receiverAddress, int timeBetweenSetOfTests /*in milliseconds*/, int timeBetweenTests /*in milliseconds*/,
        //                  int numIntervals, int numOfTestsInInterval, double slotParameter /*in seconds*/,
        //                  String typePDescriptor_Type, String typePDescriptor_Value)

        if (!session.processCIsession(args[0], args[22], args[23], args[24], args[25], args[26],
                                      args[27], args[8], args[9],
                                     (int)(Double.parseDouble(args[28].trim())*60*1000) /*in milliseconds*/,
                                     (int)(Double.parseDouble(args[29].trim())*1000) /*in milliseconds*/,
                                     Integer.parseInt(args[30].trim()),
                                     Integer.parseInt(args[31].trim()),
                                     slotParameter /*in seconds*/, args[18], args[19])) {
          System.err.println(
             "Error: Error processing the Confidence Interval session.");
          System.exit(1);
        }

      }catch (NumberFormatException e) {
        System.err.println(e);
        System.err.println("Error: One argument value read is not valid.");
        System.err.println("Error: Error processing session.");
        System.exit(1);
      }

      System.exit(0);  // In this case we must terminate the execution at the end of the processing of
                       // the requested session.
    }


    // If the number of parameters introduced is not equal to 27, 29, 32 or 34 we must use normal operation of Control-Client.

    String line = "";     // Used to receive the string read from keybord.
    boolean testError = false;     // Used to verify if an error occurred
    InetAddress serverAddress = null;
    String address = null;
    String addressAux = null;
    int serverTCPportNumber = DEFAULT_SEVER_PORT;  // TCPportNumber to use in OWAMP-Control with Server.
    boolean IPv6 = false;
    int i=0;

    do{
      testError = false;
      /*
      * The server address should be in the following format:
      *
      *     IPv4_address:TCPportNumber    // for IPv4 address
      *     [IPv6_address]:TCPportNumber  // for IPv6 address
      *
      * The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the
      * Control-Client and the respective OWAMP-Server. If the user wants to use the default TCPportNumber (22368)
      * the server address can be in the following format:
      *
      *     IPv4_address    // for IPv4 address
      *     [IPv6_address]  // for IPv6 address
      */
      System.out.print("Introduce the address of the Server: ");
      try{
        address = systemInReader.readLine().trim();
      }                                        // if not, an UnknownHostException is thrown.
      catch (IOException e) {
        System.err.println(e);
        System.err.println("Error: You must introduce a valid IP address.");
        testError = true;
      }

      if (!testError) {
        addressAux = address;
        serverTCPportNumber = DEFAULT_SEVER_PORT; // TCPportNumber to use in OWAMP-Control with Server.
        IPv6 = false;

        try {

          if (address.indexOf("[") != -1) { // IPv6 address
            IPv6 = true;
            if (address.indexOf("]:") != -1) { // TCPportNumber to use in OWAMP-Control with sender specified.
              serverTCPportNumber = Integer.parseInt(address.substring(address.
                  indexOf("]:") + 2));
            }
            address = address.substring(address.indexOf("[") + 1,
                                        address.indexOf("]"));
          }
          else { // IPv4 address
            if (address.indexOf(":") != -1) { // TCPportNumber to use in OWAMP-Control with sender specified.
              serverTCPportNumber = Integer.parseInt(address.substring(address.
                  indexOf(":") + 1));
              address = address.substring(0, address.indexOf(":"));
            }
          }

          if (serverTCPportNumber < 0 || serverTCPportNumber >= 65536) {
            System.err.println("Error: Sender TCP port number out of range (" +
                               serverTCPportNumber + ").");
            System.err.println();
            throw new Exception();
          }

          InetAddress inetAddress[] = InetAddress.getAllByName(address); // The array of all the IP addresses for a given host name.

          for (i = 0; i < inetAddress.length; i++) {
            if (IPv6 && (inetAddress[i].getAddress().length == 16)) { // IPv6 address
              serverAddress = inetAddress[i];
              break;
            }

            if (!IPv6 && (inetAddress[i].getAddress().length == 4)) { // IPv4 address
              serverAddress = inetAddress[i];
              break;
            }

            if (i == (inetAddress.length - 1)) { // last cicle.
              System.err.println("Error: There isn't a valid IPv" + (IPv6 ? 6 : 4) +
                                 " address for the server address introduced (" +
                                 address + ").");
              System.err.println();
              testError = true;
            }
          }

        }
        catch (NumberFormatException e) {
          System.err.println(
              "Error: The the server address read is not a valid address (" +
              addressAux + ").");
          System.err.println(
              "The server address should be in the following format:\n\n" +
              "    IPv4_address:TCPportNumber    // for IPv4 address\n" +
              "    [IPv6_address]:TCPportNumber  // for IPv6 address\n\n" +
              "The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the \n" +
              "Control-Client and the respective OWAMP-Server. If the user wants to use the default TCPportNumber (22368) \n" +
              "the server address can be in the following format:\n\n" +
              "    IPv4_address    // for IPv4 address\n" +
              "    [IPv6_address]  // for IPv6 address\n");
          System.err.println();
          testError = true;
        }
        catch (Exception e2) {
          System.err.println(
              "Error: The the server address read is not a valid address (" +
              addressAux + ").");
          System.err.println(
              "The server address should be in the following format:\n\n" +
              "    IPv4_address:TCPportNumber    // for IPv4 address\n" +
              "    [IPv6_address]:TCPportNumber  // for IPv6 address\n\n" +
              "The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the \n" +
              "Control-Client and the respective OWAMP-Server. If the user wants to use the default TCPportNumber (22368) \n" +
              "the server address can be in the following format:\n\n" +
              "    IPv4_address    // for IPv4 address\n" +
              "    [IPv6_address]  // for IPv6 address\n");
          System.err.println();
          testError = true;
        }

      }

    }while(testError);
    System.out.println();


    // Read the NTP server's address
    String  ntpServerAddress = null;  // The address of the NTP server to be used to get the time reference.
    int updateInterval = -1;          // The interval duration (in minutes) between two consecutives updates to time reference.

    try {
      System.out.print(
          "Do you want to get the time reference directly from a NTP server? (Default NO) ");
      line = systemInReader.readLine().trim();
      if ((line.compareToIgnoreCase("yes") == 0)||(line.compareToIgnoreCase("y") == 0)) {

        do {
          System.out.print("Introduce the address of the NTP server: ");
          ntpServerAddress = systemInReader.readLine().trim();
          testError = false;
          try { // Test if the NTP server address introduced is a valid address.
            if(InetAddress.getByName(ntpServerAddress).isLoopbackAddress()){ // The IP addresse for a given host name.
              System.err.println(
                  "Error: The NTP server's address read is not a valid address (" +
                  InetAddress.getByName(ntpServerAddress) + ").\n");
              testError = true;
            }
          }
          catch (Exception e) {
            testError = true;
            System.err.println(
                "Error: The NTP server's address read is not a valid address (" +
                ntpServerAddress + ").");
            System.err.println();
          }
        }while (testError);

      }
      System.out.println();

      do {
        System.out.print("Introduce the value (in minutes) to be used for the interval time between two consecutives updates to the time reference (Default 2): ");
        try{
          testError = false;
          line = systemInReader.readLine().trim();
          updateInterval = Integer.parseInt(line);
        }catch (NumberFormatException e2) {
          if (line.length() == 0) // 0 is the default option  - Default 2
            updateInterval = 2;   // 2 minutes
          else{
            testError = true;
            System.err.println("Error: " + e2 + ".");
            System.err.println("Error: The value of the interval time between two consecutives updates to the time reference read is not a valid.");
            System.err.println();
          }
        }

        if ((updateInterval < 0)&&(!testError)) {
          testError = true;
          System.err.println(
              "Error: The value of updateInterval can not be a negative number (" +
              updateInterval + ").");
          System.err.println();
        }
      }while (testError);
      System.out.println();

    }catch (IOException e) {
      System.err.println("Error: " + e + ".");
      System.exit(1);
    }

    // Define the object hiResTimer used to get the system time.
    HiResTimer hiResTimer = new HiResTimer(ntpServerAddress,updateInterval);

    OWAMP_ControlClient session = new OWAMP_ControlClient(serverAddress,serverTCPportNumber,systemInReader,hiResTimer); // OWAMP_ControlClient(InetAddress serverAddress, int serverTCPportNumber, BufferedReader r, HiResTimer hiResTimer)

    int mode = 1;
    RequestSession dataRS = new RequestSession();
    ScheduleSlotDescriptions[] arrayOfdataSSD = null;
    boolean test = false;
    int senderTCPportNumber = 0;
    int numberOfPackets = 0;
    int paddingLenght = 0;
    double mean = 0;
    int year = 0, aux = 0;
    int month = 0;
    int date = 0;
    int hour = 0;
    int minute = 0;
    int second = 0;
    double timeout = 0;
    byte typePDescriptor = 0;
    int typePvalue = 0;
    int numberOfScheduleSlots = 0;
    int scheduleSlotType = 0;
    int interval = 0;
    int numTest = 0;
    int testInterval = 0;
    long testStartTime = 0;
    long startTime =  0;
    long numIntervals = 0;
    boolean closeConnection = false;
    double totalTimeToWait = 0;
    int j = 0;
    boolean defineSenderPort = false;


    do{
      session.showMenu();

      if(session.option == 6)
        System.exit(1);

      switch(session.option){
        case 0: // process connectionSetup(int modes)

          // Test socket connection

          if(session.isSocketConnected(session.socketToServer)){
            System.err.println("Error: The client is already connected to this server. You can not process this operation again.");
            break;
          }

          // We should set the socket to server's timeout to 2 * TIMEOUT because the server waits
          // for TIMEOUT ms to receive any packet from Receiver, and then it sends it to Control-Client.
                                     // connectionSetup(int mode, InetAddress ia, int port, int timeout)
          session.socketToServer = session.connectionSetupRequest(mode,serverAddress,serverTCPportNumber,2*TIMEOUT);

          if(session.socketToServer == null){
            System.err.println("Error: The Server specified doesn't exist.\n");
            System.err.println("Error: Error in processing Connection Setup.");
            System.exit(1);
          }

          // Reset the number of request sessions and the arrays used to save the sessions configurations.
          session.numRS = 0;    // Number of Request Session made - there are no Request Session yet.
          session.mode = mode;  // The mode value to be used in the current OWAMP control and test sessions.
          session.arrayOfSocketToSender = new Socket[1];
          session.arrayOfSenderTCPportNumber = new int[1];
          session.arrayOfdataRS = new RequestSessionToBeSend[1];
          session.arrayOfdataSSD = new ScheduleSlotDescriptionsToBeSend[1][];

          break;
        case 1: // process requestSession

          // Test socket connection

          if((!session.socketToServer.isConnected())||(session.socketToServer.isClosed())){
            System.err.println("Error: You should connect to a server first, before trying this operation.");
            break;
          }

          try{

            testError = false;     // Used to verify if an error occurred
            do{
              System.out.println();
              System.out.println("Process Request Session:");
              System.out.println();
              System.out.println("Choose one of the options (Default 0): ");
              System.out.println("0 - Read configurations from the config file");
              System.out.println("1 - Read configurations from keyboard");
              System.out.println("2 - Choose a time interval and read the initial configurations from the config file to calculate confidence intervals");
              try {
                line = systemInReader.readLine().trim();
                session.option = Integer.parseInt(line);
              }catch (NumberFormatException e) {
                if (line.length() == 0) // 0 is the default option  - Default 0
                  session.option = 0;
                else {
                  System.err.println(e);
                  session.option = -1;
                }
              }

              if((session.option == -1)||(session.option < 0)||(session.option > 2))
                System.err.println("The option introduced is not valid.");
              if(session.option == 2 && (session.numRS != 0)){
                System.err.println("You can not choose this option when other RS where made previously.");
                testError = true;
              }
            }while((session.option == -1)||(session.option < 0)||(session.option > 2));

            if(testError)
              break;

            testError = false;     // Used to verify if an error occurred

            switch(session.option){
              case 0:
                if(!session.readConfigFile(session.numRS)){
                  System.err.println("Error: Error reading from configuration file.");
                  testError = true;
                }
                break;
              case 1:
              {

                dataRS.confSender = 0;
                dataRS.confReceiver = 1;

                numberOfPackets = 0;
                do{
                  System.out.print("Introduce the number of test packet to be sent (Default 100): ");
                  try {
                    line = systemInReader.readLine().trim();
                    numberOfPackets = Integer.parseInt(line);
                  }catch (NumberFormatException e) {
                    if (line.length() == 0) // 0 is the default option  - Default 100
                      numberOfPackets = 0;
                    else {
                      System.err.println(e);
                      numberOfPackets = -1;
                    }
                  }
                  if(numberOfPackets == -1)
                    System.err.println("The number of test packet introduced is not valid.");
                }while(numberOfPackets == -1);
                dataRS.numberOfPackets=(numberOfPackets == 0)?100:numberOfPackets;

                dataRS.senderPort = 0;
                dataRS.receiverPort = 0;

                test = false;
                senderTCPportNumber = 0;
                dataRS.senderAddress = InetAddress.getByName("127.0.0.1").getAddress();

                do{
                  test = true;

                  // sender address

                  /* The sender address should be in the following format:

                        IPv4_address:TCPportNumber    // for IPv4 address
                        [IPv6_address]:TCPportNumber  // for IPv6 address

                     The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the
                     Control-Client and the respective OWAMP-Sender. If the user wants to use the default TCPportNumber
                     the sender address can be in the following format:

                        IPv4_address    // for IPv4 address
                        [IPv6_address]  // for IPv6 address
                  */

                  if(InetAddress.getByAddress(dataRS.senderAddress).isLoopbackAddress()){ // If sender address is ok and receiver not, keep it

                    System.out.print("Introduce the address of the Sender: ");
                    address = systemInReader.readLine().trim();
                    addressAux = address;

                    senderTCPportNumber = DEFAULT_SENDER_PORT; // TCPportNumber to use in OWAMP-Control with sender.
                    IPv6 = false;

                    try {

                      if (address.indexOf("[") != -1) { // IPv6 address
                        IPv6 = true;
                        if (address.indexOf("]:") != -1) { // TCPportNumber to use in OWAMP-Control with sender specified.
                          senderTCPportNumber = Integer.parseInt(address.
                              substring(address.indexOf("]:") + 2));
                        }
                        address = address.substring(address.indexOf("[") + 1,
                            address.indexOf("]"));
                      }
                      else { // IPv4 address
                        if (address.indexOf(":") != -1) { // TCPportNumber to use in OWAMP-Control with sender specified.
                          senderTCPportNumber = Integer.parseInt(address.
                              substring(address.indexOf(":") + 1));
                          address = address.substring(0, address.indexOf(":"));
                        }
                      }

                      if (senderTCPportNumber < 0 || senderTCPportNumber >= 65536) {
                        System.err.println(
                            "Error: Sender TCP port number out of range (" +
                            senderTCPportNumber + ").");
                        throw new Exception();
                      }

                      InetAddress inetAddress[] = InetAddress.getAllByName(
                          address); // The array of all the IP addresses for a given host name.

                      for (i = 0; i < inetAddress.length; i++) {
                        if (IPv6 && (inetAddress[i].getAddress().length == 16)) { // IPv6 address
                          dataRS.senderAddress = inetAddress[i].getAddress();
                          break;
                        }

                        if (!IPv6 && (inetAddress[i].getAddress().length == 4)) { // IPv4 address
                          dataRS.senderAddress = inetAddress[i].getAddress();
                          break;
                        }

                        if (i == (inetAddress.length - 1)) { // last cicle.
                          System.err.println("Error: There isn't a valid IPv" +
                                             (IPv6 ? 6 : 4) +
                              " address for the sender address introduced (" +
                                             address + ").");
                          System.err.println();
                          test = false;
                        }
                      }

                      // The sender and receiver addresses can only be both loopback addresses if the Control-Client
                      // is on the same machine were the Server is running.
                      if(test && (InetAddress.getByAddress(dataRS.senderAddress).isLoopbackAddress()) && !session.isAnyLocalAddress(session.serverAddress)){
                        System.err.println("Error: The sender address can only be loopback address if the Control-Client is on the same machine were the Server is running.");
                        System.err.println();
                        test = false;
                      }

                    }
                    catch (NumberFormatException e) {
                      System.err.println(
                          "Error: The the sender address read is not a valid address (" +
                          addressAux + ").");
                      System.err.println(
                          "The sender address should be in the following format:\n\n" +
                          "    IPv4_address:TCPportNumber    // for IPv4 address\n" +
                          "    [IPv6_address]:TCPportNumber  // for IPv6 address\n\n" +
                          "The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the \n" +
                          "Control-Client and the respective OWAMP-Sender. If the user wants to use the default TCPportNumber \n" +
                          "the sender address can be in the following format:\n\n" +
                          "    IPv4_address    // for IPv4 address\n" +
                          "    [IPv6_address]  // for IPv6 address\n");
                      System.err.println();
                      test = false;
                    }
                    catch (Exception e2) {
                      System.err.println(
                          "Error: The the sender address read is not a valid address (" +
                          addressAux + ").");
                      System.err.println(
                          "The sender address should be in the following format:\n\n" +
                          "    IPv4_address:TCPportNumber    // for IPv4 address\n" +
                          "    [IPv6_address]:TCPportNumber  // for IPv6 address\n\n" +
                          "The TCPportNumber is the TCP port to be used in OWAMP-Control communication between the \n" +
                          "Control-Client and the respective OWAMP-Sender. If the user wants to use the default TCPportNumber \n" +
                          "the sender address can be in the following format:\n\n" +
                          "    IPv4_address    // for IPv4 address\n" +
                          "    [IPv6_address]  // for IPv6 address\n");
                      System.err.println();
                      test = false;
                    }

                  }

                  if(test){
                    // receiver address

                    /* The receiver address should be in the following format:
                           IPv4_address     // for IPv4 address
                           [IPv6_address]   // for IPv6 address

                       The TCP port to be used in OWAMP-Control communication between the Server and receiver should be defined
                       when running OWAMP_Server, or the default TCP port will be used.
                     */

                    System.out.print("Introduce the address of the Receiver: ");
                    address = systemInReader.readLine().trim();
                    addressAux = address;

                    IPv6 = false;

                    try {

                      if (address.indexOf("[") != -1) { // IPv6 address
                        IPv6 = true;
                        dataRS.mbz_iPVN = (byte) 6;
                        if (address.indexOf("]:") != -1) { // TCPportNumber to use in OWAMP-Control with receiver specified.
                          System.err.println("Error: The TCP port to be used in OWAMP-Control communication between the Server and receiver should be defined when running OWAMP_Server.");
                          System.err.println();
                          throw new Exception();
                        }else
                          address = address.substring(address.indexOf("[") + 1, address.indexOf("]"));
                      }
                      else { // IPv4 address
                        dataRS.mbz_iPVN = (byte) 4;
                        if (address.indexOf(":") != -1) { // TCPportNumber to use in OWAMP-Control with receiver specified.
                          System.err.println("Error: The TCP port to be used in OWAMP-Control communication between the Server and receiver should be defined when running OWAMP_Server.");
                          System.err.println();
                          throw new Exception();
                        }
                      }

                      InetAddress inetAddress[] = InetAddress.getAllByName(address); // The array of all the IP addresses for a given host name.

                      for (i = 0; i < inetAddress.length; i++) {
                        if (IPv6 && (inetAddress[i].getAddress().length == 16)) { // IPv6 address
                          dataRS.receiverAddress = inetAddress[i].getAddress();
                          break;
                        }

                        if (!IPv6 && (inetAddress[i].getAddress().length == 4)) { // IPv4 address
                          dataRS.receiverAddress = inetAddress[i].getAddress();
                          break;
                        }

                        if (i == (inetAddress.length - 1)) { // last cicle.
                          System.err.println("Error: There isn't a valid IPv" + (IPv6 ? 6 : 4) +
                                             " address for the receiver address introduced (" +
                                             address + ").");
                          System.err.println();
                          test = false;
                        }
                      }

                      // If sender address is loopback address receiver should be loopback address too. If receiver address is loopback address sender should be loopback address too.
                      if(test&&((InetAddress.getByAddress(dataRS.senderAddress).isLoopbackAddress()&&!InetAddress.getByAddress(dataRS.receiverAddress).isLoopbackAddress())||(!InetAddress.getByAddress(dataRS.senderAddress).isLoopbackAddress()&&InetAddress.getByAddress(dataRS.receiverAddress).isLoopbackAddress()))){
                        if(InetAddress.getByAddress(dataRS.senderAddress).isLoopbackAddress())
                          System.err.println("Error: The sender address is loopback address and receiver address isn't.");
                        if(InetAddress.getByAddress(dataRS.receiverAddress).isLoopbackAddress())
                          System.err.println("Error: The receiver address is loopback address and sender address isn't.");
                        System.err.println();
                        test = false;
                        dataRS.senderAddress = InetAddress.getByName("127.0.0.1").getAddress(); // To force the system to ask the sender address again.
                      }

                      if (test && (dataRS.senderAddress.length != dataRS.receiverAddress.length)) { // The sender and receiver address should have the same IP version.
                        System.err.println(
                            "Error: The sender and receiver addresses should have the same IP version.");
                        System.err.println();
                        test = false;
                        dataRS.senderAddress = InetAddress.getByName("127.0.0.1").getAddress(); // To force the system to ask the sender address again.
                      }

                    }
                    catch (NumberFormatException e) {
                      System.err.println(
                          "Error: The the receiver address read is not a valid address (" +
                          addressAux + ").");
                      System.err.println(
                          "The receiver address should be in the following format:\n\n" +
                          "    IPv4_address    // for IPv4 address\n" +
                          "    [IPv6_address]  // for IPv6 address\n");
                      System.err.println();
                      test = false;
                    }
                    catch (Exception e2) {
                      System.err.println(
                          "Error: The the receiver address read is not a valid address (" +
                          addressAux + ").");
                      System.err.println(
                          "The receiver address should be in the following format:\n\n" +
                          "    IPv4_address    // for IPv4 address\n" +
                          "    [IPv6_address]  // for IPv6 address\n");
                      System.err.println();
                      test = false;
                    }
                  }

                }while(!test);

                paddingLenght = 0;
                do{
                  System.out.print("Introduce the padding length of the test packets to be sent (Default 0): ");
                  try {
                    line = systemInReader.readLine().trim();
                    paddingLenght = Integer.parseInt(line);
                  }catch (NumberFormatException e) {
                    if (line.length() == 0) // 0 is the default option  - Default 0
                      paddingLenght = 0;
                    else {
                      System.err.println(e);
                      paddingLenght = -1;
                    }
                  }

                  if(paddingLenght == -1)
                    System.err.println("The padding length of test packet introduced is not valid.");
                }while(paddingLenght == -1);
                dataRS.paddingLenght=(paddingLenght == 0)?0:paddingLenght;

                // Ask for the StartTime of this Session. Note that session with StartTime in the past are
                // rejected by the Session-Sender.

                Calendar startTimeAux = Calendar.getInstance();
                startTimeAux.setTimeInMillis((long)hiResTimer.milliTime());
                System.out.println("\nIntroduce the start time of this session");
                year = 0;
                aux = 0;
                do {
                  aux = startTimeAux.get(startTimeAux.YEAR);
                  System.out.print("Introduce the year (Default "+aux+"): ");
                  try {
                    line = systemInReader.readLine().trim();
                    year = Integer.parseInt(line);
                  }catch(NumberFormatException e){
                    if(line.length() == 0)
                      year = aux;
                    else{
                      System.err.println(e);
                      year = -1;
                    }
                  }
                  if ( (year < 1970) || (year > 2036))
                    System.err.println(
                        "Error: The value of year read is not valid.");
                }while ( (year < 1970) || (year > 2036));

                month = 0;
                do {
                  aux = startTimeAux.get(startTimeAux.MONTH);
                  System.out.print("Introduce the month (Default "+(aux+1)+"): ");
                  try {
                    line = systemInReader.readLine().trim();
                    month = (Integer.parseInt(line)-1);  // 0 represents January and 11 December.
                  }catch(NumberFormatException e){
                    if(line.length() == 0)
                      month = aux;
                    else{
                      System.err.println(e);
                      month = -1;
                    }
                  }
                  if ( (month < 0) || (month > 11))
                    System.err.println(
                        "Error: The value of month read is not valid.");
                }while ( (month < 0) || (month > 11));

                date = 0;
                do {
                  aux = startTimeAux.get(startTimeAux.DATE);
                  System.out.print("Introduce the date (Default "+aux+"): ");
                  try {
                    line = systemInReader.readLine().trim();
                    date = Integer.parseInt(line);
                  }catch(NumberFormatException e){
                    if(line.length() == 0)
                      date = aux;
                    else{
                      System.err.println(e);
                      date = -1;
                    }
                  }
                  if ( (date < 1) || (date > 31))
                    System.err.println(
                        "Error: The value of date read is not valid.");
                }while ( (date < 1) || (date > 31));

                hour = 0;
                do {
                  aux = startTimeAux.get(startTimeAux.HOUR_OF_DAY);
                  System.out.print("Introduce the hour (Default "+aux+"): ");
                  try {
                    line = systemInReader.readLine().trim();
                    hour = Integer.parseInt(line);
                  }catch(NumberFormatException e){
                    if(line.length() == 0)
                      hour = aux;
                    else{
                      System.err.println(e);
                      hour = -1;
                    }
                  }
                  if ( (hour < 0) || (hour > 23))
                    System.err.println(
                        "Error: The value of hour read is not valid.");
                }while ( (hour < 0) || (hour > 23));

                minute = 0;
                do {
                  aux = startTimeAux.get(startTimeAux.MINUTE);
                  System.out.print("Introduce the minutes (Default "+aux+"): ");
                  try {
                    line = systemInReader.readLine().trim();
                    minute = Integer.parseInt(line);
                  }catch(NumberFormatException e){
                    if(line.length() == 0)
                      minute = aux;
                    else{
                      System.err.println(e);
                      minute = -1;
                    }
                  }
                  if ( (minute < 0) || (minute > 59))
                    System.err.println(
                        "Error: The value of minutes read is not valid.");
                }while ( (minute < 0) || (minute > 59));

                second = 0;
                do {
                  aux = 0;
                  System.out.print("Introduce the seconds (Default "+aux+"): ");
                  try {
                    line = systemInReader.readLine().trim();
                    second = Integer.parseInt(line);
                  }catch(NumberFormatException e){
                    if(line.length() == 0)
                      second = aux;
                    else{
                      System.err.println(e);
                      second = -1;
                    }
                  }
                  if ( (second < 0) || (second > 59))
                    System.err.println(
                        "Error: The value of second read is not valid.");
                }while ( (second < 0) || (second > 59));
                System.out.println();

                startTimeAux.set(year, month, date, hour, minute, second);
                startTimeAux.set(startTimeAux.MILLISECOND, 0);

                dataRS.startTime= new Timestamp(startTimeAux.getTimeInMillis(), (short)1); // Time in milliseconds since 1Jan1970

                timeout = 0;  // in seconds
                do{
                  System.out.print("Introduce the timeout value to use for test packets (Default 10): ");
                  try {
                    line = systemInReader.readLine().trim();
                    timeout = Double.parseDouble(line);
                  }catch (NumberFormatException e) {
                    if (line.length() == 0) // 0 is the default option  - Default 10
                      timeout = 10;
                    else {
                      System.err.println(e);
                      timeout = -1;
                    }
                  }

                  if(timeout < 0)
                    System.err.println("The timeout value to use for test packets introduced is not valid.");
                }while(timeout < 0);
                dataRS.timeout = new Timestamp(timeout, (short)0);    // Timeout in seconds since 1Jan1900 = timeout seconds * 1000

                typePDescriptor = 0;
                typePvalue = 0;
                dataRS.typeP_Descriptor = session.generateTypeP_Descriptor(typePDescriptor,typePvalue);

                numberOfScheduleSlots = 0;
                do{
                  System.out.print("Introduce the number of packets Schedule slot descriptions (Default 1): ");
                  try {
                    line = systemInReader.readLine().trim();
                    numberOfScheduleSlots = Integer.parseInt(line);
                  }catch (NumberFormatException e) {
                    if (line.length() == 0) // 0 is the default option  - Default 1
                      numberOfScheduleSlots = 1;
                    else {
                      System.err.println(e);
                      numberOfScheduleSlots = -1;
                    }
                  }

                  if(numberOfScheduleSlots == -1)
                    System.err.println("The number of packets Schedule slot descriptions introduced is not valid.");
                }while(numberOfScheduleSlots == -1);
                dataRS.numberOfScheduleSlots = numberOfScheduleSlots;

                arrayOfdataSSD = new ScheduleSlotDescriptions[dataRS.numberOfScheduleSlots];

                scheduleSlotType = 0;
                do{
                  System.out.println("Choose which type of schedule do you want (Default 0):");
                  System.out.println("0 - All Exponentially distributed pseudo-random quantity");
                  System.out.println("1 - All Fixed quantity");
                  System.out.println("2 - Mixed type");
                  try {
                    line = systemInReader.readLine().trim();
                    scheduleSlotType = Integer.parseInt(line);
                  }catch (NumberFormatException e) {
                    if (line.length() == 0) // 0 is the default option  - Default 0
                      scheduleSlotType = 0;
                    else {
                      System.err.println(e);
                      scheduleSlotType = -1;
                    }
                  }

                  if((scheduleSlotType == -1)||(scheduleSlotType < 0)||(scheduleSlotType>2))
                    System.err.println("The schedule slot type introduced is not valid.");
                }while((scheduleSlotType == -1)||(scheduleSlotType < 0)||(scheduleSlotType>2));

                if(scheduleSlotType != 2){

                  mean = 0;

                  switch(scheduleSlotType){
                    case 0:
                      do{
                        System.out.print("Introduce the mean value of the Exponentially distributed pseudo-random quantity in seconds (Default 1): ");
                        try {
                          line = systemInReader.readLine().trim();
                          mean = Double.parseDouble(line);
                        }catch (NumberFormatException e) {
                          if (line.length() == 0) // 0 is the default option  - Default 1
                            mean = 1;
                          else {
                            System.err.println(e);
                            mean = -1;
                          }
                        }
                        if(mean < 0)
                          System.err.println("The mean value of the Exponentially distributed pseudo-random quantity introduced is not valid.");
                      }while(mean < 0);

                      for(i=0;i<dataRS.numberOfScheduleSlots;i++)
                        arrayOfdataSSD[i] = new ScheduleSlotDescriptions((byte)scheduleSlotType, session.generateSlotParameter(mean));

                      break;
                    case 1:
                      for(i=0;i<dataRS.numberOfScheduleSlots;i++){
                        do{
                          System.out.print("Introduce the fixed quantity value in seconds (Default 1): ");
                          try {
                            line = systemInReader.readLine().trim();
                            mean = Double.parseDouble(line);
                          }catch (NumberFormatException e) {
                            if (line.length() == 0) // 0 is the default option  - Default 1
                              mean = 1;
                            else {
                              System.err.println(e);
                              mean = -1;
                            }
                          }

                          if(mean < 0)
                            System.err.println("The fixed quantity value introduced is not valid.");
                        }while(mean < 0);

                        arrayOfdataSSD[i] = new ScheduleSlotDescriptions((byte)scheduleSlotType,  new Timestamp(mean, (short)0));
                      }
                  }

                }else{

                  for(i=0;i<dataRS.numberOfScheduleSlots;i++){

                    do{
                      System.out.println("Choose which type of schedule do you want (Default 0):");
                      System.out.println("0 - Exponentially distributed pseudo-random quantity");
                      System.out.println("1 - Fixed quantity");
                      try {
                        line = systemInReader.readLine().trim();
                        scheduleSlotType = Integer.parseInt(line);
                      }catch (NumberFormatException e) {
                        if (line.length() == 0) // 0 is the default option  - Default 0
                          scheduleSlotType = 0;
                        else {
                          System.err.println(e);
                          scheduleSlotType = -1;
                        }
                      }

                      if((scheduleSlotType == -1)||(scheduleSlotType < 0)||(scheduleSlotType>1))
                        System.err.println("The schedule slot type introduced is not valid.");
                    }while((scheduleSlotType == -1)||(scheduleSlotType < 0)||(scheduleSlotType>1));

                    mean = 0;

                    switch(scheduleSlotType){
                      case 0:
                        do{
                          System.out.print("Introduce the mean value of the Exponentially distributed pseudo-random quantity in seconds (Default 1): ");
                          try {
                            line = systemInReader.readLine().trim();
                            mean = Double.parseDouble(line);
                          }catch (NumberFormatException e) {
                            if (line.length() == 0) // 0 is the default option  - Default 1
                              mean = 1;
                            else {
                              System.err.println(e);
                              mean = -1;
                            }
                          }

                          if(mean < 0)
                            System.err.println("The mean value of the Exponentially distributed pseudo-random quantity introduced is not valid.");
                        }while(mean < 0);

                        arrayOfdataSSD[i] = new ScheduleSlotDescriptions((byte)scheduleSlotType, session.generateSlotParameter(mean));

                        break;
                      case 1:
                        do{
                          System.out.print("Introduce the fixed quantity value in seconds (Default 1): ");
                          try {
                            line = systemInReader.readLine().trim();
                            mean = Double.parseDouble(line);
                          }catch (NumberFormatException e) {
                            if (line.length() == 0) // 0 is the default option  - Default 1
                              mean = 1;
                            else {
                              System.err.println(e);
                              mean = -1;
                            }
                          }

                          if(mean < 0)
                            System.err.println("The fixed quantity value introduced is not valid.");
                        }while(mean < 0);

                        arrayOfdataSSD[i] = new ScheduleSlotDescriptions((byte)scheduleSlotType, new Timestamp(mean, (short)0));
                    }
                  }

                }

                // Initialize the Request Session Packet to be sent to the Server.

                session.arrayOfdataRS[session.numRS] = new RequestSessionToBeSend(dataRS);
                session.arrayOfSenderTCPportNumber[session.numRS] = senderTCPportNumber;

                // Initialize the list of the ScheduleSlotDescriptions packets to be sent.

                session.arrayOfdataSSD[session.numRS] = new ScheduleSlotDescriptionsToBeSend[dataRS.numberOfScheduleSlots];
                for(i=0;i<dataRS.numberOfScheduleSlots;i++)
                  session.arrayOfdataSSD[session.numRS][i] = new ScheduleSlotDescriptionsToBeSend(arrayOfdataSSD[i]);

                break;
              }
              case 2:
              {
                System.out.println("\nIntroduce the ending date and time of these Sessions: ");
                Calendar stopTime = Calendar.getInstance();
                stopTime.setTimeInMillis((long)hiResTimer.milliTime());
                year = 0;
                aux = 0;
                do {
                  aux = stopTime.get(stopTime.YEAR);
                  System.out.print("\nIntroduce the year (Default "+aux+"): ");
                  try {
                    line = systemInReader.readLine().trim();
                    year = Integer.parseInt(line);
                  }catch(NumberFormatException e){
                    if(line.length() == 0)
                      year = aux;
                    else{
                      System.err.println(e);
                      year = -1;
                    }
                  }
                  if ( (year < 1970) || (year > 2036))
                    System.err.println(
                        "Error: The value of year read is not valid.");
                }while ( (year < 1970) || (year > 2036));

                month = 0;
                do {
                  aux = stopTime.get(stopTime.MONTH);
                  System.out.print("Introduce the month (Default "+(aux+1)+"): ");
                  try {
                    line = systemInReader.readLine().trim();
                    month = (Integer.parseInt(line)-1);  // 0 represents January and 11 December.
                  }catch(NumberFormatException e){
                    if(line.length() == 0)
                      month = aux;
                    else{
                      System.err.println(e);
                      month = -1;
                    }
                  }
                  if ( (month < 0) || (month > 11))
                    System.err.println(
                        "Error: The value of month read is not valid.");
                }while ( (month < 0) || (month > 11));

                date = 0;
                do {
                  aux = stopTime.get(stopTime.DATE);
                  System.out.print("Introduce the date (Default "+aux+"): ");
                  try {
                    line = systemInReader.readLine().trim();
                    date = Integer.parseInt(line);
                  }catch(NumberFormatException e){
                    if(line.length() == 0)
                      date = aux;
                    else{
                      System.err.println(e);
                      date = -1;
                    }
                  }
                  if ( (date < 1) || (date > 31))
                    System.err.println(
                        "Error: The value of date read is not valid.");
                }while ( (date < 1) || (date > 31));

                hour = 0;
                do {
                  aux = stopTime.get(stopTime.HOUR_OF_DAY);
                  System.out.print("Introduce the hour (Default "+aux+"): ");
                  try {
                    line = systemInReader.readLine().trim();
                    hour = Integer.parseInt(line);
                  }catch(NumberFormatException e){
                    if(line.length() == 0)
                      hour = aux;
                    else{
                      System.err.println(e);
                      hour = -1;
                    }
                  }
                  if ( (hour < 0) || (hour > 23))
                    System.err.println(
                        "Error: The value of hour read is not valid.");
                }while ( (hour < 0) || (hour > 23));

                minute = 0;
                do {
                  aux = stopTime.get(stopTime.MINUTE);
                  System.out.print("Introduce the minutes (Default "+aux+"): ");
                  try {
                    line = systemInReader.readLine().trim();
                    minute = Integer.parseInt(line);
                  }catch(NumberFormatException e){
                    if(line.length() == 0)
                      minute = aux;
                    else{
                      System.err.println(e);
                      minute = -1;
                    }
                  }
                  if ( (minute < 0) || (minute > 59))
                    System.err.println(
                        "Error: The value of minutes read is not valid.");
                }while ( (minute < 0) || (minute > 59));

                second = 0;
                do {
                  aux = 0;
                  System.out.print("Introduce the seconds (Default "+aux+"): ");
                  try {
                    line = systemInReader.readLine().trim();
                    second = Integer.parseInt(line);
                  }catch(NumberFormatException e){
                    if(line.length() == 0)
                      second = aux;
                    else{
                      System.err.println(e);
                      second = -1;
                    }
                  }
                  if ( (second < 0) || (second > 59))
                    System.err.println(
                        "Error: The value of second read is not valid.");
                }while ( (second < 0) || (second > 59));
                System.out.println();

                stopTime.set(year, month, date, hour, minute, second);
                stopTime.set(stopTime.MILLISECOND, 0);

                interval = 0;
                do{
                  System.out.print("Introduce the time interval between the set of tests in minutes: ");
                  try {
                    line = systemInReader.readLine().trim();
                    interval = Integer.parseInt(line);
                  }catch(NumberFormatException e){
                    if(line.length() != 0)
                      System.err.println(e);
                    interval = -1;
                  }
                  if((interval < 1)||(interval > 1440))  // 24 hours
                    System.err.println("Error: The value of time interval between set of tests read is not valid.");
                }while((interval < 1)||(interval > 1440));
                System.out.println();

                numTest = 0;
                do{
                  System.out.print("Introduce the number of tests to be placed in each interval: ");
                  try {
                    line = systemInReader.readLine().trim();
                    numTest = Integer.parseInt(line);
                  }catch(NumberFormatException e){
                    if(line.length() != 0)
                      System.err.println(e);
                    numTest = -1;
                  }
                  if((numTest < 1)||(numTest > 3600))  // 3600 testes
                    System.err.println("Error: The value of number of tests read is not valid.");
                }while((numTest < 1)||(numTest > 3600));
                System.out.println();

                testInterval = 0;
                do{
                  System.out.print("Introduce the time interval between tests in seconds: ");
                  try {
                    line = systemInReader.readLine().trim();
                    testInterval = Integer.parseInt(line);
                  }catch(NumberFormatException e){
                    if(line.length() != 0)
                      System.err.println(e);
                    testInterval = -1;
                  }
                  if((testInterval < 1)||(testInterval > 3600))  // 1 hour
                    System.err.println("Error: The value of time interval between tests read is not valid.");
                }while((testInterval < 1)||(testInterval > 3600));
                System.out.println();

                // Read the first configuration from configuration file

                if(!session.readConfigFile(session.numRS)){
                  System.err.println("Error: Error reading from configuration file.");
                  testError = true;
                }else{ // If everything is ok continue

                  // Determine the number of intervals of test between start and end of tests.
                  testStartTime = 0;
                  startTime = (long)session.arrayOfdataRS[session.numRS].startTime.timeInMillisecondsSince1970();
                  if(startTime < (long)session.hiResTimer.milliTime()){
                    System.err.println("\nError: The Start time introduced is in the past.");
                    testError = true;
                    break;
                  }

                  numIntervals = (stopTime.getTimeInMillis()-startTime)/((interval*60*1000));
                  if(numIntervals < 0){
                    System.err.println("\nError: The Stop time introduced is not valid (Stop time < Start time).");
                    testError = true; //Client will not request a one more RS (see below)
                    break;
                  }

                  if(numIntervals == 0){
                    System.err.println("\nError: For this configuration the number of intervals is equal to sero.");
                    testError = true; //Client will not request a one more RS (see below)
                    break;
                  }

                  closeConnection = false;

                  totalTimeToWait = 0;
                  j = 0;
                  for(i=0, j = 0;i<session.arrayOfdataRS[session.numRS].numberOfPackets;i++,j++){

                    totalTimeToWait = totalTimeToWait + session.arrayOfdataSSD[session.numRS][j].slotParameter.timeInMillisecondsSince1900();
                    if(j == (session.arrayOfdataSSD[session.numRS].length -1))  // If we are at the end of the array of SSD we have to
                      j = -1;                            // turn around.  j = -1 -> when incremented by the cycle
                                                         // for, it will have the correct value (zero).
                  }

                  // Verify total tests duration and interval duration
                  if((numTest*testInterval > (interval*60))||((testInterval*(numTest-1)*1000+totalTimeToWait) > (interval*60*1000))){
                    System.err.println("Error: The total tests duration is longer than interval between set of tests.");
                    System.err.println();
                    System.err.println("  Possible errors:");
                    System.err.println();
                    System.err.println("  i)   The number of tests times test duration is longer than interval between set of tests.");
                    System.err.println("  ii)  The sum of start time of the last test and total test duration is longer than start time of set of tests");
                    System.err.println("       plus total duration of a set of tests.");
                    System.err.println();
                    numIntervals = 0;
                  }else
                    System.out.println("Number of intervals = "+numIntervals);

                  // The user can define which sender UDP port he wants to use during test session. In this case
                  // the sender port is defined as a value different of zero. If the sender port is zero a free
                  // port will be used. For the Confidence Interval sessions we copy the configurations of the first RS to process
                  // all other RS. If the senderPort is defined as zero in the first RS it will have a value of a free
                  // port after the RS is processed. So we must guarantee that its value is zero for the other RS. To
                  // do so we use the variable defineSenderPort.
                  defineSenderPort = false;
                  if(session.arrayOfdataRS[0].senderPort == 0)
                    defineSenderPort = true;

                  for(i=0;i<numIntervals;i++){  // for each interval
                    testStartTime = startTime + i*interval*60*1000; // Test start time for this interval
                    for(j=0;j<numTest;j++){ // for each test in this interval
                      session.arrayOfdataRS[session.numRS].startTime.timestamp(testStartTime+j*(testInterval*1000),(short)1);
                      if(!session.requestSession(closeConnection)){
                        System.err.println("Error: Error in processing Request Session.");
                        System.err.println("Error: The Confidence Interval session will not be processed.");
                        testError = true;
                        if(session.numRS != 0)  // send the SS to Server and Session-Senders. They will free the UDP port allocated to this session (on senders and receivers).
                          StopSessions.stopAllSessions(session.socketToServer,session.arrayOfSocketToSender,session.numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
                        break;
                      }

                      session.arrayOfdataRS[session.numRS] = new RequestSessionToBeSend(session.arrayOfdataRS[0]);
                      session.arrayOfSenderTCPportNumber[session.numRS] = session.arrayOfSenderTCPportNumber[0];

                      if(defineSenderPort) // Used to verify if a free port should be used.
                        session.arrayOfdataRS[session.numRS].senderPort = 0;

                      session.arrayOfdataSSD[session.numRS] = new  ScheduleSlotDescriptionsToBeSend[session.arrayOfdataRS[0].numberOfScheduleSlots];
                      for(int c=0;c<session.arrayOfdataRS[0].numberOfScheduleSlots;c++)
                        session.arrayOfdataSSD[session.numRS][c] = new  ScheduleSlotDescriptionsToBeSend(session.arrayOfdataSSD[0][c]);
                    }
                    if(testError)
                      break;
                  }

                  // For Confidence Interval sessions the startingTestSessions should be processed immediately after
                  // the RS process.
                  if((!testError)&&(numIntervals != 0)){
                    if (!session.startingTestSessions(true)) { // startingTestSessions(boolean saveArrayOfRStoFiles)
                      System.err.println("Error: Error in processing Starting Test Sessions.\n");
                      System.err.println(
                          "Error: The Confidence Interval session will not be processed.");
                    }
                  }
                  testError = true; //Client will not request a one more RS (see below)
                }
              }
            }

            if(!testError){
              closeConnection = false;

              if(!session.requestSession(closeConnection))
                System.err.println("Error: Error in processing Request Session.");
            }

          }
          catch (IOException e) {
            System.err.println(e);
          }

          break;
        case 2:
          if (!session.startingTestSessions(true))  // startingTestSessions(boolean saveArrayOfRStoFiles)
            System.err.println("Error: Error in processing Starting Test Sessions.\n");
          break;
        case 3:
          System.err.println("Error: You can not process a StopSession before Requesting a Start Test Session.");
          break;
        case 4:
          if((session.numRS != 0)&&(!session.socketToServer.isClosed())){
            System.err.println("Error: You have requested a test session, but you can not execute a Fetch-Session before starting a Test-Session.");
            break;
          }

          OWAMP_FetchClient fetch = new OWAMP_FetchClient(serverAddress,serverTCPportNumber, new Socket(), systemInReader, session.hiResTimer);
          while(fetch.processFetchSession(true));  // processFetchSession(boolean readArrayOfRSfromFiles)
          break;
        case 5:
          // Test socket connection

          if((!session.socketToServer.isConnected())||(session.socketToServer.isClosed())){
            System.err.println("Error: You should connect to a server first, before trying this operation.");
            System.err.println("Error: Error in processing Starting Test Sessions.\n");
            break;
          }

          if((session.numRS != 1)&&(session.numRS != 0)){
             System.err.println("Error: You can not execute a Start-Session and Fetch with more than one RS.");
             break;
          }

          if (!session.startingTestSessions(true))  // startingTestSessions(boolean saveArrayOfRStoFiles)
            System.err.println("Error: Error in processing Starting Test Sessions.\n");

          break;

      }

      System.out.println("Press Enter to continue...");
      try {
        systemInReader.readLine();
      }
      catch (IOException e) {
        System.err.println(e);
      }

    }while(true);

  }

}


