/**
 * <p>Title: OWAMP Protocol</p>
 * <p>Description: This section of this project contains the class used to implement the OWAMP Server.</p>
 * <p>Copyright: Copyright (c) 2005</p>
 * <p>Company: University of Aveiro / Institute of Telecommunications - Aveiro pole</p>
 * @authors: Helder Veiga and Teresa Pinho
 * @version: 2.1
 * @Last updated: 9-12-2005
 *
 * Copyright (C) 2005  Helder Veiga and Teresa Pinho
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.  A HTML version of
 * the GNU General Public License can be seen at
 * http://www.gnu.org/licenses/gpl.html.
 *
 */


import java.net.*;
import java.io.*;
import java.util.*;

public class OWAMP_Server extends OWAMP_Base{

  // Constructors

  /**
   * Constructor for class OWAMP_Server.
   * @param socketToClient Socket - represents the socket connection to the Client.
   * @param upTime Timestamp - uptime is a timestamp representing the time when the current instantiation of
   * the server started operating, in milliseconds since 1Jan1970.
   * @param receiverTCPport int - TCP port number to be used in OWAMP-Control with all OWAMP Session-Receivers.
   * @param hiResTimer HiResTimer - represents the high resolution timer used to get the time.
   */

  OWAMP_Server(Socket socketToClient, Timestamp upTime, int receiverTCPport, HiResTimer hiResTimer){

    rand = new Random();

    this.socketToClient = new Socket();
    this.socketToClient = socketToClient;
    try{
      this.socketToClient.setSoTimeout(TIMEOUT);
    }
    catch(IOException e){
      System.err.println(e);
    }

    buffer = new byte[DEFAULT_PACKET_LENGTH];

    arrayOfSocketToReceiver = new Socket[1];
    arrayOfdataRS= new RequestSessionReceiver[1];
    arrayOfdataSSD = new ScheduleSlotDescriptionsReceiver[1][];

    this.upTime = new Timestamp(upTime.integerPartOfSeconds, upTime.fractionalPartOfSeconds);
    this.receiverTCPport = receiverTCPport;

    this.hiResTimer = hiResTimer;
    dataSSr = new StopSessionsReceiver();

    try{
      serverSocket = new ServerSocket();    // Used to verify if a socket connection is closed
                                            // using method isSocketConnected.
    }catch(IOException e){
      System.err.println(e);
    }

  }


  // Instance variables

  private Socket socketToClient;                  // Represents the socket connection to the Client.
  private Socket[] arrayOfSocketToReceiver;       // Array of sockets connections to all Receivers of a OWAMP
                                                  // test session, with more than one test session.
  private InetAddress receiverInetAddress = null;
  private byte sid[] = new byte[16];
  private int mode;                               // The mode value to be used in the
                                                  // current OWAMP control and test sessions.

  private RequestSessionReceiver[] arrayOfdataRS;  /* Array of datagrams Request Session (it has all request
                                                    * session made by the client).
                                                    */

  private short numRS;          // Number of Request Sessions made by the client. It's automatically incremented
                                // for each new Request Sessions made.

  private ScheduleSlotDescriptionsReceiver[][] arrayOfdataSSD; // For each Request Session made by the client
                                                               // we have an array of ScheduleSlotDescriptions.
  private AcceptSessionReceiver dataASr = new AcceptSessionReceiver();
  private StartingTestSessionsToBeSend dataSTS = new StartingTestSessionsToBeSend();
  private Start_AckReceiver dataSAr = new Start_AckReceiver();
  private SessionDescriptionRecordReceiver[] arrayOfdataSDRr;  // Used to receive the SDR from Client
  private SessionDescriptionRecordToBeSend dataSDRs;  // SDR of each Client
                                                      // used to send the SS to Receiver.
  private static File serverRSdir = new File("serverRSdir");   // The directory of all RS received from
                                                               // Clients in this Server.
  private int receiverTCPport;       // TCP port number to be used in OWAMP-Control with all OWAMP Session-Receivers.


  // Instance methods

  /**
   * This method start a new process to implement the Server.
   */

  public void run(){

    try{

                                     // Process Connection Setup Server Greeting
      if(!connectionSetupResponse(mode, socketToClient, upTime)){    //connectionSetupResponse(int modes, Socket socket, Timestamp upTime)
        System.err.println("Error: Error in processing Connection Setup with client in socket " +socketToClient);
        closeSocket(socketToClient);
        return;
      }

      numRS = 0;    // Number of Request Session made - there are no Request Session yet.

      while(true){  // The server keeps listening for the incoming packets.
        if(socketToClient.isClosed()){
          System.out.println("End of the current Session with Client's machine on "+socketToClient.toString());
          System.out.println();
          return;
        }

        socketToClient.setSoTimeout(0);  // Configure time to wait for packets from Client to infinite.

        int readn = 0;                   // number of bytes read.

        // To receive a packet we just must read the first byte to determine which packet was received.

        readn = receivePacket(socketToClient,buffer,0,1);   // receivePacket (Socket socket, byte[] _buffer, int off, int len)
        socketToClient.setSoTimeout(TIMEOUT);

        if(readn != 1){
          if(isSocketConnected(socketToClient))  // An error occurred reading from socket
            System.err.println("Error: Error receiving Packet.");
          else{
            // Note: A client may not fecth the session results after its end, and close the socket to server.
            // In this case the error socket to client no longer available shouldn't be shown.
            // When all StopSessions are received from Client numSS = numRS, the numRS is made zero.
            if(numRS != 0){
              System.err.println("Error: The socket connection to Client on " +
                                 socketToClient +
                  " is no longer available. All sessions will be finished.\n");
              System.out.println("End of the current Session.\n");
            }else
              System.out.println("End of the current Session with Client's machine on "+socketToClient.toString()+".\n");
          }

          if(numRS != 0)      // send the SS to Client (it will be send to Client only if the socket is connected) and to Session-Receivers.
            StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)

          closeSocket(socketToClient);
          return;
        }

        if(!processPacketReceived()){
          closeSocket(socketToClient);
          for(int i=0;i<numRS;i++)
            closeSocket(arrayOfSocketToReceiver[i]);       // Close the TCP connection to the Session-Receiver
          return;
        }

      }

    }
    catch(Exception ex){     // catches any type of exception.
      ex.printStackTrace();
    }
    finally{
      closeSocket(socketToClient);
      for(int i=0;i<numRS;i++)
        closeSocket(arrayOfSocketToReceiver[i]);       // Close the TCP connection to the Session-Receiver

      // Runs the finalization methods of any objects pending finalization. Then, run the garbage colector.
      System.runFinalization();
      System.gc();
    }
  }


  /**
   *  Method to receive the packet Request Session send by a client.
   *  To each Request-Session message, an OWAMP server MUST respond with an Accept-Session message
   * @param accept short - specify if the server accepts the request. In this message, zero in the
   * Accept field means that the server is willing to conduct the session.  A value of 1 indicates
   * rejection of the request.  All other values are reserved.
   * @return boolean - This method returns true if the request operation was ok, otherwise it returns false.
   */

  public boolean requestSessionResponse(short accept){

    // In variable byte[] buffer we have the packet Request Session received

    // Process the complete Request Session reception ( RS + SSD + IZP)

    System.out.println("Request Session from Client on "+socketToClient);
    System.out.println();

    arrayOfdataRS[numRS] = new RequestSessionReceiver();
    System.arraycopy(buffer,0,arrayOfdataRS[numRS].buffer,0,arrayOfdataRS[numRS].PACKET_LENGTH);

                               // Initialize the list to receive the ScheduleSlotDescriptions packets.

    int numberOfScheduleSlots = Packets.getIntFromByteArray(buffer,4);  // numberOfScheduleSlots to receive
    arrayOfdataSSD[numRS] = new ScheduleSlotDescriptionsReceiver[numberOfScheduleSlots];

    if(!arrayOfdataRS[numRS].receiveCompleteRequestSession(arrayOfdataSSD[numRS],socketToClient)){
      System.err.println("Error: Error processing complete Request Session, from Client on "+socketToClient);

      // Send the Stop All Session to Client. In case of failure,
      // the Server should stop all sessions and stop this OWAMP-Control session.

      StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS+1,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)

      return false;
    }

    // Test confSender and confReceiver - At least one of Conf-Sender and Conf-Receiver MUST be 1.

    if ( (arrayOfdataRS[numRS].confSender == 0) && (arrayOfdataRS[numRS].confReceiver == 0)) {
      System.err.println("Error: At least one of Conf-Sender and Conf-Receiver MUST be 1.");
      System.err.println();

      // Send the AcceptSession packet

      System.out.println("AcceptSession to client on " + socketToClient);
      System.out.println("The Server rejected the Request Session.");
      System.out.println();
      accept = 1;
      AcceptSessionToBeSend dataAS = new AcceptSessionToBeSend(accept, 0, sid);
      dataAS.sendAS(socketToClient);

      return true; // The server should not close the TCP connection in case of error. The client
                   // may close it.
    }

    /* SID is the session identifier.  It can be used in later sessions as
     an argument for the Fetch-Session command.  It is meaningful only if
     Conf-Receiver is 0.  This way, the SID is always generated by the
     receiving side.*/

    if(arrayOfdataRS[numRS].confReceiver != 0){
      sid = generateSid();                // If confReceiver is set, the value of sid is generated
                                          // by the server.
      System.arraycopy(sid,0,arrayOfdataRS[numRS].sid,0,sid.length);
    }


    // Sends the RS to the receiver

                                  // Determine Session-Receiver address

    receiverInetAddress = determineSeReAddress(arrayOfdataRS[numRS].mbz_iPVN, arrayOfdataRS[numRS].receiverAddress);


    // Process the connection Setup Request to the Session-Receiver

    System.out.println("connectionSetupRequest to receiver "+receiverInetAddress);
    System.out.println();

    arrayOfSocketToReceiver[numRS] = connectionSetupRequest(mode,receiverInetAddress,receiverTCPport,TIMEOUT);


    if(arrayOfSocketToReceiver[numRS] == null){

      System.err.println("Error: The Session-Receiver specified doesn't exist.\n");
      System.err.println("Error: Error processing connection Setup");
      System.err.println();

      // Send the AcceptSession packet

      System.out.println("AcceptSession to client on "+socketToClient);
      System.out.println("\nThe Server rejected the Request Session.");
      System.out.println();
      accept = 1;
      AcceptSessionToBeSend dataAS = new AcceptSessionToBeSend(accept,0,sid);
      dataAS.sendAS(socketToClient);

      return true;  // The server should not close the TCP connection in case of error. The client
                    // may close it.

    }

    System.out.println("Request Session to Receiver "+(numRS+1)+" on "+arrayOfSocketToReceiver[numRS]);
    System.out.println();

    if(!requestSessionToReceiver(numRS)){  // requestSessionToReceiver(int index)
      System.err.println("Error: Error processing Request Session to Session-Receiver on "+arrayOfSocketToReceiver[numRS]);
      accept = 1;

      closeSocket(arrayOfSocketToReceiver[numRS]);       // Close the TCP connection to the Session-Receiver

    }else
      accept = 0;

    /* To each Request-Session message, an OWAMP server MUST respond with an Accept-Session message.
     * If the server rejects a Request-Session command, it SHOULD not close the TCP connection.  The
     * client MAY close it if it gets negative response to Request-Session.
     * The meaning of Port in the response depends on the values of Conf-Sender and Conf-Receiver in the
     * query that solicited the response. If both were set, Port field is unused.  If only Conf-Sender was
     * set, Port is the port to expect OWAMP-Test packets from.  If only Conf- Receiver was set, Port is the
     * port to send OWAMP-Test packets to.
     * If only Conf-Sender was set, SID field in the response is unused. Otherwise, SID is a unique server-
     * generated session identifier.  It can be used later as handle to fetch the results of a session.
     */

    // Send the AcceptSession packet

    System.out.println("AcceptSession to client on "+socketToClient);
    System.out.println();

    AcceptSessionToBeSend dataAS = new AcceptSessionToBeSend(accept,arrayOfdataRS[numRS].receiverPort,arrayOfdataRS[numRS].sid);
    dataAS.sendAS(socketToClient);

    if(accept != 0){        // The server rejected the requested session.
      closeSocket(arrayOfSocketToReceiver[numRS]);       // Close the TCP connection to the Session-Receiver
      System.err.println("Error: The server rejected the requested session.");
      System.err.println();
      System.err.println("  Possible errors:");
      System.err.println();
      System.err.println("  i)   The Session-Receiver doesn't exist.");
      System.err.println("  ii)  An error occurred processing connection to Receiver.");
      System.err.println("  iii) The Session-Receiver has no free UDP port to process this test session because all\n"+
                         "       the UDP ports in the range specified when the Session-Receiver was started are binded.");
      System.err.println("  iv)  An error occurred processing Request Session with Receiver.");
      System.err.println("  v)   The receiver rejected the requested session.");
      System.err.println();
      return true;  // The server should not close the TCP connection in case of error. The client
                    // may close it.
    }else{
      // An OWAMP client may issue zero or more Request-Session messages to an OWAMP server. So we have to increment
      // the number of Request Session made (numRS) and the size of the arrays used to keep the sessions
      // configurations. The new incremented position will be used by the next session configured.

      numRS++;   // Increment the number of Request Session made

      Socket[] arrayOfSocketToReceiverAux = new Socket[numRS+1];
      System.arraycopy(arrayOfSocketToReceiver,0,arrayOfSocketToReceiverAux,0,numRS);
      arrayOfSocketToReceiver = arrayOfSocketToReceiverAux;

      RequestSessionReceiver[] arrayOfdataRSAux = new RequestSessionReceiver[numRS+1];
      System.arraycopy(arrayOfdataRS,0,arrayOfdataRSAux,0,numRS);
      arrayOfdataRS = arrayOfdataRSAux;

      ScheduleSlotDescriptionsReceiver[][] arrayOfdataSSDAux = new ScheduleSlotDescriptionsReceiver[numRS+1][];
      for(int i=0;i<numRS;i++){
        arrayOfdataSSDAux[i] = new ScheduleSlotDescriptionsReceiver[arrayOfdataSSD[i].length];
        System.arraycopy(arrayOfdataSSD[i],0,arrayOfdataSSDAux[i],0,arrayOfdataSSD[i].length);
      }
      arrayOfdataSSD = arrayOfdataSSDAux;
    }

    return true;

  }


  /**
   * Method to send the packet Request Session to the Session-Receiver.
   * To each Request-Session an OWAMP Session-Receiver MUST respond with an Accept-Session message.
   * @param index int - The variable index represents the current Request Session in array arrayOfdataRS.
   * @return boolean - It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean requestSessionToReceiver(int index){

    // Test socket connection

    if(!isSocketConnected(arrayOfSocketToReceiver[index])){
      System.err.println("Error: You should connect to a Session-Receiver first, before trying this operation.");
      return false;
    }

    // Process the complete RequestSession command

    arrayOfdataRS[index].sendTheCompleteRSreceived(arrayOfdataSSD[index],arrayOfSocketToReceiver[index]);

    //  To each Request-Session message, an OWAMP Receiver MUST respond with an Accept-Session message.

    // Receive the AcceptSession packet

    if(!dataASr.receiveAS(arrayOfSocketToReceiver[index])){
      System.err.println("Error: Error receiving packet AcceptSession in operation Request session from receiver on "+arrayOfSocketToReceiver[index]);
      closeSocket(arrayOfSocketToReceiver[index]);
      return false;
    }

    System.out.println("AcceptSession from receiver on "+arrayOfSocketToReceiver[index]);
    System.out.println();

    boolean error = false;
    switch (dataASr.accept) {
      case 0: // The Session-Receiver accepts the RS
        break;
      case 1: // Failure, reason unspecified (catch-all).
        System.err.println(
            "Error: Message from receiver - Failure, reason unspecified (catch-all).");
        System.err.println();
        error = true;
        break;
      case 2: // Internal error.
        System.err.println("Error: Message from receiver - Internal error.");
        System.err.println();
        error = true;
        break;
      case 3: // Some aspect of request is not supported.
        System.err.println(
            "Error: Message from receiver - Some aspect of request is not supported.");
        System.err.println();
        error = true;
        break;
      case 4: // Cannot perform request due to permanent resource limitations.
        System.err.println("Error: Message from receiver - Cannot perform request due to permanent resource limitations.");
        System.err.println();
        error = true;
        break;
      case 5: // Cannot perform request due to temporary resource limitations.
        System.err.println("Error: Message from receiver - Cannot perform request due to temporary resource limitations.");
        System.err.println();
        error = true;
        break;
      default: // Reserved values
        System.err.println(
            "Error: The Server doesn't support the accept mode received from the receiver.");
        System.err.println();
        error = true;
    }

    if (error) { // If an error occurred, return false.
      System.err.println("Request Session rejected by the receiver on "+arrayOfSocketToReceiver[index]);
      closeSocket(arrayOfSocketToReceiver[index]);
      return false;
    }

    /* Test confSender and confReceiver:
       Conf-Sender and Conf-Receiver MUST be set to 0 or 1 by the client.
       The server MUST interpret any non - zero value as 1.If the value is
       1, the server is being asked to configure the corresponding agent
       (sender or receiver).In this case, the corresponding Port value
       SHOULD be disregarded by the server.At least one of Conf - Sender and
       Conf - Receiver MUST be 1. (Both can be set, in which case the server
       is being asked to perform a session between two hosts it can configure.)
     */
    switch(arrayOfdataRS[index].confReceiver){
      case 0:
        if(arrayOfdataRS[index].confSender == 1)
          arrayOfdataRS[index].senderPort = dataASr.port;
        break;
      case 1:
      default:
        if(arrayOfdataRS[index].confSender == 0)
          arrayOfdataRS[index].receiverPort = dataASr.port;
    }

    return true;

  }


  /**
   * Having requested one or more test sessions and received affirmative Accept-Session responses, an OWAMP
   * client may start the execution of the requested test sessions by sending a Start-Sessions message to
   * the server. The Server should send the StartingTestSession command to each Session-Receiver. If all
   * receiver accept these requests, the Server sends an Control_Ack packet to Client with accept equal to
   * 0, otherwise accept value is 1.
   * @param accept byte - If Accept is 1, the Start-Sessions request was rejected; zero means that the
   * command was accepted.  All other values are reserved.  The server MAY and the client SHOULD close the
   * connection in the case of a negative response.
   * @return boolean -  It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean startingTestSessionsResponse(byte accept){ // process Starting Test Sessions Response

    // In variable byte[] buffer we have the packet startingTestSessions received
    // Test the startingTestSessions packet

    StartingTestSessionsReceiver dataSTSr = new StartingTestSessionsReceiver();
    if(!dataSTSr.testIntegrityZeroPadding(buffer,(short)16,(short)16)){ // testIntegrityZeroPadding(byte[] _buffer, short pos, short length)
      System.err.println("Error: Error receiving packet startingTestSessions from client on "+socketToClient);

      // Send the Stop All Session to Client. In case of failure,
      // the Server should stop all sessions and stop this OWAMP-Control session.

      StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)

      return false;
    }

    System.out.println("Starting Test Sessions from Client on "+socketToClient);
    System.out.println();

    if(numRS < 1){  // The client didn't ask for a Request Session yet.
      System.err.println("Error: There is no sessions to start.");
      accept = 1;
    }

    Start_AckToBeSend dataSAtoClient;

    switch(accept){
      case 0:
        // The server SHOULD start all OWAMP-Test streams immediately after it sends the response or
        // immediately after their specified start times, whichever is later.
        // Send the StartingTestSessions packet to each receiver to start the test sessions. Each request
        // is considered separately in a different socket connection.

        System.out.println("Starting Test Sessions to Receivers.");
        System.out.println();

        for(int i = 0; i<numRS;i++){     // For each RS made.

          // Test socket connection

          if(!isSocketConnected(arrayOfSocketToReceiver[i])){
            System.err.println("Error: Error in Starting Test Session operation.");
            System.err.println("Error: The socket connection to receiver "+(i+1)+" - "+arrayOfSocketToReceiver[i]+" - is closed.");
            System.err.println("Error: All sessions requested so far will be finished.");
            System.err.println();
            accept = 1;
            break;
          }

          // Send StartingTestSessions packet to the Receiver

          dataSTS.sendSTS(arrayOfSocketToReceiver[i]);

          // Receive the Start-Ack packet

          System.out.println("Start_Ack from Receiver "+(i+1)+" on "+arrayOfSocketToReceiver[i]);
          System.out.println();

          if(!dataSAr.receiveSA(arrayOfSocketToReceiver[i])){
            System.err.println("Error: Error in Starting Test Session operation.");
            System.err.println("Error: Error receiving packet Control-Ack in operation Starting Test Session, from receiver "+(i+1)+": "+arrayOfSocketToReceiver[i]+".");
            System.err.println("Error: All sessions requested so far will be finished.");
            System.err.println();
            accept = 1;
            break;
          }

          boolean error = false;
          switch (dataSAr.accept) {
            case 0: // The Session-Seceiver accepts the STS
              break;
            case 1: // Failure, reason unspecified (catch-all).
              System.err.println(
                  "Error: Message from receiver - Failure, reason unspecified (catch-all).");
              System.err.println();
              error = true;
              break;
            case 2: // Internal error.
              System.err.println(
                  "Error: Message from receiver - Internal error.");
              System.err.println();
              error = true;
              break;
            case 3: // Some aspect of request is not supported.
              System.err.println(
                  "Error: Message from receiver - Some aspect of request is not supported.");
              System.err.println();
              error = true;
              break;
            case 4: // Cannot perform request due to permanent resource limitations.
              System.err.println("Error: Message from receiver - Cannot perform request due to permanent resource limitations.");
              System.err.println();
              error = true;
              break;
            case 5: // Cannot perform request due to temporary resource limitations.
              System.err.println("Error: Message from receiver - Cannot perform request due to temporary resource limitations.");
              System.err.println();
              error = true;
              break;
            default: // Reserved values
              System.err.println(
                  "Error: The server doesn't support the accept mode received from the Session-Receiver.");
              System.err.println();
              error = true;
          }

          if (error) { // If an error occurred
            System.err.println("Error: Error in Starting Test Session operation.");
            System.err.println("Error: The Session-Receiver "+(i+1)+" - "+arrayOfSocketToReceiver[i]+" - has rejected the Starting Test Session request.");
            System.err.println("Error: All sessions requested so far will be finished.");
            System.err.println();
            accept = 1;
            break;
          }

        }

        // If a receiver didn't accept a request session, the Server should inform the Client sending the
        // Start_Ack packet with accept equal to 1 and terminate all sessions.

        // Send datagram Start_AckToBeSend to Client

        System.out.println("Start_Ack to Client on "+socketToClient);
        System.out.println();
        dataSAtoClient = new Start_AckToBeSend(accept);
        dataSAtoClient.sendSA(socketToClient);

                       // Stop all sessions started till now.

        if(accept == 1){

          System.err.println("Error: The Start-Sessions request was rejected.");
          System.err.println("Error: All sessions requested so far will be finished.");
          System.err.println();
          System.err.println("  Possible errors:");
          System.err.println();
          System.err.println("  i)   A Session-Receiver doesn't exists.");
          System.err.println("  ii)  A failure occurred in processing Starting Test Sessions between the Server and a Receiver.");
          System.err.println("  iii) A Session-Receiver has rejected the Starting Test Session request.");
          System.err.println("  iv)  A failure occurred receiving the Starting Test Sessions command from Client "+socketToClient);
          System.err.println();

          // Stop all sessions started till now. In case of failure,
          // the Server should stop all sessions and stop this OWAMP-Control session.

          StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,false);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)

        }

        long startTime = (long)hiResTimer.milliTime();


                                        // Calculate the total time to wait
        double totalTimeToWait = 0;
        double ti;

        for(int i = 0; i<numRS;i++){
          ti = totalTimeToWait(arrayOfdataRS[i], arrayOfdataSSD[i], hiResTimer);  // totalTimeToWait(RequestSession dataRS, ScheduleSlotDescriptions[] arrayOfdataSSD, HiResTimer hiResTimer)
          if(totalTimeToWait < ti)
            totalTimeToWait = ti;
        }

        arrayOfRStoFiles();  // Record the RS that have been accepted and started to a file. This information
                             // will be used in Fetch-Sessions.

        /* During total time to wait test if the client didn't request a Stop Session or if a Stop Session
         * haven't been received from a Receiver.
         */

        if(!testStopSession(startTime, totalTimeToWait)) // testStopSession(long startTime, long totalTimeToWait)
          return false;

        break;
      case 1:

        // Send datagram Start_AckToBeSend to Client

        System.out.println("Start_Ack to Client on "+socketToClient);
        System.out.println();
        dataSAtoClient = new Start_AckToBeSend(accept);
        dataSAtoClient.sendSA(socketToClient);

        // The server didn't accept the command Start-Sessions request.
        System.err.println("Error: The server didn't accept the command Start-Sessions request from client on "+socketToClient);
        System.out.println();

                      // Stop all sessions started till now.

        // Stop all sessions started till now. In case of failure,
        // the Server should stop all sessions and stop this OWAMP-Control session.

        StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,false);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)

        break;
      default:
        // Send datagram Start_AckToBeSend to Client.

        System.out.println("Start_Ack to Client on " + socketToClient);
        System.out.println();
        dataSAtoClient = new Start_AckToBeSend(accept);
        dataSAtoClient.sendSA(socketToClient);
        System.err.println("Error: The accept value returned by the server is not valid.");

                       // Stop all sessions started till now.

        // Stop all sessions started till now. In case of failure,
        // the Server should stop all sessions and stop this OWAMP-Control session.

        StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,false);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)

        return false;
    }

    return true;

  }


  /**
   * During total time to wait, this method is used to test if the user didn't request a Stop Session.
   * @param startTime long - this variable represents the time that the current session started operating.
   * @param totalTimeToWait long - the time the Client should wait to finish the current session.
   * @return boolean -  It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean testStopSession(long startTime, double totalTimeToWait){

    // Wait for command Stop Session or end of total time to wait for the end of all sessions.

    // Wait for the StopSessions packets during totalTimeToWait

    short stopVal = 0;  // Auxiliary variable used to verify if a session was stopped.
    totalTimeToWait = totalTimeToWait + 2*TIMEOUT; // The Server should wait for totalTimeToWait + 2 * TIMEOUT sec
                                                   // to guaranty that, at the end of totalTimeToWait the
                                                   // Server and the Client don't send the Stop Session
                                                   // at same time.
    double totalTimeToWaitAux = 3*TIMEOUTTOWAIT;


                        // wait for Stop Session

    int i = 0, index = 0,
          numOfSS = 0;          // Number of stopped sessions.
    do {

      // Test if a Stop Session was sent by the Control-Client and, if yes, process it.

      stopVal = dataSSr.receiveSS(socketToClient, totalTimeToWaitAux);

      if (stopVal == -2){ // The socket connections was reset.
        StopSessions.stopAllSessions(socketToClient, arrayOfSocketToReceiver, numRS, false);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
        System.err.println("All sessions will be finished.");
        System.err.println();
        return false;
      }

      if ( (stopVal == -1) || ( (stopVal == 1) && (dataSSr.accept != 0))) { // If stopSession's accept val = 1 or an error ocurred stop operation.
        if (stopVal != -1) {
          System.out.println("StopSession from Client on " + socketToClient);
          verifyAcceptValue(dataSSr.accept, "Client",socketToClient, "");
          StopSessions.stopAllSessions(socketToClient, arrayOfSocketToReceiver, numRS, false);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
        }else
          StopSessions.stopAllSessions(socketToClient, arrayOfSocketToReceiver, numRS, true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
        System.err.println("All sessions will be finished.");
        System.err.println();
        return false;
      }

      if (stopVal == 1) { // If there is a StopSession and accept = 0, the Server should receive the
                          // complete SS and send the SS to the repectives Receivers.
        System.out.println("StopSession from Client on " + socketToClient);
        System.out.println();

        // Receive the Session Description Records and record them in the arrayOfSessionDescriptionRecords.
        arrayOfdataSDRr = new SessionDescriptionRecordReceiver[dataSSr.numberOfSessions];  // Used to receive the SDR from Client
        arrayOfdataSDRr = dataSSr.receiveCompleteSS(socketToClient);

        if(arrayOfdataSDRr == null){
          System.err.println("Error: Error in StopSession operation from Client.");
          System.err.println("All sessions will be finished.");
          System.err.println();
          StopSessions.stopAllSessions(socketToClient, arrayOfSocketToReceiver, numRS, true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
          return false;
        }

        dataSSs = new StopSessionsToBeSend( (short) 0, 1);
        int numberOfSessions = dataSSr.numberOfSessions;
        for(i = 0; i < numberOfSessions; i++, numOfSS++){
          dataSDRs = new SessionDescriptionRecordToBeSend(
              arrayOfdataSDRr[i].sid, arrayOfdataSDRr[i].nextSeqno,
              arrayOfdataSDRr[i].numberOfSkipRanges,
              arrayOfdataSDRr[i].skipRangeDescriptions);
          index = findSession(arrayOfdataSDRr[i].sid);
          dataSSs.sendCompleteSS(dataSDRs, arrayOfSocketToReceiver[index]);
          System.out.println("StopSessions to Receiver "+(index+1)+" on " + arrayOfSocketToReceiver[index]);
          System.out.println();

          // Receive the SS response from receiver
          stopVal = dataSSr.receiveSS(arrayOfSocketToReceiver[index], TIMEOUT);

          if (stopVal == -2){ // The socket connections was reset.
            System.err.println("Error: The socket connections to Receiver "+(index+1)+" on " +
                               arrayOfSocketToReceiver[index]+" was reset.");
            closeSocket(arrayOfSocketToReceiver[index]); // Close socket connection to Receiver
            StopSessions.stopAllSessions(socketToClient, arrayOfSocketToReceiver, numRS, true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
            System.err.println("All sessions will be finished.");
            System.err.println();
            return false;
          }

          if ( (stopVal != 1) || (dataSSr.accept != 0)) {
            if(stopVal == 1)
              verifyAcceptValue(dataSSr.accept, "Receiver",arrayOfSocketToReceiver[index], "");
            System.err.println(
                "Error: Error in StopSession operation with Receiver "+(index+1)+" on " +
                arrayOfSocketToReceiver[index]);
            System.err.println("All sessions will be finished.");
            System.err.println();
            closeSocket(arrayOfSocketToReceiver[index]);
            StopSessions.stopAllSessions(socketToClient, arrayOfSocketToReceiver, numRS, true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
            return false;
          }
          else
            System.out.println("StopSessions response from Receiver "+(index+1)+" on " +
                               arrayOfSocketToReceiver[index]);
          System.out.println();

          closeSocket(arrayOfSocketToReceiver[index]);
        }

        System.out.println("StopSessions response to Client on " +socketToClient);
        System.out.println();
        dataSSs = new StopSessionsToBeSend( (short) 0, 0);
        dataSSs.sendSS(socketToClient);

        // Test if not all session has been finished

        if (numOfSS == numRS) { // all session has been finished
          System.out.println("End of all session between this Server and the Client on " +
                             socketToClient);
          System.out.println();
          numRS = 0; // It will be used to verify if all sessions are finished in case the socket from client is closed before
                     // the fetch-session, on method run. Note: A client may not fecth the session results after its end and close
                     // the socket to server. In this case the error socket to client no longer available shouldn't be shown.
          return true;
        }

        // If there is a stopped session, test total time to wait again because the current
        // TotalTimeToWait maybe has respect to a stopped RS.

        double ti = 0;
        totalTimeToWait = 0;
        for (i = 0; i < numRS; i++) {
          if (!arrayOfSocketToReceiver[i].isClosed()) {
            ti = totalTimeToWait(arrayOfdataRS[i], arrayOfdataSSD[i], hiResTimer); // totalTimeToWait(RequestSession dataRS, ScheduleSlotDescriptions[] arrayOfdataSSD, HiResTimer hiResTimer)
            if (totalTimeToWait < ti)
              totalTimeToWait = ti;
          }
        }

        totalTimeToWait = totalTimeToWait + 2*TIMEOUT;

      }

      // Verify the socketToClient. If it is closed before totalTimeToWait
      // It means that a failure has occurred, possibly a loss of connection
      // to Client. The session must be finished.

      if(!isSocketConnected(socketToClient)){
        System.err.println();
        System.err.println("Error: The socket connection to Client on "+socketToClient+" is no longer available. All sessions will be finished.");
        StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,false);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
        return false;
      }

      // Test if the socket connection to the receivers was not reset by a receiver. In this case this session must be finished.
      // To detect if a socket was reset by a Receiver we can try to receive a SS from the receiver, if stopVal is -2 it means
      // that the connection to receiver was reset.
      // Note: The receivers will not send a SS to Server. This method is only used to detect if a receiver reset the
      // connection to Server.

      for (i = 0; i < numRS; i++){
        if (!arrayOfSocketToReceiver[i].isClosed()) {
          stopVal = dataSSr.receiveSS(arrayOfSocketToReceiver[i], TIMEOUTTOWAIT);

          if (stopVal == -2) { // The socket connections was reset.
            closeSocket(arrayOfSocketToReceiver[i]); // Close socket connection to Sender
            StopSessions.stopAllSessions(socketToClient, arrayOfSocketToReceiver, numRS, true); // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
            System.err.println("All sessions will be finished.");
            System.err.println();
            return false;
          }
        }
      }

    }while ((startTime + totalTimeToWait) > ((long)hiResTimer.milliTime()));


    // If totalTimeToWait occurred
    // Send a StopSession to Client

    System.out.println("TotalTimeToWait occurred");
    System.out.println("StopSession to Client on "+socketToClient);
    System.err.println("All sessions will be finished.");
    StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,true); // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)

    return true;

  }


  /* Method to process the Server's Fetch Session response. The server will be the intermediate between
   * the Control-Client and the Session-Receiver of the requested session. It will determine the address
   * of that Session-Receiver and will forward the fetch session request to it. Then it will forward
   * the controlAck message sent by the session receiver to the client. Finally, if Session-Receiver sends
   * any results, the Server will forward results to the client.
   * @return boolean - It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean processFetchSession(){

    Fetch_AckToBeSend dataFAs;

    // In variable byte[] buffer we have the packet FetchSession received

    FetchSessionReceiver dataFS = new FetchSessionReceiver();
    System.arraycopy(buffer,0,dataFS.buffer,0,dataFS.PACKET_LENGTH);
    if(!dataFS.receiveFS()){
      System.err.println("Error: Error in receiving FetchSession packet from Client on "+socketToClient);

      // Send the Fetch_Ack packet to the client

      System.out.println("Fetch-Ack to Client on "+socketToClient);
      System.out.println();
      dataFAs = new Fetch_AckToBeSend((byte)1);
      dataFAs.sendFA(socketToClient);
      System.err.println("Error: The Fetch Session command was rejected by the Server.");

      closeSocket(socketToClient);
      return false;
    }

    System.out.println("Fetch Session from Client on "+socketToClient);
    System.out.println();

    ScheduleSlotDescriptions[] arrayOfdataSSDAux = dataRSfromFile(dataFS.sid);
    if(arrayOfdataSSDAux == null){     // Read the dataRS from a file and record it in arrayOfdataRS[0].
                                       // This file, if it exits, it's name is a string representation
                                       // of the session SID and it should has the RS accepted by the
                                       // Sender and Receiver of test packets.(So, we are sore that
                                       // this session was started)

      System.err.println("Error: This Session doesn't exist.");

      // Send the Fetch_Ack packet to the client

      System.out.println("Fetch-Ack to Client on "+socketToClient);
      System.out.println();
      dataFAs = new Fetch_AckToBeSend((byte)1);
      dataFAs.sendFA(socketToClient);
      System.err.println("Error: The Fetch Session command was rejected by the Server.");
      System.err.println();
      closeSocket(socketToClient);

      return false;
    }

    // Determine the receiver's address that has the results of the session required

    receiverInetAddress = determineSeReAddress(arrayOfdataRS[0].mbz_iPVN, arrayOfdataRS[0].receiverAddress);

    // Open a TCP connection to the Session-Receiver

    Socket socketToReceiver = null;                // Represents the socket connection to a Receiver.

    System.out.println("connectionSetupRequest to receiver "+receiverInetAddress);
    System.out.println();

                         // connectionSetup(int mode, InetAddress ia, int port, int timeout)
    socketToReceiver = connectionSetupRequest(mode,receiverInetAddress,receiverTCPport,TIMEOUT);

    if(socketToReceiver == null){

      System.err.println("Error: The Session-Receiver specified doesn't exist.\n");
      System.err.println("Error: Error processing connection Setup");
      System.err.println();

      // Send the Fetch_Ack packet to the client

      System.out.println("Fetch-Ack to Client on "+socketToClient);
      System.out.println();
      dataFAs = new Fetch_AckToBeSend((byte)1);
      dataFAs.sendFA(socketToClient);
      System.err.println("Error: The Fetch Session command was rejected by the Server.");
      System.err.println();
      closeSocket(socketToClient);

      return false;
    }

    // Send packet Fetch-Session to receiver

    // Send the dataFS packet to the receiver ( In variable byte[] dataFS.buffer we have the packet dataFS received from the client,
    // and is the same packet that will be send to the session receiver)

    System.out.println("Fetch Session to Receiver on "+socketToReceiver);
    System.out.println();

    dataFS.sendTCP(socketToReceiver,dataFS.PACKET_LENGTH);    // send the fetch session message to the session receiver.

    // Receive the fetch-Ack packet

    System.out.println("Fetch-Ack from Receiver on "+socketToReceiver);
    System.out.println();

    Fetch_AckReceiver dataFA = new Fetch_AckReceiver();

    if(!dataFA.receiveFA(socketToReceiver)){
      System.err.println("Error: Error in receiving the packet Fetch-Ack from receiver on "+socketToReceiver);
      System.err.println();
      closeSocket(socketToReceiver);
      dataFA.buffer[0] = (byte)1;  // dataFA.accept = 1
    }

    // Send the Fetch_Ack packet to the client

    System.out.println("Fetch-Ack to Client on "+socketToClient);
    System.out.println();

    dataFA.sendTCP(socketToClient,dataFA.PACKET_LENGTH);

    boolean error = false;
    switch (dataFA.accept) {
      case 0: // The Session-Receiver accepts the FS
        break;
      case 1: // Failure, reason unspecified (catch-all).
        System.err.println(
            "Error: Message from receiver - Failure, reason unspecified (catch-all).");
        System.err.println();
        error = true;
        break;
      case 2: // Internal error.
        System.err.println(
            "Error: Message from receiver - Internal error.");
        System.err.println();
        error = true;
        break;
      case 3: // Some aspect of request is not supported.
        System.err.println(
            "Error: Message from receiver - Some aspect of request is not supported.");
        System.err.println();
        error = true;
        break;
      case 4: // Cannot perform request due to permanent resource limitations.
        System.err.println("Error: Message from receiver - Cannot perform request due to permanent resource limitations.");
        System.err.println();
        error = true;
        break;
      case 5: // Cannot perform request due to temporary resource limitations.
        System.err.println("Error: Message from receiver - Cannot perform request due to temporary resource limitations.");
        System.err.println();
        error = true;
        break;
      default: // Reserved values
        System.err.println(
            "Error: The server doesn't support the accept mode received from the Session-Receiver.");
        System.err.println();
        error = true;
    }

    if (error) { // If an error occurred
      // Fetch-Session request was rejected
      closeSocket(socketToClient);
      closeSocket(socketToReceiver);
      System.err.println("The Fetch-Session request was rejected.");
      System.err.println();
      return false;
    }


    // If Accept was 0, the server sends the Fetch ack to the client. The session receiver
    // then will send the OWAMP-Test session data in question, followed by 16 octets of
    // Integrity Zero Padding.

    // Send the OWAMP Test Session Data to the Fetch-Client.
    // The OWAMP-Test session data consists of the following (concatenated):


    // I - A reproduction of the Request-Session command that was used to start the session; it is modified
    // so that actual sender and receiver port numbers that were used by the OWAMP-Test session always appear
    // in the reproduction.

    System.out.println("OWAMP-Test session data to Client on " + socketToClient);
    System.out.println();

    // Send the complete reproduction of the RequestSession command

    arrayOfdataRS[0].sendTheCompleteRSreceived(arrayOfdataSSDAux,socketToClient);

    // II - Zero or more (as specified) Skip Range descriptions.

    if( dataFA.numSkipRanges != 0){
      Packets packet = new Packets(8);
      int readn = 0;
      for (int i = 0; i < dataFA.numSkipRanges; i++) {
        readn = receivePacket(socketToReceiver, packet.buffer, 0, 8); // receivePacket (Socket socket, byte[] _buffer, int off, int len)
        if (readn == 8) { // All Skip Range descriptions received
          packet.sendTCP(socketToClient,8);
        }else {
          closeSocket(socketToClient);
          closeSocket(socketToReceiver);
          System.err.println("Error: Error receiving the OWAMP-Test session data Skip Range descriptions in operation Fetch session.");
          System.err.println();
          return false;
        }
      }

      // Each Skip Range Descriptions has only 8 bytes, so the last (possibly full, possibly incomplete)
      // block (16 octets) of data MUST be padded with zeros, if necessary (numberOfSkipRanges is odd).
      // This ensures that the next session description record starts on a block boundary.
      if ((dataFA.numSkipRanges % 2) == 1) {
        readn = receivePacket(socketToReceiver, packet.buffer, 0, 8); // receivePacket (Socket socket, byte[] _buffer, int off, int len)
        if (readn == 8) { // A Skip Range descriptions received
          packet.sendTCP(socketToClient,8);
        }else {
          closeSocket(socketToClient);
          closeSocket(socketToReceiver);
          System.err.println("Error: Error receiving the OWAMP-Test session data Skip Range descriptions in operation Fetch session.");
          System.err.println();
          return false;
        }
      }

    }

    // III - 16 octets of IZP.

    // Receive the IntegrityZeroPadding packet

    IntegrityZeroPaddingReceiver izp = new IntegrityZeroPaddingReceiver(16);
    if (!izp.receiveIZP(socketToReceiver)) {
      closeSocket(socketToClient);
      closeSocket(socketToReceiver);
      System.err.println("Error: Error receiving the OWAMP-Test session data IntegrityZeroPadding in operation Fetch session.");
      System.err.println();
      return false;
    }

    // Send integrityZeroPadding packet to Client

    izp.sendTCP(socketToClient,izp.PACKET_LENGTH);

    // IV - Zero or more (as specified) packet records.
    // Receive each PacketRecords from Receiver and send them to Client

    OwampTestPacketRecordReceiver packetRecordReceiver = new
        OwampTestPacketRecordReceiver();

    for (int i = 0; i < dataFA.numRecords; i++) { // For each one of the ScheduleSlots Packets
      if (!packetRecordReceiver.receivePR(socketToReceiver)) {
        System.err.println("Error: Error receiving packet PacketRecord " +
                           (i + 1) + " in operation Fetch session.");
        System.err.println();
        closeSocket(socketToClient);
        closeSocket(socketToReceiver);
        return false; // All the packet records must be received properly.
      }
      else {
        packetRecordReceiver.sendTCP(socketToClient,packetRecordReceiver.PACKET_LENGTH);
      }
    }

    // V - 16 octets of IZP.

    // Receive the IntegrityZeroPadding packet

    if (!izp.receiveIZP(socketToReceiver)) {
      closeSocket(socketToClient);
      closeSocket(socketToReceiver);
      System.err.println("Error: Error receiving the OWAMP-Test session data IntegrityZeroPadding in operation Fetch session.");
      System.err.println();
      return false;
    }

    // Send integrityZeroPadding packet to Client

    izp.sendTCP(socketToClient,izp.PACKET_LENGTH);

    closeSocket(socketToClient);

    closeSocket(socketToReceiver);

    return true;

  }


  /**
   * This method is used to determine which packet was received and process it.
   * The first byte read represents the type of the packet.
   * @return boolean - It returns true if everything goes ok, otherwise it returns false.
   */

  public boolean processPacketReceived(){

    int readn = -1;   // number of bytes read
    switch((int)(buffer[0] & 0xFF)){  // First byte read, it represents the type of the packet.
      case 1:  //Request Session
        readn = receivePacket(socketToClient, buffer, 1, RequestSession.PACKET_LENGTH -1);   // receivePacket (Socket socket, byte[] _buffer, int off, int len)
        if(readn == (RequestSession.PACKET_LENGTH -1)){   // Packet Request Session
          short accept = 0;
          //int port = DEFAULT_RECEIVER_UDP_PORT;
          if (!requestSessionResponse(accept)) { // requestSession(short accept)
            System.err.println("Error: Error in Request-Session operation.");
            System.err.println();
            return false;
          }
          return true;
        }else
          System.err.println("Error: The packet received in socket "+socketToClient+" is not a Request-Session packet.\n");
        break;
      case 2:  // Starting Test Sessions
        readn = receivePacket(socketToClient,buffer,1,StartingTestSessions.PACKET_LENGTH -1);   // receivePacket (Socket socket, byte[] _buffer, int off, int len)
        if(readn == (StartingTestSessions.PACKET_LENGTH -1)){   // Packet Starting Test Sessions

          byte accept = 0;
          if (!startingTestSessionsResponse(accept)) {  // startingTestSessionsResponse(short accept)
            System.err.println("Error: Error in Starting Test Sessions operation.");
            System.err.println();
            return false;
          }
          return true;
        }else
          System.err.println("Error: The packet received in socket "+socketToClient+" is not a Starting Test Sessions packet.\n");
        break;
      case 3:  // Stop Sessions. If a SS is received before the Starting Test session command, this session should be finished.

        readn = receivePacket(socketToClient,buffer,1,StopSessions.PACKET_LENGTH -1);   // receivePacket (Socket socket, byte[] _buffer, int off, int len)
        if(readn == (StopSessions.PACKET_LENGTH -1)){   // Packet Stop Sessions
            // Create the stopSessions packet
          System.arraycopy(buffer,0,dataSSr.buffer,0,dataSSr.PACKET_LENGTH);
          dataSSr.create_Datagram_StopSessions();

          System.out.println("Stop Session from Client on "+socketToClient);
          System.err.println("Error: A Stop Session was received before Starting Test session command.");
          System.err.println("All sessions requested so far will be finished.");
          System.err.println();

          // The Server should stop all sessions and stop this OWAMP-Control session.

          if(dataSSr.accept != 1)
            StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)
          else
            StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,false); // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)

          closeSocket(socketToClient);

          return true;
        }else
          System.err.println("Error: The packet received in socket "+socketToClient+" is not a Stop Session packet.\n");
        break;
      case 4:   // Fetch-Session
        readn = receivePacket(socketToClient,buffer,1,FetchSession.PACKET_LENGTH -1);   // number of bytes read
        if(readn == (FetchSession.PACKET_LENGTH -1))   // Packet Fetch Session
          return processFetchSession();
        else
          System.err.println("Error: The packet received in socket "+socketToClient+" is not a Fetch-Session packet.\n");
        break;
      default:
        System.err.println("Error: The packet received doesn't exits in the protocol.");
    }

    // If we reach this point it means that the packet received doesn't respects the protocol or an error occurred.
    // So, we should send the Stop All Session to Client and stop this OWAMP-Control session.

    if(readn > 0)   // Packet received has a valid type but not a valid length.
      System.err.println("Error: The packet received doesn't exits in the protocol.\n");

    if(readn == 0)  // Timeout to wait occurred.
      System.err.println("Error: Timeout to wait for the packet occurred.\n");

    System.err.println("All sessions requested so far will be finished.");
    System.err.println();
    StopSessions.stopAllSessions(socketToClient,arrayOfSocketToReceiver,numRS,true);  // stopAllSessions(Socket socket, Socket[] arrayOfSocketToS_R, int numRS, boolean send)

    return false;

  }


  /**
   * This method is used to find the index in arrayOfdataRS of a session that has the same sid as the sid
   * that it was given.
   * @param sid byte[] - sid of the session to be find.
   * @return int - this method returns the index of the session if one of the elements of arrayOfdataRS has
   * the same sid, otherwise it returns -1.
   */

  public int findSession(byte[] sid){

    String str1 = new String(sid);  // Auxiliary string
    for(int i = 0; i<numRS; i++){
      if (str1.compareTo(new String(arrayOfdataRS[i].sid)) == 0){
        return i;
      }
    }

    return -1;

  }


  /**
   * Method used to record the arrayOfdataRS and the arrayOfdataSSD, for each RS, in a file to be used in
   * Fetch Session. For each RS we have a different file, which name is the session SID as a string.
   * In this case we can not have the same file to all sessions because, if the Servers is communicating
   * with more then one Client at same time, they would use the same file. Which can not be possible. The
   * only way to resolve this problem is to have different files, with different names.
   * @return boolean -  This method returns true if everything goes ok, otherwise it returns false.
   */

  public boolean arrayOfRStoFiles(){

    try{

      // Test serverRSdir directory, to prevent errors if a user deletes it. If it doesn't exist it should be created.
      if (!serverRSdir.exists())
        serverRSdir.mkdir();

      for(int i=0;i<numRS;i++){
        if (!arrayOfSocketToReceiver[i].isClosed()) {  // It means that this session was started
                                                       // record the arrayOfdataRS

          File outputFileName = new File(serverRSdir.getPath()+serverRSdir.separator+owampHexEncode(arrayOfdataRS[i].sid)+".rs");  // If outputFile already exists, delete it
          if(outputFileName.exists())
            outputFileName.delete();

          RandomAccessFile outputFile = new RandomAccessFile(outputFileName,"rw");  // Create a new outputFile
          if(!saveRStoFile(arrayOfdataRS[i],arrayOfdataSSD[i],outputFile)){
            outputFile.close();
            return false;
          }

          outputFile.close();
        }
      }

    }
    catch(IOException ex){
      System.err.println(ex);
      return false;
    }

    return true;
  }


  /**
   * Method used to read the dataRS and the arrayOfdataSSD from a file to be used in Fetch Session.
   * For each started session we have a file which name is the SID of the session.
   * @return ScheduleSlotDescriptions[] -  This method returns the ScheduleSlotDescriptions[] if everything goes ok,
   * otherwise it returns null.
   */

  public ScheduleSlotDescriptions[] dataRSfromFile(byte[] sid){

    // Test serverRSdir directory, to prevent errors if a user deletes it. If it doesn't exist return false.
    if (!serverRSdir.exists())
      return null;

    String strSid = owampHexEncode(sid);
    File inputFileName = new File(serverRSdir.getPath()+serverRSdir.separator+strSid+".rs");  // Test if inputFile exists.
    if(!inputFileName.exists()){
      System.err.println("Error: There is no Sessions to Fetch (the Server's file of RS doesn't exits).");
      return null;
    }

    ScheduleSlotDescriptions[] arrayOfdataSSDAux = null;
    RandomAccessFile inFile = null;    // file with the RS made (RS is the combination of the arrayOfdataRS and the arrayOfdataSSD).

    try{

      inFile = new RandomAccessFile(inputFileName,"r");

      // read the arrayOfdataRS

      arrayOfdataRS[0] = new RequestSessionReceiver();

      arrayOfdataSSDAux = readRSfromFile(arrayOfdataRS[0],inFile);
      if(arrayOfdataSSDAux == null){
        inFile.close();
        return null;
      }

      inFile.close();

    }catch(IOException ex){
      System.err.println(ex);
      return null;
    }finally{
      try {
        inFile.close();
      }catch (IOException ex) {
        System.err.println(ex);
      }
    }

    return arrayOfdataSSDAux;

  }


  // Main function

  public static void main(String[] args) throws IOException{

    InetAddress serverAddress = null; // The local IP address to bind to. The serverAddress can be used on a
                                      // multi-homed host for a ServerSocket that will only accept connect requests
                                      // to one of its addresses. If serverAddress is null, it will default
                                      // accepting connections on any/all local addresses.
    int serverTCPportNumber = -1,     // TCP port number to be used in OWAMP-Control.
        receiverTCPport = -1;         // TCP port number to be used in OWAMP-Control with all OWAMP Session-Receivers.
    String  ntpServerAddress = null;  // The address of the NTP server to be used to get the time reference.
                                      // If not defined the system time will be used to get the time reference.
    int updateInterval = -1;          // The interval duration (in minutes) between two consecutives updates to time reference.

    if((args.length == 1)&&((args[0].compareToIgnoreCase("-?") == 0)||(args[0].compareToIgnoreCase("-help") == 0))){  // Print the help message
      System.out.println("\nUsage: OWAMP_Server [-options]\n");
      System.out.println("where options include:");
      System.out.println("\n   -a | -address             \tThe local IP address to bind to. This address can be used on a multi-homed host");
      System.out.println("                               \tfor a Server that will only accept connection requests to one of its addresses.");
      System.out.println("                               \tIf not defined, it will by default accept connections on any/all local addresses.");
      System.out.println("\n   -p | -port port           \tThe TCP port number to be used in OWAMP-Control.");
      System.out.println("\n   -r | -receiverTCPport port\tTCP port number to be used in OWAMP-Control with all OWAMP Session-Receivers.");
      System.out.println("\n   -n | -ntpServer           \tThe address of the NTP server to be used to get the time reference. If not defined,");
      System.out.println("                               \tthe system time will be used to get the time reference.");
      System.out.println("\n   -u | -updateInterval      \tThe value (in minutes) to be used for the time interval between two consecutives");
      System.out.println("                               \tupdates of the time reference.");
      System.out.println("\n   -? | -help                \tPrints this help message.");
      System.exit(1);
    }

    if(args.length == 0){
      System.out.println("\nDefault port will be used on OWAMP-Control to this Server ("+DEFAULT_SEVER_PORT+").");
      System.out.println("\nDefault port will be used on OWAMP-Control between this Server and all OWAMP Session-Receivers ("+DEFAULT_RECEIVER_PORT+").");
      serverTCPportNumber = DEFAULT_SEVER_PORT;
      receiverTCPport = DEFAULT_RECEIVER_PORT;
    }else{
      try{

        int numOfArg = 0; // Used to determine if all arguments are ok.

        // option -address (It's the local address to bind to)
        for (int i = 0; i < args.length; ) {
          if ((args[i].compareToIgnoreCase("-a") == 0)||(args[i].compareToIgnoreCase("-address") == 0)) {
            try{ // Test if the server address introduced is a valid address.
              serverAddress = InetAddress.getByName(args[i + 1]);
              if(!OWAMP_Base.isAnyLocalAddress(serverAddress)){
                System.err.println("Error: The OWAMP server's address read is not a valid local address (" + args[i + 1] + ").");
                System.exit(1);
              }
            }catch (Exception e) {
              System.err.println(
                  "Error: The OWAMP server's address read is not a valid address (" +
                  args[i + 1] + ").");
              System.exit(1);
            }
            numOfArg++;
            break;
          }
          i += 2;
        }

        // option -port
        for (int i = 0; i < args.length; ) {
          if ((args[i].compareToIgnoreCase("-p") == 0)||(args[i].compareToIgnoreCase("-port") == 0)) {
            serverTCPportNumber = Integer.parseInt(args[i + 1]);
            if (serverTCPportNumber < 0 || serverTCPportNumber >= 65536) {
              System.err.println("Error: Server TCP port number out of range (" +
                                 serverTCPportNumber + ").");
              System.err.println();
              System.exit(1);
            }
            numOfArg++;
            break;
          }
          i += 2;
        }

        // option -receiverTCPport
        for (int i = 0; i < args.length; ) {
          if ((args[i].compareToIgnoreCase("-r") == 0)||(args[i].compareToIgnoreCase("-receiverTCPport") == 0)) {
            receiverTCPport = Integer.parseInt(args[i + 1]);
            if (receiverTCPport < 0 || receiverTCPport >= 65536) {
              System.err.println("Error: Receiver TCP port number out of range (" +
                                 receiverTCPport + ").");
              System.err.println();
              System.exit(1);
            }
            numOfArg++;
            break;
          }
          i += 2;
        }

        // option -ntpServer
        for (int i = 0; i < args.length; ) {
          if ((args[i].compareToIgnoreCase("-n") == 0)||(args[i].compareToIgnoreCase("-ntpServer") == 0)) {
            ntpServerAddress = args[i + 1];
            try{ // Test if the NTP server address introduced is a valid address.
              if(InetAddress.getByName(ntpServerAddress).isLoopbackAddress()){ // The IP addresse for a given host name.
                System.err.println(
                    "Error: The NTP server's address read is not a valid address (" +
                    InetAddress.getByName(ntpServerAddress) + ").");
                System.exit(1);
              }
            }catch (Exception e) {
              System.err.println(
                  "Error: The NTP server's address read is not a valid address (" +
                  ntpServerAddress + ").");
              System.exit(1);
            }
            numOfArg++;
            break;
          }
          i += 2;
        }

        // option -updateInterval
        for (int i = 0; i < args.length; ) {
          if ((args[i].compareToIgnoreCase("-u") == 0)||(args[i].compareToIgnoreCase("-updateInterval") == 0)) {
            updateInterval = Integer.parseInt(args[i + 1]);
            if (updateInterval < 0) {
              System.err.println("Error: The value of updateInterval can not be a negative number (" +
                                 updateInterval + ").");
              System.err.println();
              System.exit(1);
            }
            numOfArg++;
            break;
          }
          i += 2;
        }

        if((numOfArg<(double)args.length/2)||(serverAddress == null)&&(serverTCPportNumber==-1)&&(receiverTCPport==-1)&&(ntpServerAddress==null)&&(updateInterval==-1))
          throw new IllegalArgumentException();

        if(serverTCPportNumber==-1){ // Only receiverTCPport option was used.
          System.out.println("\nDefault port will be used on OWAMP-Control to this Server ("+DEFAULT_SEVER_PORT+").");
          serverTCPportNumber = DEFAULT_SEVER_PORT;
        }

        if(receiverTCPport==-1){ // Only port option was used.
          System.out.println("\nDefault port will be used on OWAMP-Control between this Server and all OWAMP Session-Receivers ("+DEFAULT_RECEIVER_PORT+").");
          receiverTCPport = DEFAULT_RECEIVER_PORT;
        }else
          System.out.println("\nTCP port "+receiverTCPport+" will be used on OWAMP-Control between this Server and all OWAMP Session-Receivers.");

      }catch (Exception e){
        System.err.println("\nIllegalArgumentException in thread main: Usage: OWAMP_Server [-options]\n");
        System.err.println("where options include:");
        System.err.println("\n   -a | -address             \tThe local IP address to bind to. This address can be used on a multi-homed host");
        System.err.println("                               \tfor a Server that will only accept connection requests to one of its addresses.");
        System.err.println("                               \tIf not defined, it will by default accept connections on any/all local addresses.");
        System.err.println("\n   -p | -port port           \tThe TCP port number to be used in OWAMP-Control.");
        System.err.println("\n   -r | -receiverTCPport port\tTCP port number to be used in OWAMP-Control with all OWAMP Session-Receivers.");
        System.err.println("\n   -n | -ntpServer           \tThe address of the NTP server to be used to get the time reference. If not defined,");
        System.err.println("                               \tthe system time will be used to get the time reference.");
        System.err.println("\n   -u | -updateInterval      \tThe value (in minutes) to be used for the time interval between two consecutives");
        System.err.println("                               \tupdates of the time reference.");
        System.err.println("\n   -? | -help                \tPrints this help message.");
        System.exit(1);
      }
    }

    System.out.println("\nStarting on port "+serverTCPportNumber);
    System.out.println();

    try{
      ServerSocket serverSocket = null;
      if(serverAddress == null)
        serverSocket = new ServerSocket(serverTCPportNumber); // initialize a server socket on port 'port'
      else
        serverSocket = new ServerSocket(serverTCPportNumber,0,serverAddress); // initialize a server socket on port 'port'

      // The hiResTimer should be definied here to guarante that
      // all Server instantiation has same time reference. It will
      // be passed has an argument of the Server constructor.
      HiResTimer hiResTimer = null;
      if(updateInterval == -1){  // Use default updateInterval defined on HiResTimer class.
        hiResTimer = new HiResTimer(ntpServerAddress);
      }else
        hiResTimer = new HiResTimer(ntpServerAddress,updateInterval);

      Timestamp upTime = new Timestamp(hiResTimer.milliTime(), (short)1);
                                             // Uptime is a timestamp representing the time when the
                                             // current instantiation of the server started operating, in
                                             // milliseconds since 1Jan1970.

      // Test serverRSdir directory. If it exists it should be cleaned (delete it and create it again). This
      // directory will have files, each with a RS made to this Server during a test session.

      if(OWAMP_Server.serverRSdir.exists())
        try{
          OWAMP_Server.deleteDir(OWAMP_Server.serverRSdir);
        }
        catch(IOException ex){
          ex.printStackTrace();
        }

      OWAMP_Server.serverRSdir.mkdir();

      // wait for requests connections

      while(true){
        Socket client = serverSocket.accept();                                              // Accept the client socket connection.
        OWAMP_Server server = new OWAMP_Server(client,upTime,receiverTCPport, hiResTimer);  // deliver this socket to the server
                                                                                            // OWAMP_Server(Socket socketToClient, Timestamp upTime, int receiverTCPport, HiResTimer hiResTimer)
        server.mode = 1;          // Mode of authentication.
        server.start();           // initialize thread.
      }
    }
    catch (SocketException se){
      System.err.println(se);
    }

  }

}

